/* PALS5.C								*/
/* fichier contenant les entetes et les corps de fonctions concernant	*/
/* la gestion des palindromes.						*/
/* 03/09/94								*/
/* 21/02/95	   							*/


#include <stdio.h>

#include "pals5.h"

/* definition des fonctions */


/* verifie que 2 palindromes P1 et P2 sont juxtaposes et qu'il   */
/* n'existe aucun palindrome juxtapose entre			  */
int  Juxta0(int *Mjuxta[],int i, int j)

     {
        int t ;

	if (Mjuxta[i][j] == 1) {
	    t = i+1 ;
            while (t < j && !(Mjuxta[i][t] && Mjuxta[t][j]))
                   t++ ;
            return (t==j) ;

        }
        else
            return 0 ;
     }


  /* verifie la juxtaposition de 2 palindromes */
int Juxtapose(int Tolerance,Palindrome P1,Palindrome P2)
    {
      return (P1.Fqueue - Tolerance < P2.Dtete) ;
    }


 /* verifie l'imbrication d'un palindrome dans un autre */
int Imbrique(int Tolerance,Palindrome P1, Palindrome P2)
    {
       return (P1.Ftete - Tolerance < P2.Dtete &&
	       P2.Fqueue - Tolerance < P1.Dqueue) ;
    }


 /* verifie si la tete de P2 est incluse dans la boucle de P1 */
int Quinconce(int Tolerance,Palindrome P1, Palindrome P2)
    {
       return (P1.Ftete - Tolerance < P2.Dtete &&
	       P2.Ftete - Tolerance < P1.Dqueue &&
	       P1.Fqueue - Tolerance < P2.Dqueue) ;
    }

 /* verifie si un palindrome contient au moins une liaison impure */
int Impure(Palindrome P1)
    {
      return (P1.Code == 0) ;
    }

void PrintPal(FILE * sortie,Palindrome P,chaine suffixe)
{

   fprintf(sortie,"[%d-%d-%d-%d-%d]",
	   P.Dtete,P.Ftete,
	   P.Dqueue,P.Fqueue,
	   P.Code) ;
   fprintf(sortie,suffixe) ;
}


/* verifie l'egalite de 2 palindromes	*/
int EgalPal(Palindrome P1, Palindrome P2)
{
   return (P1.Dtete == P2.Dtete &&
	   P2.Ftete == P2.Ftete &&
	   P1.Dqueue == P2.Dqueue &&
	   P1.Fqueue == P2.Fqueue &&
	   P1.Code == P2.Code
	  ) ;
}




/* verifie que 2 palindromes sont compatibles (strictement disjoints) */
/* Compatible accepte les pseudo-noeuds (Quinconce)	       	    */
int Compatible(int Tolerance, Palindrome P1, Palindrome P2)

{
    return(
	    Place(Tolerance, P2.Dtete,P2.Ftete,P1) &&
	    Place(Tolerance, P2.Dqueue,P2.Fqueue,P1) 
          ) ;
}




/* verifie que 2 palindromes sont compatibles (strictement disjoints) */
/* Compatible2 n'accepte pas les pseudo-noeuds (Quinconce)	      */
int Compatible2(int Tolerance, Palindrome P1, Palindrome P2)

{   
    int a ; 
 
      return(Place(Tolerance, P2.Dtete,  P2.Ftete,P1) &&
	     Place(Tolerance, P2.Dqueue, P2.Fqueue,P1) &&
             !Quinconce(Tolerance,P1,P2) &&
             !Quinconce(Tolerance,P2,P1)
            ) ;
 }


   
    



int Place(int Tolerance, int Debut, int Fin, Palindrome P)
{
   return(
	 P.Fqueue < Debut + Tolerance ||
	 (P.Ftete < Debut + Tolerance &&  Fin < P.Dqueue + Tolerance) ||
	 Fin < P.Dtete + Tolerance
	) ;
}   







/* verifie qu'un ensemble de palindromes est compatibles 2 a 2 		*/
/*
int Compatibilite(int Tolerance, Palindrome P[], int n)
{
   int i ;
   int k ;
   int ok ;

   ok = 1 ;
   i = 0 ;
   while (ok && i < n-1) {
	 k = i + 1 ;
	 while (ok && k < n)
	       if (!Compatible(int Tolerance, P[i],P[k]))
		  ok = 0 ;
	       else
		  k++ ;
	 i++ ;
   }
   return ok ;
}

*/


/* lit une sequence  AVEC SCORE 					*/
int  LireSeqScore(FILE *entree, int Distingue, int* DebPart, int* FinPart,
                  Palindrome Pals[],char *Seq,int *NbPals)

{
  char sep[5]  ;
  int j        ;
  int k        ;

   if (Distingue)   
      fscanf(entree,"%d %d", DebPart, FinPart) ;
   
   fscanf(entree,"%s",Seq) ;

   (*NbPals) = 0 ;

   if (!feof(entree)) 
   {

      while (fscanf(entree,"%d %d %d %d %d %d",
		    &(Pals[*NbPals].Dtete), &(Pals[*NbPals].Ftete),
		    &(Pals[*NbPals].Dqueue),&(Pals[*NbPals].Fqueue),
		    &(Pals[*NbPals].Code),&(Pals[*NbPals].Score))
	    )
	     (*NbPals)++ ;
      fscanf(entree,"%s",sep) ;
      return 1 ;
   }
   else
      return 0 ;

}




/* lit une sequence  AVEC SCORE 					*/
int  LireSeqScoreS(FILE *entree,Palindrome Pals[],char *Seq,int *NbPals)

{
  char sep[5]  ;
  int j        ;
  int k        ;

				/* creation 4 palindromes artificiels	*/
				/* representant structure 2110		*/

   (*NbPals) = 4 ;

   fscanf(entree,"%s",Seq) ;
   if (!feof(entree)) {

      while (fscanf(entree,"%d %d %d %d %d %d",
		    &(Pals[*NbPals].Dtete), &(Pals[*NbPals].Ftete),
		    &(Pals[*NbPals].Dqueue),&(Pals[*NbPals].Fqueue),
		    &(Pals[*NbPals].Code),&(Pals[*NbPals].Score))
	    )
	     (*NbPals)++ ;
      fscanf(entree,"%s",sep) ;
      return 1 ;
   }
   else
      return 0 ;

}
