/* ==================================================== */
/* Gloup Jul. 92 					*/
/* Copyright (c) Atelier de BioInformatique		*/
/* @file: libkbab1.c					*/
/* KBranch&Bound (KBaB) library				*/
/* Version etendue a la sauvegarde des K meilleurs	*/
/* solutions						*/
/* Fevrier 1995						*/
/* ==================================================== */
 
#include <stdio.h>

			/* inclusions  		*/

#include  "../../include/gtypes.h"
#include  "../../include/libset.h"
#include  "../../include/libkbab.h"



/* ------------------------------------------------------------ */
/* recherche dans l'ensemble L (non vide), l'element de plus 	*/
/* fort poids. ord est le vecteur des index des elts tries	*/
/* par poids decroissants. renvoi l'index de cet elt.		*/
/* ------------------------------------------------------------ */

static Int32 FindMaxElt(L, ord, i0)
   SetPtr L;
   Int32  *ord, i0;
{
   Int32 i;

   for (i = i0, ord += i0 ; i < L->dim ; i++) {
	if (BelongSet(L, *ord++))
	    return i;
   }

   return -1;   /* huh ? */
}




/* ------------------------------------------------------------ */
/* Recherche la place de la nouvelle solution			*/
/* ------------------------------------------------------------ */

void  InsererSolution(K,nK,PoidSol,Solution,Wb,Eb,IndexPoids)
      Int32     K		;
      Int32 	*nK		;
      Int32	PoidSol   	;
      SetPtr    Solution	;
      Int32     Wb[]		;
      SetPtr    Eb[]	        ;
      Int32     IndexPoids[]	;  /* vecteur des indices des   */
				   /* solution par poids decroi.*/

{
      Int32 j       ;
      Int32 t       ;
      Int32 Indice  ;
      Int32 Sortant ;

      j = 0 ;

      while ((j < *nK) && (Wb[IndexPoids[j]] >= PoidSol))
	     j++ ;

	     /* j pointe sur le premier poids inferieur a Poids	*/
	     /* si j vaut nK => Poids est inferieur a tous	*/
	     /* donc nouvel element				*/

      if (*nK == K)
	     /* on ne peut plus inserer de solution		*/
	     /* la plus faible solution est sortante		*/
	 Sortant = IndexPoids[K-1] ;

      for (t = *nK - 1 ; t >= j ; t--)

	     /* on decale les elements de plus faible poids	*/

	     if (t < K - 1)
		IndexPoids[t+1] = IndexPoids[t] ;

      if (*nK < K) {
	    /* on peut ajouter un element 			*/
	 Indice = *nK        ;
	 (*nK)++             ;
      }
      else
	 Indice = Sortant ;

      IndexPoids[j] = Indice       ;
      Wb[Indice] = PoidSol         ;
      CopySet(Eb[Indice],Solution) ;

}



/* ------------------------------------------------------------ */
/* procedure recursive KBabar (c'est elle qui fait tout)	*/
/*                      parametres ---------------------------- */
/* K  : nombre de solutions a conserver				*/
/* nK : nombre de meilleurs solutions deja retenues		*/
/* k  : profondeur						*/
/* Ek : ensemble des elements deja retenus			*/
/* wk : poids de l'ensemble Ek					*/
/* Lk : ensemble des elements compatibles avec Ek		*/
/* Eb : ensemble de plus fort poids trouve			*/
/* wb : poids de l'ensemble Eb					*/
/* mx    : nombre max d'elements dans la solution		*/
/* Poids : vecteur des poids de tous les elements		*/
/* Order : vecteur des index tries par poids decroissants	*/
/* ok    : index courant dans Order				*/
/* Rel   : matrice de la relation de compatiblite		*/
/*         sous la forme d'un vecteur d'ensembles (cf BaB)	*/
/* IndexPoids : vecteur des indices des solutions classees	*/
/*              par ordre decroissant de poids			*/
/* ------------------------------------------------------------ */

static void KBabar(K,nK,k, Ek, wk, Lk, Eb, wb, mx, Poids, Order, ok, Rel,
		   IndexPoids)
   Int32   *nK,k, wk, wb[], mx   ;
   Int32   *Poids, *Order        ;
   SetPtr  Ek, Lk, Eb[], *Rel    ;
   Int32   IndexPoids[]          ;
{
   Int32   ikk, pkk, wkk;
   SetPtr  Ekk, Lkk;



   Ekk = NewSet(Ek->dim);		/* nouvel ens Ek+1	*/
   CopySet(Ekk, Ek);			/* Ek+1 = Ek 		*/

   Lkk = NewSet(Lk->dim);		/* nouvel ens Lk+1      */

   /* verification si Lkk est vide : auquel une feuille a ete 
      atteinte a l'appel precedent 				*/
   while (! IsEmptySet(Lk)) 
   {

	ok  = FindMaxElt(Lk, Order, ok);/* element de plus 	*/
	ikk = Order[ok];		/* fort poids dans Lk	*/
	pkk = Poids[ikk];		/* = ikk	 	*/

	

	if (*nK == K)                    /* deja K solutions    */

	   if ((wk + (mx - k) * pkk) <= wb[IndexPoids[K-1]]) 	/* impasse	*/
	       break;

	AddElt(Ekk, ikk);		/* Ek+1 = Ek u {ikk}	*/
	InterSet(Lkk, Lk, Rel[ikk]);	/* Lk+1 = ens des i c Lk*/
	RemElt(Lkk, ikk);		/* compatibles avec ikk	*/

	wkk = wk + pkk;

	if ((*nK < K) || (wkk > wb[IndexPoids[K-1]])) {	/* meilleure config.	*/
	    InsererSolution(K,nK,wkk,Ekk,wb,Eb,IndexPoids) ;
	}

					/* recursion si pas mx	*/
					/* elts dans solution	*/

					/* ATTENTION methode 	*/
					/* exacte s'il n'existe */
					/* pas solution de plus */
					/* de mx elts		*/
					/* sinon heuristique	*/


	if (k + 1 < mx)
	   KBabar(K,nK,k+1, Ekk, wkk, Lkk, Eb, wb, mx, Poids,
		  Order, ok, Rel,IndexPoids) ;

					/* retour de recursion	*/
	RemElt(Ekk, ikk);		/* Ek+1 = Ek		*/

	RemElt(Lk, ikk);		/* retrait ikk		*/
    }

    FreeSet(Ekk);			/* liberation memoire   */
    FreeSet(Lkk);
}

/* ------------------------------------------------------------ */
/* Construction d'un vecteur d'index des elts tries par poids	*/
/* decroissants	(p = vecteur des poids)			       	*/
/* ------------------------------------------------------------ */
static int compare(i,j)	/* fonction de comparaison pour qsort	*/
Int32 *i, *j;
{
     return(*j - *i);
}


static Int32 * BuiltOrder(p, n)	/* fonction de construction	*/
   Int32 *p, n;
{
   Int32 i, j, imax, *ord, *tmp;

   tmp = NEWN(Int32, 2*n);
   ord = NEWN(Int32, n);

   for (i = j = 0 ; i < n ; i++) {
	tmp[j++] = p[i];
	tmp[j++] = i;
   }

   qsort(tmp, n, 2*sizeof(Int32), compare);	/* quick sort	*/

   for (i = 0, j = 1 ; i < n ; i++, j += 2)
	ord[i] = tmp[j];

   FREE(tmp);

   return ord;
}



/* ------------------------------------------------------------ */
/* Procedure appel B&B						*/
/* ceci est la seule fonction publique				*/
/*                      parametres ---------------------------- */
/* Poids  : vecteur des poids de tous les elements (entiers)	*/
/* N      : nombre d'elements					*/
/* bool   : matrice (char) de la relation de compatibilte entre	*/
/*	   elements (NxN chars)					*/
/* MaxElt : nombre maxi d'elements admissible dans la solution	*/
/* Wbest  : (Int32 *) recoit le poids de la solution trouvee	*/
/* Best   : (Int32[]) recoit les index des elts de la solution	*/
/* NbElt  : NbElt[k] recoit le nombre d'element de la solu. k	*/
/* ------------------------------------------------------------ */

int KBaB( K,nK,Poids, N, bool, MaxElt, Wbest, Best, NbElt)
                 
         Int32  K,*nK,N, MaxElt, *Poids, Wbest[], *Best[] , NbElt[] ;
         Bool   *bool;
{

   Int32   i, j, k,*Order        ;
   SetPtr  E0, L0, *Eb, *Erel    ;
   Int32   *IndexPoids           ;
   Int32   *Wbest2		 ;

   fprintf(stderr,"debut de la recherche...\n") ;

				/* ---------------------------- */
				/* construction des index tries */
				/* ---------------------------- */
   Order = BuiltOrder(Poids, N);

				/* ---------------------------- */
				/* construction de la matrice	*/
				/* des relations		*/
				/* ---------------------------- */
   Erel = NEWN(SetPtr, N);
   if (!Erel) {
      fprintf(stderr,"Probleme d'allocation pour Erel\n") ;
      return 0 ;
   }

   for (i = 0 ; i < N ; i++) {
	Erel[i] = NewSet(N);
        for (j = 0 ; j < N ; j++)
	    if (*bool++)
		AddElt(Erel[i], j);
   }
				/* construction de l'index des  */
				/* poids			*/
   IndexPoids = NEWN(Int32,K) ;
   if (!IndexPoids) {
      fprintf(stderr,"Probleme d'allocation pour IndexPoids\n") ;
      return 0 ;
   }

				/* ---------------------------- */
				/* construction des ensembles	*/
				/* E0 , Eb et L0		*/
				/* ---------------------------- */
   E0 = NewSet(N);		/* E0 = {}			*/

   Eb = NEWN(SetPtr,K) ;	/* Eb = {}			*/
   if (!Eb) {
      fprintf(stderr,"Probleme d'allocation pour Eb\n") ;
      return 0 ;
   }
   for (k = 0 ; k < K ; k++)
       Eb[k] = NewSet(N) ;

   L0 = NewSet(N);		/* L0 = E			*/

   for (i = 0 ; i < N ; i++)
	AddElt(L0, i);

				/* ---------------------------- */
				/* appel de Babar		*/
				/* ---------------------------- */

   Wbest2 = NEWN(Int32,K) ;
   if (!Wbest2) {
      fprintf(stderr,"Probleme d'allocation pour Wbest2\n") ;
      return 0 ;
   }

   for (k = 0 ; k < K ; k++)
       Wbest2[k] = 0  ;


   KBabar(K,nK,0,E0, 0, L0, Eb, Wbest2, MaxElt, Poids, Order, 0, Erel,
	  IndexPoids) ;


   fprintf(stderr,"fin de la recherche\n") ;

				/* ---------------------------- */
				/* transfert Resultat		*/
				/* ---------------------------- */
   for (k = 0 ; k < *nK ; k++) {
				/* pour chaque solution		*/

	Wbest[k] = Wbest2[IndexPoids[k]] ;
	NbElt[k] = 0			 ;

	for (i = 0 ; i < N ; i++)
	    if (BelongSet(Eb[IndexPoids[k]], i))  {
		Best[k][i] = 1 ;
		(NbElt[k])++   ;
	    }
	    else
		Best[k][i] = 0 ;
   }

				/* ---------------------------- */
				/* liberation memoire et fin	*/
				/* ---------------------------- */
   for (i = 0 ; i < N ; i++)
	FreeSet(Erel[i]);

   FREE(Erel);

   FREE(Order);

   FreeSet(E0);
   FreeSet(L0);

   for (k = 0 ; k < K ; k++)
	FreeSet(Eb[k]) ;

   FREE(Eb) ;

   FREE(Wbest2) ;

   FREE(IndexPoids) ;

   fprintf(stderr,"fin de la procedure\n") ;

   return 1 ;
}   
