/* ============================================================ */
/*	   							*/
/* fichier solkbab1.c						*/
/* isole les K meilleures solutions (ensembles de palindromes	*/
/* compatibles) de chaque sequence 				*/
/* avec Branch & Bound (version les K meilleures)		*/
/* en fonction de scores issus du modele energetique de SALSER	*/
/* Decembre 1994						*/
/* Revision Septembre 1995					*/
/* Revision Mars 96						*/
/* introduit compatibilite faible 				*/
/* 								*/
/*								*/
/* Arguments d'appels du main :					*/
/*								*/
/* valeurs par defaut entre parentheses ;			*/
/*								*/
/* -s<entier> : nombre de sequences a traiter   (1)		*/
/*								*/
/* -t<entier> : nombre max d'elements dans une solution	(15)	*/
/*               (intervient dans le Bound => peut donner	*/
/*               des solutions contenant plus d'elements	*/
/*								*/
/* -j<entier> : nombre de solutions a rechercher  (200) 	*/
/*								*/
/* -p<entier> : nombre max de pal dans une sequence  (500)	*/
/*								*/
/* -n         : interdit les pseudo-noeuds			*/								/*								*/
/*								*/
/* -r         : introduit la compatibilite faible		*/
/*								*/
/* -k         : lire debut et fin zone particuliere de l'ARN 	*/
/*								*/
/* ============================================================ */




#include <stdlib.h>
#include <stdio.h>

		/* fichiers inclus 				*/

#include "../../include/param.h"
#include "../../include/gtypes.h"
#include "../../include/pals5.h"
#include "../../include/libkbab.h"

#define LONGSEQ 500
#define MAXNOM 30



/* ==================================================================== */
/* 			  MACROS					*/
/* ==================================================================== */

#define PointeurNul(a,b) if(!a) {fprintf(stderr,"Probleme allocation sur %s\n",b) ; return 0 ; }
 
#define ARGUMENT(a,b,c) a = atoi(c) ;if (!a) {fprintf(stderr,"Mauvais argument -%c %s\n",b,c); return 0 ;}


/* ==================================================================== */
/* 	               VARIABLES GLOBALES				*/
/* ==================================================================== */

 int     PseudoNoeuds = 1 ;        /* pseudo-noeuds pas interdits       */



/* ==================================================================== */
/* 	               FONCTIONS LOCALES				*/
/* ==================================================================== */


/* -------------------------------------------------------------------- */
/* ecrit les solutions de la sequence courante dans le fichier 		*/
/* -------------------------------------------------------------------- */

static void SauvegardeSolutions(sortie, Distingue, DebPart,FinPart,
                                Seq,NbPals,Pals,nK,Best,NbElt)
       FILE       *sortie 	;
       int        Distingue	;
       int 	  DebPart	;
       int 	  FinPart	;
       char       *Seq		;
       Int32      NbPals        ;
       Palindrome Pals[]	;
       Int32      nK		;
       Int32      *Best[]	;
       Int32      NbElt[]	;
{

   Int32 i ;
   Int32 k ;


  /* ecrit debut et fin zone particuliere de l'ARN			*/

  if (Distingue)
     fprintf(sortie,"%d %d\n", DebPart, FinPart) ;


   /* ecrit la sequence							*/

   fprintf(sortie,"%s\n",Seq) ;

   /* ecrit le nombre de palindromes dans la sequence			*/

   
   fprintf(sortie,"%d\n",NbPals) ;

   /* ecrit les palindromes 						*/

   for (i = 0 ; i < NbPals ; i++)
       fprintf(sortie,"%d %d %d %d %d %d ",
		      Pals[i].Dtete  , Pals[i].Ftete,
		      Pals[i].Dqueue , Pals[i].Fqueue,
		      Pals[i].Code   , Pals[i].Score   ) ;

   fprintf(sortie,"\n") ;


   /* ecrit le nombre de solutions		*/

   fprintf(sortie,"%d\n",nK) ;

   for (k = 0 ; k < nK ; k++) 
   {
       /* pour chaque solution		*/

       /* ecrit le nombre de pals dans la solution			*/

       fprintf(sortie,"%d\n",NbElt[k]) ;

       /* ecrit les indices des pals de la solution			*/

       for (i = 0 ; i < NbPals ; i++)
	   if (Best[k][i])
	       fprintf(sortie,"%d ",i) ;

       fprintf(sortie,"\n") ;

    } 
}



/* -------------------------------------------------------------------- */
/* Creer la matrice de compatibilite entre les pals d'une solution	*/
/* -------------------------------------------------------------------- */
static void MatriceCompatibilite(CompatibiliteFaible,bool,NbPals,Pals)
       int        CompatibiliteFaible ;
       Bool       *bool		;
       Int32      NbPals	;
       Palindrome Pals[]        ;
{
       Int32 i			;
       Int32 j			;

       Int32 m = 0		;


       for (i = 0 ; i < NbPals ; i++)
	   for (j = 0 ; j < NbPals ; j++) 
           {
               
               if (PseudoNoeuds)
	           bool[m] = Compatible(CompatibiliteFaible, Pals[i],Pals[j]) ;
               else
                   bool[m] = Compatible2(CompatibiliteFaible, Pals[i],Pals[j]) ;

	       m++ ;
	   }
}





int main (argc,argv)
    int argc	    ;
    char *argv[]    ;
{
   Palindrome *Pals             ;  /* Palindromes de chaque sequence    */
   char Seq[LONGSEQ]       	;  /* sequence				*/
   int  MaxPal = 2000		;  /* nombre max pal dans sequence	*/
   int  NbPals  	   	;  /* nb palindromes dans sequence	*/
   int  Nseq	           	;  /* nb sequences traitees		*/
   Bool         *bool           ;  /* matrice de compatibilite		*/
   Int32 	*NbElt		;  /* vecteur nb elements par solution  */
   Int32        *Wbest          ;  /* vecteur poids meilleures solu.	*/
   Int32        *Poids          ;  /* vecteur des poids des palindro.   */
   Int32       **Best           ;  /* vecteur d'indices des elts	*/
				   /* appartenant a meilleures solu.	*/
   Int32        nK              ;  /* nb meilleures solutions retenues  */
   Int32        k               ;  /* indice des solutions		*/
   Int32        i               ;  /* indice des palindromes		*/
   Int32      K      = 200      ;  /* nombre de solutions recherchees   */
   Int32      MaxElt = 15       ;  /* nb max d'elements par solution    */
				   /* valeur par defaut			*/
   Int32      NbSeq  =  1	;  /* nombre de sequence a traiter	*/
   char       Symb              ;  /* symbole argument de main		*/
   char      Argument[50]	;  /* argument de main			*/
   int       CompatibiliteFaible = 0 ; /* compatibilite forte par defaut */
   int       Distingue		 = 0 ; /* 0 si pas zone particuliere	*/
   int       DebPart		 = -1 ;
   int       FinPart		 = -1 ;
  


   /* debut du programme principal */


				   /* recherche des arguments de main	*/

   
    while (SelectArg("-k -s -t -p -n -r -j",&Symb,Argument,argc,argv)) 
    {

	  switch (Symb) 
          {
		 case 't' : ARGUMENT(MaxElt,Symb,Argument)      ;
			    break	                        ;
		 case 's' : ARGUMENT(NbSeq,Symb,Argument)       ;
			    break	                        ;
		 case 'j' : ARGUMENT(K,Symb,Argument)           ;
			    break	                        ;
		 case 'p' : ARGUMENT(MaxPal,Symb,Argument)      ;
			    break	                        ;
                 case 'n' : PseudoNoeuds = 0     		;
			    break 				;
                 case 'r' : CompatibiliteFaible = 1             ;
                            break				;
		 case 'k' : Distingue = 1 		        ;
                            break				;
	  }
    }


   fprintf(stderr,"Recherche des %d meilleures solutions 1\n",K) ;

   Nseq = 0 ;
   NbPals = 0 ;

					/* Solutions			*/
   Best =  NEWN(Int32*,K) ;
   PointeurNul(Best,"Best") ;

					/* Poids des solutions		*/

   Wbest = NEWN(Int32, K) ;
   PointeurNul(Wbest,"Wbest") ;

					/* Nb elts par solution		*/

   NbElt = NEWN(Int32,K) ;
   PointeurNul(NbElt,"NbElt") ;

   Pals = NEWN(Palindrome,MaxPal) ;
   PointeurNul(Pals,"Pals") ;


   fprintf(stderr,"Pseudo-noeuds = %d",PseudoNoeuds) ;

   fprintf(stdout,"%d\n",NbSeq) ;	/* inscrit Nombre sequences	*/

	while
	        (
		 Nseq < NbSeq &&
		 LireSeqScore(stdin,Distingue, &DebPart, &FinPart,
                              Pals,Seq, &NbPals )
	       ) 
        {
					/* pour chaque sequence		*/

	      fprintf(stderr,"Sequence  %d : \n",Nseq) ;

					/* matrice compatibilite	*/

	      bool = NEWN(Bool,NbPals * NbPals) ;
	      PointeurNul(bool,"Bool")          ;

	      MatriceCompatibilite(CompatibiliteFaible,bool,NbPals,Pals) ;

					/* Vecteur des poids		*/

	      Poids = NEWN(Int32,NbPals) ;

	      for (i = 0 ; i < NbPals ; i++)
		  Poids[i] = Pals[i].Score ;

					/* solutions			*/

	      for (k = 0 ; k < K ; k++) 
	      {
		   Best[k] = NEWN(Int32,NbPals) ;
		   PointeurNul(Best[k],"Best") ;
	       }

	      nK = 0 ;

	      KBaB(K,&nK,Poids,NbPals,bool,MaxElt,Wbest,Best,NbElt) ;

	      SauvegardeSolutions(stdout,Distingue, DebPart, FinPart,
 				  Seq,NbPals,Pals,nK,Best,NbElt) ;

	      free(bool)  ;
	      free(Poids) ;

	      for (k = 0 ; k < K ; k++)
		   free(Best[k]) ;

	      Nseq++ ;
	}

					/* desalocation			*/
   free(Best) ;
   free(Wbest) ;
   free(NbElt) ;
					/* fermeture fichiers		*/

   fprintf(stderr,"\nfin du programme\n") ;

   return 1 ;
}      
