/* ====================================================================	*/
/*    		 							*/
/*	Copyright (c) Atelier de BioInformatique			*/
/*	Juillet 95.		 					*/
/* 	Fichier : Kmrstr8.c  	    					*/
/*	Objet	:							*/
/* 	   	  Recherche des structures secondaires repetees en 	*/
/* 	          considerant N sequences, t[i] solutions (ensembles	*/
/*	   	  de palindromes deux a deux compatibles) pour chaque  	*/
/* 	          sequence i, et p[i][j] palindromes pour chaque	*/
/*	          solution j d'une sequence i.				*/
/*	          Une structure est repetee q fois si elle apparait	*/
/* 	          dans au moins 1 solution de q sequences		*/
/*	Version : Avril 1995						*/
/*                version avec sauvegarde des taux de repetitions       */
/*                de chaque structure de longueur 1 a 4                 */
/*	          version avec description longueur, purete des pals	*/
/*      Version : Mai 95						*/
/*	          notion de distance entre palindromes meme sol		*/
/*      Version : Juillet 95						*/
/*                apprentissage valeurs des attributs			*/
/*									*/
/*      Parametres de main : valeur par defaut entre parentheses	*/
/*									*/
/*	-f<Nomfichier>	: affichage structures	(fichier modele alea)	*/
/*	 		  						*/
/*      -q<entier>      : pourcentage minimum de repetition (80)	*/
/* 									*/
/*   	-e<nomfichier>  : sauvegarde taux des structures                */
/*									*/
/*	-a<nomfichier>	: calcul des accroissements relatifs		*/
/*									*/
/*      -v<entier>      : taux d'affichage [0, 1 ou 2])	     (2)	*/
/*									*/
/*      -m<entier>      : taille maximale des structures    (15)	*/
/*									*/
/*      -n<entier>      : taille minimale des structures     (1)	*/
/*									*/
/*      -l<entier>      : taille maxi d'une sequence       (250)	*/
/*									*/
/*	-d<Model alea>	: "dot plot"					*/
/*									*/
/*      -x		: "dot plot" / sequence sinon /structure/seq	*/
/*									*/								
/*	-y<MaxPoids>	: Nb occurrences appariements pour pixel max	*/
/*									*/
/*									*/
/*	-z		: les sequences sont alignees			*/
/*									*/	
/*	-w<FAppariement>: nom fichier "dot plot"			*/
/*									*/
/*  	-u<NbStructures>: considere NbStructures structures candidates	*/
/*			  0 <=> toutes les structures candidates	*/
/*									*/						
/* ====================================================================	*/

   
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "../../include/gtypes.h"
#include "../../include/pals6.h"
#include "../../include/kmrstr8.h"
#include "../../include/param.h"
#include "../../include//libstki.h"
#include "../../include/libsysks.h"
#include "../../include/libwbs.h"
#include "../../include/libkmrs8.h" 


/* ==================================================================== */
/* constantes								*/
/* ==================================================================== */

#define MAX_LONGUEUR_STRUCTURES 50
#define ENTETE "                " 
#define SHIFT 		    	16   /* 16 blancs avant d'ecrire la seq	*/
#define MIN_PLAUSIBILITE     -1000
#define MIN_PIXEL 	     	 0.0
#define MAX_PIXEL 	         1.0
#define MIN_V			 0.0  
#define INFINI 		      1000
#define SEUIL   	         0.5 /* seuil pour considerer apparie	*/
#define NB_STRUCTURES            5   /* # meilleures structures		*/ 
#define COUVERTURE_MIN	       100   /* 1% des sequences		*/



/* ==================================================================== */
/* Macros standards							*/
/* ==================================================================== */

#define CheckNullPtr(Mes,ptr) if (!(ptr)) {fprintf(stderr,"%s",Mes) ; return 0 ;}



/* ==================================================================== */
/* definition des fonctions locales					*/
/* ==================================================================== */



void SwapInt(int* a, int *b)
{
 int Aux = *b ;
     *b = *a  ;
     *a = Aux ;
}



void EnleveBlancs(char chaine[]) 
{
  int i = 0 ;
	
  while (chaine[i] != 0)
  {
  	if (chaine[i] == ' ')
	{
	    if (chaine[i+1] == 0)
	        chaine[i] = 0 ;
	    else
	        chaine[i] = chaine[i+1] ; 
	}
	else 
		i++ ;
  }	
}


/* --------------------------------------------------------------------	*/
/* Initialise un tableau de pointeurs					*/
/* --------------------------------------------------------------------	*/
static void InitialiseTabPtr(Int32 N, void * Ptr[])
{

    Int32 i ;

  for (i = 0 ; i < N ; i++)
      Ptr[i] = NULL ;
}



/* --------------------------------------------------------------------	*/
/* convertit argument chaine du programme en entier			*/
/* --------------------------------------------------------------------	*/
static void ARGUMENT(int* a, char b, char c[])
{
     if (c[0] != '0') 
     {
          *a = atoi(c) ;
          if (!a) 
          {
		fprintf(stderr,"Mauvais argument -%c %s\n",b,c); 
                exit(0) ;
	  }
     }
}



/* --------------------------------------------------------------------	*/
/* Lecture des solutions de chaque sequence				*/
/* --------------------------------------------------------------------	*/
static int LectureDonnees(
        
	   int Distingue ,
           Int32** DebPart,
           Int32** FinPart,
           FILE *fichier ,      /* fichier d'entree			*/
	   Int32 *N	 ,	/* nombre de sequences			*/
           Int32 Lmax    ,      /* Longueur max d'une sequence		*/
	   char**   s    ,      /* tableau des sequences		*/
	   Int32 *ns 	 ,      /* nombre total de solutions		*/
	   Int32 *n	 ,	/* nombre total de palindromes		*/
	   Int32 **t 	 ,	/* nombre solutions seq i		*/
	   Int32 ***p    ,  	/* p[i][j] nombre pal seq i solu j  	*/
	   Int32 **NbPal ,      /* nombre de pals sequence i		*/
	   Palindrome  ***Pal  ,/* Pal[i][m] = pal m seq i 	        */
	   Int32 ****Sol        /* Sol[i][j][m] indice du pal m         */
				/* sol j seq i dans Pal[i]		*/
			)

{
   Int32  i		;
   Int32  l		;
   Int32  j		;
   Int32  m		;
 
	    		/* lit nombre de sequence			*/

   fscanf(fichier,"%d",N) 		   ;


   fprintf(stderr,"Lecture des donnees en cours...\n") ;
  

   *s   = (char*) malloc(*N *Lmax) ;
   CheckNullPtr("Pas assez de memoire pour allouer s\n",*s)
   

   *t   = (Int32*)    malloc(*N * sizeof(Int32)) 	;
   CheckNullPtr("Pas assez de memoire pour allouer t !\n",*t)

   *p   = (Int32**)   malloc(*N * sizeof(Int32*)) 	;
   CheckNullPtr("Pas assez de memoire pour allouer p\n",*p)
   InitialiseTabPtr(*N,(void**) *p)   ;

   *Pal = (Palindrome **) malloc(*N * sizeof(Palindrome*))	;
   CheckNullPtr("Pas assez de memoire pour allouer Pal\n",*Pal)
   InitialiseTabPtr(*N,(void**) *Pal) ;

   *NbPal = (Int32*) malloc(*N * sizeof(Int32)) ;
   CheckNullPtr("Pas assez de memoire pour allouer NbPal\n",*NbPal)

   *Sol = (Int32 ***) malloc(*N * sizeof(Int32**)) 	;
   CheckNullPtr("Pas assez de memoire pour allouer Sol\n",*Sol)
   InitialiseTabPtr(*N,(void**) *Sol) ;

   *DebPart = (Int32*) malloc(*N * sizeof(Int32)) ;
   CheckNullPtr("Pas assez de memoire pour allouer DebPart\n",*DebPart)

   *FinPart = (Int32*) malloc(*N * sizeof(Int32)) ;
   CheckNullPtr("Pas assez de memoire pour allouer FinPart\n",*FinPart)
   

   for (i = 0 ; i < *N ; i++) 
   {
       /* pour chaque sequence				*/
       
       
       fprintf(stderr,"Sequence %d\n",i) ;

	if (Distingue) 
        {
            fscanf(fichier, "%d", *DebPart+i) ;
            fscanf(fichier, "%d", *FinPart+i) ;
        }
        

        		/* lit contenu de la sequence			*/

       fscanf(fichier,"%s",(*s) + i * Lmax) ; 
      
      
        if (!Distingue)
	{
	    (*DebPart)[i] = 1 			        ;
	    (*FinPart)[i] = strlen((*s) + i * Lmax)     ;
	}
                     /* lit nombre de palindromes dans la sequence	*/
       fscanf(fichier,"%d",&((*NbPal)[i])) ;

       (*Pal)[i] = (Palindrome*)  malloc((*NbPal)[i] * sizeof(Palindrome)) ;
       CheckNullPtr("Pas assez de memoire pour allouer Pal[i]\n",(*Pal)[i])


       for (l = 0 ; l < (*NbPal)[i] ; l++) 
       {
	   /* lit les palindromes			*/
	   fscanf(fichier,"%d %d %d %d %d %d",
		  &((*Pal)[i][l].DTete),&((*Pal)[i][l].FTete),
		  &((*Pal)[i][l].DQueue),&((*Pal)[i][l].FQueue),
		  &((*Pal)[i][l].Code),&((*Pal)[i][l].Score)) ;
       }

       /* lit le nombre de solutions pour la sequence	*/
      fscanf(fichier,"%d", &((*t)[i])) ;

      *ns       += (*t)[i]
			     ;
      (*p)[i]      = (Int32*) malloc((*t)[i] * sizeof(Int32))  ;
      CheckNullPtr("pas assez de memoire pour allouer p[i]\n",(*p)[i])



       (*Sol)[i] = (Int32 **) malloc((*t)[i] * sizeof(Int32*)) ;
       CheckNullPtr("Pas assez de memoire pour allouer Sol[i]\n",(*Sol)[i])
       InitialiseTabPtr((*t)[i],(void**)(*Sol)[i]) ;

      for (j = 0 ; j < (*t)[i] ; j++) 
      {
	  /* pour chaque solution 			*/


	  /* lit le nombre de pal dans la solution	*/
	  fscanf(fichier,"%d",&((*p)[i][j]))	;



	  (*Sol)[i][j] = (Int32*) malloc((*p)[i][j] * sizeof(Int32)) ;
	  CheckNullPtr("Pas assez de memoire pour allouer Sol[i][j]\n",
                       (*Sol)[i][j])

	  for (m = 0 ; m < (*p)[i][j] ; m++) 
          {
	      /* lit indice dans Pal[i][j] du pal de 	*/
	      /* chaque solution			*/
	      fscanf(fichier,"%d",&((*Sol)[i][j][m])) ;
	      (*n)++ ;
	  }				/* fin m	*/
      }					/* fin j	*/
   }					/* fin i	*/
   close(fichier) ;
   return 1       ;
}



/* --------------------------------------------------------------------	*/
/* Constitue la matrice des structures de chaque solution		*/
/* --------------------------------------------------------------------	*/
static Int32 
ConstituerMatriceInitiale(
       Int32 ***** M 	, 	/* M[i][j] represente matrice diago	*/
				/* Seq i Sol j				*/
       Int32 N  	,	/* nombre de sequences			*/
       Int32 *t      	,	/* nombre de solutions / sequence	*/
       Int32 **p	 	/* nombre de pals / solution		*/
                         )
{
  Int32 i ;
  Int32 j ;
  Int32 k ;


  fprintf(stderr,"Constitution des matrices initiales...\n") ;



  *M = (Int32 ****) malloc(N * sizeof(Int32 ***)) ;
  CheckNullPtr("Pas assez de memoire pour allouer M\n",*M) ;
  InitialiseTabPtr(N,(void**) *M) ;

  for (i = 0 ; i < N ; i++) 
  {
      /* pour chaque sequence						*/

      fprintf(stderr,"%d\n",i) ;

      (*M)[i] = (Int32 ***) malloc(t[i] * sizeof(Int32 **)) ;
      CheckNullPtr("Pas assez de memoire pour allouer M[i]\n",(*M)[i])
      InitialiseTabPtr(t[i],(void **) (*M)[i]) ;

      for (j = 0 ; j < t[i] ; j++) 
      {
	  /* pour chaque solution					*/

	  (*M)[i][j] = (Int32 **) malloc(p[i][j] * sizeof(Int32 *)) ;
	  CheckNullPtr("Pas assez de memoire pour allouer M[i][j]\n",(*M)[i][j])
	  InitialiseTabPtr(p[i][j],(void **) (*M)[i][j]) ;

	  for (k = 0 ; k < p[i][j] ; k++) 
	  {
	      /* pour chaque palindrome de la solution			*/
	      /* on cree une ligne dans la matrice			*/
   
	      (*M)[i][j][k] = (Int32 *) malloc ((k+1) * sizeof(Int32)) ;
	      CheckNullPtr("Pas assez de memoire pour allouer  M[i][j][k]\n",
                           (*M)[i][j][k])
	      InitialiseTabPtr(k+1,(void **) (*M)[i][j][k]) ;

	      /* initialement structure de taille 1 codage 0		*/
	      (*M)[i][j][k][k] = 0 ;
	  }
      }
  }

  fprintf(stderr,"Fin constitution des matrices\n") ;
  return 1 ;

}



/* --------------------------------------------------------------------	*/
/* Constitue le vecteur de correspondance entre une position		*/
/* dans l'index Va et les sequence, solution et palindrome 		*/
/* correspondant a cet element.						*/
/* --------------------------------------------------------------------	*/
static Int32 
ConstituerPosition(
       TPosition **Pos	,	/* index des positions			*/
       Int32        n   ,       /* nb pals dans seq et sol		*/
       Int32	    N	,       /* nombre de sequence			*/
       Int32	   *t   ,	/* nb solutions par seq i		*/
       Int32	  **p		/* nb pal par solution j		*/
	          )
{
 Int32 i     ;
 Int32 j     ;
 Int32 k     ;
 Int32 r = 0 ;



 fprintf(stderr,"Consitution du tableau de correspondance...\n") ;

 *Pos = (TPosition *) malloc(n * sizeof(TPosition)) ;
 CheckNullPtr("Pas assez de memoire pour allouer Pos",*Pos)

				 	/* Pour chaque sequence 	*/

 for (i = 0 ; i < N ; i++) 
 {   					
     fprintf(stderr,"Sequence %d\n",i) ;
			
					/* Pour chaque solution         */
     for (j = 0 ; j < t[i] ; j++) 
     	 
					/* Pour chaque palindrome	*/

	 for (k = 0 ; k < p[i][j] ; k ++) 
	 {
	     (*Pos)[r].Seq = i ;
	     (*Pos)[r].Sol = j ;
	     (*Pos)[r].Pal = k ;

	     r++ ;
	 }
 }
 return 1 ;
}



/* --------------------------------------------------------------------	*/
/* Constitue l'index Va des classes des structures repetees pour k = 1	*/
/* une unique classe de numero 0.					*/
/* --------------------------------------------------------------------	*/
static Int32 
ConstituerVaInitial(
       		Int32	n,  	/* nb elements dans l'index Va		*/
       		Int32** Va      /* index 				*/
                   )
{
  Int32 r ;



  fprintf(stderr,"Constitution va \n") ;

  *Va = (Int32 *) malloc(n * sizeof(Int32)) ;
  CheckNullPtr("Pas assez de memoire pour allouer Va\n",*Va) ;

  for (r = 0 ; r < n ; r++)
       (*Va)[r] = 0 ;
      
  return 1 ;

}


/* --------------------------------------------------------------------	*/
/* insere la valeur v en position (p + shift) du tableau T		*/
/* --------------------------------------------------------------------	*/

static void 
Insere(
	int 	T[]	, 
	int 	p	, 
	int 	shift	,  
	int 	v
      )
{
   int i ;
   
   /* decalage vers la gauche des elts compris entre p+1 et p+shift	*/
   for (i = p ; i > (p-shift) ; i--)
        T[i] = T[i-1] ;
	
   T[p-shift] = v ;
       
} 



/* --------------------------------------------------------------------	*/
/* classe les tetes et queues des palindromes d'une structure		*/
/* --------------------------------------------------------------------	*/
 
static void 
ClasserElts(
		int 		k		, 
		int 		OrdreElts[]	, 
		Int32*** 	M[]		,	    
	        Int32 		Seq		, 
		Int32 		Sol		, 
		Int32 		Pal
	   ) 
{
  int u     ; /* indice du palindrome dans la structure 		*/
  	
	/* positionne tete et queue du dernier palindrome		*/
	/* queue du uieme pal reference comme elts 2u, tete comme 2u+1	*/
	/* 2k elts dans le tableau OrdreElts				*/
	
	  
				      /* 1er (pivot) pal de structure   */
        OrdreElts[0] = 0        ;     /* emplacement queue pal		*/
 	OrdreElts[1] = 1        ;     /* emplacement tete pal		*/

        for (u = 1 ; u < k ; u++)
	{    
	    OrdreElts[2*u+1] = 2*u + 1;  /* tete pal u			*/
	    Insere(OrdreElts, 2*u, M[Seq][Sol][Pal][Pal-u], 2*u) ;
  	}
}



/* --------------------------------------------------------------------	*/
/* Affiche la structure	de k pals d'indice Pal dans M		   	*/
/* --------------------------------------------------------------------	*/

static void 
AfficherStructure(
			int 		OrdreElts[]	, 
			Int32 		Seq		, 
			Int32 		Sol		, 
			Int32 		Pal		, 
			Palindrome* 	Pals[]		,  
			Int32** 	Sols[]		,  
			Int32 		k
		  )
{
  int i                          ;
  int j		                 ;
  int u                          ;
  int tete                       ;
  Int32 Np	                 ;
  int Longueur		         ;
  int position =  1		 ;
  int debut			 ;
  float Energie = 0		 ;
   
					/* calcul de l'energie 		*/	
		
    for (i = 2*k-1 ; i >= 0 ; i--)
    {
		/* pour l'element (tete ou queue)courant		*/

	  u        = OrdreElts[i] / 2 ;
	  tete     = OrdreElts[i] % 2 ;
	  Np       = Sols[Seq][Sol][Pal-u] ;
	  if (tete)
	          Energie += Pals[Seq][Np].Score ;
     }
     
     fprintf(stdout,"%9.2f  kc/m ",Energie / (-100.0)) ;
    
     for (i = 2*k-1 ; i >= 0 ; i--)
     
     {
		/* pour l'element (tete ou queue)courant		*/
	  u    = OrdreElts[i] / 2 ;
	  tete = OrdreElts[i] % 2 ;
	  
	  Np = Sols[Seq][Sol][Pal-u] ;
	  
	  if (tete)
	  {
		debut    = Pals[Seq][Np].DTete             ;
		Longueur = Pals[Seq][Np].FTete - debut + 1 ;
		Energie += Pals[Seq][Np].Score ;
	  }
	  else
	  {
	        debut    = Pals[Seq][Np].DQueue             ;
		Longueur = Pals[Seq][Np].FQueue - debut + 1 ;
	  }
	  
	  /* complete par des blancs de position a debut-1		*/
	  
	  for (j = position ; j < debut ; j++)
	      fprintf(stdout, " ") ;
	      
	  /* ecrit l'element (tete ou queue) du palindrome		*/
	  
	  
	  for (j = 0 ; j < Longueur ; j++)
	      if (tete)
		 fprintf(stdout, "%c", k-u + 64) ;	/* 65 = asc(A)	*/
 	      else
		 fprintf(stdout,"%c", k-u + 96)  ;	/* 97 = asc(a)	*/  
          

	  position = debut + Longueur ;
     }
     fprintf(stdout,"\n") ;
}
		       


/* --------------------------------------------------------------------	*/
/* verifie si la structure represente par Sol,Pal est identique a celle */
/* representee par Sol1, Pal1						*/
/* --------------------------------------------------------------------	*/
static int 
MemeInstance(
		Int32 	k	, 
		int 	Sol1	, 
		int 	Pal1	, 
		int 	Sol2	, 
		int 	Pal2	, 
                Int32**	Sols
	    )
{
    int u1           ;
    int u2           ;
    int Longueur = 0 ;
    
        
    u1 = Pal1 - k + 1 ;				
    u2 = Pal2 - k + 1 ;
    
    while ((Longueur < k) && (Sols[Sol1][u1]== Sols[Sol2][u2]))
    {
	u1++       ;
	u2++       ;
	Longueur++ ;
    }
    
    return (Longueur == k) ;        
}



/* --------------------------------------------------------------------	*/
/* met a jour le tableau des instances de structures dans une sequence  */
/* --------------------------------------------------------------------	*/
static int 
InsereInstance(
		  Int32 	ClasseInstance	, 
  		  Int32 	k		, 
		  TInstance 	Instances[]	, 
               	  int 		Sol		, 
	          int 		Pal		, 
		  int* 		NbInstances	, 
		  Int32**	Sols
               )
{
   int j = ClasseInstance ;
					/* verifie si instance deja la	*/

   while  (
	    (j < *NbInstances) &&
	    !MemeInstance(k, Sol, Pal, Instances[j].Sol, 
		          Instances[j].Pal, Sols)
	  )  
	  j++ ;
   
   if (j == *NbInstances)
   {
      Instances[j].Sol = Sol ;
      Instances[j].Pal = Pal ;
      (*NbInstances)++       ;

      return 1 ;
   }             	
   else
      return 0 ;
}


/* --------------------------------------------------------------------	*/
/* Recherche Structure dans le tableau SModel ordonne selon Best   	*/
/* --------------------------------------------------------------------	*/
static int
DansSelection(
	      StructModel   Structure	, 
	      StructModel*  SModel	, 
	      int*	    Best        ,
	      int 	    NbStructures
	     )
{ 
    int t = 0	;
    
    while (t < NbStructures && strcmp(Structure, SModel[Best[t]]))
	   t++ ;
	
    return (t < NbStructures) ;        
}
	     	     

/* --------------------------------------------------------------------	*/
/* Affiche la sequence et distingue la zone correspondant a l'ARN	*/
/* --------------------------------------------------------------------	*/
static void
AfficheSequence(
	        int     i	    , 
		int*	NbPal	    , 
		char*	s	    , 
		int	Lmax	    , 
		int	Distingue   , 
		int	DebPart	    , 		                    
		int	FinPart
	       )
{	       
    int	u	 ;
    int	j	 ;
    int	l	 ;
    int Longueur ;	
    
    
    Longueur = strlen(s+i*Lmax)   ;

    fprintf(stdout," :\n") ;
    fprintf(stderr," :\n") ;                             
    fprintf(stdout,"\n\nsequence %d\n", i)  ;
    fprintf(stdout,"contenant %d palindromes :\n",NbPal[i]) ;
    fprintf(stdout,"%s",ENTETE) ;
 	  
	  					/* affichage   dizaines */
    fprintf(stdout,"         ") ;
    for (j = 10 ; j <= Longueur ; j+=10)
    {
	 fprintf(stdout, "%1d", j/10) ;
	 for (l = 1 ; l <= 10 - (int)floor(log10(j)) ; l++)
  	      fprintf(stdout, " ") ; 
     } 
	   
						/* affichage unites	*/
    fprintf(stdout,"\n") ;
    fprintf(stdout,"%s", ENTETE) ;
    for (j = 1 ; j <= Longueur ; j++)
	 fprintf(stdout,"%1d",j % 10) ;
    fprintf(stdout,"\n") ;
			 
						/* affichage sequence	*/
    fprintf(stdout,"%s",ENTETE) ;
    fprintf(stdout,"%s\n",s+i*Lmax)     ;
    
      						/* affichage zone RNA  */			
    if (Distingue)
    {		      
       fprintf(stdout,"%s",ENTETE) ;
       for(u = 1 ; u < DebPart ; u++) 
           fprintf(stdout," ") ;
                     
       for(u = DebPart ; u <= FinPart ; u++) 
          fprintf(stdout,"x") ;
       fprintf(stdout,"\n") ;
    }
}    	



/* --------------------------------------------------------------------	*/
/* affiche des marques correspondant aux palindromes d'un tRNA		*/
/* --------------------------------------------------------------------	*/
static void Affiche_trna(
	Int32 Debut,
	Int32 Fin       )
{
  Int32 i               ;
  char ctrna[150]       ;
  int FinArnt = 72      ;
  int  j		;

 			/* place le curseur au debut du tRNA		*/

  for (i = 1 ; i < Debut ; i++)
       fprintf(stdout, " ") ;

  
   /* fprintf(stdout,"(((((((  {{{{        }}}} <<<<<       >>>>>     [[[[[       ]]]]])))))))") ;  */

 
  for (i = 0 ; i < Fin ; i++)
       ctrna[i] = ' ' ;
 

						/* 1er palindrome		*/
	for (j = 0 ; j < 7 ; j++)
        {            	
 	    ctrna[Debut+FinArnt-1-j-1] = ')' ;
 	    ctrna[Debut+1-1+j-1]       = '(' ;
        }

						/* 2eme palindrome		*/
	for (j = 0 ; j < 4 ; j++)
        {	
 	    ctrna[Debut+25-1-j-1] = '}' ;
	    ctrna[Debut+10-1+j-1] = '{' ; 	
        }

						/* 3eme palindrome		*/
	for (j = 0 ; j < 5 ; j++)	
        {
 	    ctrna[Debut+43-1-j-1] = '>' ;
	    ctrna[Debut+27-1+j-1] = '<' ; 	
        }

					        /* 4eme palindrome		*/
	for (j = 0 ; j < 5 ; j++)
	{	
 	    ctrna[Debut+FinArnt-7-1-j-1] = ']' ;
	    ctrna[Debut+49-1+j-1]        = '[' ; 	
        }

   ctrna[Fin] = 0 ; 

   fprintf(stdout,"%s", ctrna) ;
 	
}



/* --------------------------------------------------------------------	*/
/* affichage des sequences, palindromes et positions des structures	*/
/* --------------------------------------------------------------------	*/
static int 
SauvegardeStructures(
       int	    Distingue	, 
       Int32*	    DebPart	, 
       Int32*	    FinPart	, 
       Int32	    Lmax	,  /* longueur max d'une sequence	*/
       Int32	    N		,  /* nombre de sequences		*/
       char*	    s		,  /* tableau des sequences		*/
       Int32	    NbPal[]	,  /* nombre pal de chaque sequence	*/
       Palindrome   *Pals[]     ,  /* pal de chaque sequence		*/
       WordBunchPtr AllStr	,  /* positions des structures		*/
       TPosition    Pos[]	,  /* vecteur des positions des struc	*/
				   /* dans les sequences et solutions	*/
       Int32***	    Sols        , 
       Int32***	    M[]		, 
       StructModel* SModel	, 
       Int32*       Best        ,
       int          NbStructures
   	           )       

{
  
  TInstance* Instances = NULL      ; /* Occurrences des structures	*/
  Int32 NbInstances    = 0	   ; /* Nb instances de structures	*/
  Int32 ClasseInstance = 0         ; 
  Int32 MaxInstances   = 0         ; 
  Int32 NbInstancesTot = 0         ;
  int   OrdreElts[MAX_LONGUEUR_STRUCTURES] ; 
				      /* ordre tetes et queue dans str */
  Int32  i			   ;
  Int32  k			   ;
  Int32  ea         = -1	   ;
  Int32  c			   ;
  StackiHdle Structures		   ;
  Int32  val			   ;
  Int32 Seq			   ;
  Int32 Sol			   ;
  Int32 Pal		           ;
  int   premier			   ; /* temoin 1er elet d'une classe	*/
  Int32 u			   ;
  int   affiche   = 0              ;
  int   ok_classe = 1              ;
  StructModel  Structure	   ;
  StructModel  Tampon		   ;
  int   iterer = 1	           ; /* temoin arret lecture classe	*/
  
 
  fprintf(stderr,"Sauvegarde des resultats en cours...\n") ;

  fprintf(stdout,"Nombre de sequences : %d\n\n",N)  ;

						    /* sauvegarde positions structures */			
  CursiToTop(AllStr->nbwords) ;
  CurspToTop(AllStr->words)   ;

						    /* sauve nombre de longueurs	*/
    
  for (k = AllStr->maxlen ; k >= AllStr->minlen ; k--) 
  {
						    /* sauve longueur k			*/

      fprintf(stdout,"\n\nlongueur           : %d \n",k) ;

						    /* sauve nombre classes longueur k	*/
     
      if (!ReadiDown(AllStr->nbwords,&ea))
      {
	  fprintf(stderr,"probleme lecture ea (%d)\n", ea) ;
          return 0 ;
      }

      fprintf(stdout,"nombre de classes  : %d\n\n",ea)     ;

						    /* ecrit les positions debut structure */
						    /* (pal + a droite) de chaque classe   */

      ReadpDown(AllStr->words,(Ptr *) &Structures) ;

						    /* curseur haut chaque pile		    */
     for (c = 0 ; c < ea ; c++)
          CursiToTop(Structures[c]) ;

						    /* pour chaque sequence		    */
     for (i = N-1 ; i >= 0 ; i--)
     {
	 fprintf(stderr,"sequence %d\n",i) ;

         NbInstancesTot = 0        ;
	 Instances      = NULL     ; 
         MaxInstances   = 0        ;

         for ( c = 0 ; c < ea ; c++)
	       MaxInstances += Structures[c]->top ;
	  
	 Instances =  malloc(MaxInstances * sizeof(TInstance))			 ;
         CheckNullPtr("Pas assez de memoire pour allouer Instances\n",Instances) ;

				    		/* pour chaque classe longueur k	    */
        for (c = 0 ; c < ea ; c++)  
        {
          affiche = 0 ;	
	  premier = 1 ;

          NbInstances    = 0 ;
          ClasseInstance = 0 ;
	  iterer         = 1 ;
	 
	  while (iterer && ReadiDown(Structures[c],&val)) 
	  {
		 
		  Seq  = Pos[val].Seq      ; 
		  Sol  = Pos[val].Sol      ;
		  Pal  = Pos[val].Pal      ;	
	  
	   				 			     		    		   		         	              	      					
		  if (premier)
		  {			   			          					        			    			                               			   
			  premier = 0         ;			 			  
			  Structure[0] = '\0' ;	
		  	  				/* recupere classe structure	*/
			  for (u = k - 1 ; u >= 0 ; u--)	
 			  {		 
                              Tampon[0] = '\0' ;

                              if (u == 0)
				  sprintf(Tampon, "%d", M[Seq][Sol][Pal][Pal-u])  ;
 			      else
				  sprintf(Tampon, "%d ",M[Seq][Sol][Pal][Pal-u])  ;	

			      strcat(Structure,Tampon) ;
			  }

			  ok_classe = DansSelection(Structure, SModel, Best, NbStructures) ;
			  
			  if (ok_classe && Seq == i)
                          {
			     affiche = 1							  ;
			     AfficheSequence (i,  NbPal, s, Lmax,  Distingue, DebPart[i], FinPart[i]) ;
			  	    
                             fprintf(stdout,"%s", Structure) ;
	
					/* affichage du tRNA				*/
                             /* fprintf(stdout,"         ") ;
		             Affiche_trna(DebPart[i], FinPart[i]) ; */
		             fprintf(stdout,"\n") ;
			     fprintf(stderr,"%s\n", Structure) ;	
		          } /* fin de ok_classe */
			             
		    }  /* Fin de if premier		*/
		       
		    if (ok_classe && Seq == i)
		    {
		           ClasserElts(k, OrdreElts, M, Seq, Sol, Pal) ;
                           if (InsereInstance(ClasseInstance, k, Instances, Sol, Pal, 
		                              &NbInstances, Sols[i]))    
			        AfficherStructure(OrdreElts, Seq, Sol, Pal, Pals,  Sols, k) ;
		    }   /* fin de ok_classe && Seq == i */
		    else
		    {
			  iterer = 0 ;		       
			  	   	/* element lu appartient a sequence < i */		  			
		 	  	        /* repositionne pointeur sur element 	*/
			  (Structures[c]->cursor)++ ;
	            } 			   		                       					         
		    		    
           }  /* fin de While continue ...*/
           
           if (affiche)
           {
               fprintf(stdout,"Instances = %d\n",NbInstances) ;
               NbInstancesTot += NbInstances ; 
           }
        }   /* fin de for c */
    
    free(Instances) ;
    fprintf(stdout,"Instances cumulees = %d\n",NbInstancesTot) ;

    }  /* fin de for i */

  } /* fin de for k */

 
  fprintf(stderr,"Fin sauvegarde des resultats.\n") ;
  return 1	  ;
} 
/* fin de SauvegardesStructures */



/* --------------------------------------------------------------------	*/
/* Creation des tableaux SModel et Best					*/
/* --------------------------------------------------------------------	*/
static int 
CreationModele(
		char*         	NfichierAc , 
		StructModel**	SModel	   , 
		int**		Best       ,
		Int32*		SFic	   ,
		Int32**         Model                    
               ) 
{
   int 		i		;
   int		long1	   = -1	;
   int 		x	   = -1	;
   int		u		;
   StructModel 	Structure	;
   FILE*	fichier		;

   *Model  = (Int32 *) malloc(SFic[LMS] * sizeof(Int32)) ;  
   if (!(*Model)) 
   {
      fprintf(stderr,"Probleme de memoire impossible allouer Model\n") ;
      return 0 ;
   }


   *SModel = (StructModel*) malloc (SFic[LMS] * sizeof(StructModel)) ;

   if (!*SModel) 
   {
      fprintf(stderr,"Probleme de memoire impossible allouer SModel\n") ;
      return 0 ;
   }

  *Best = (Int32 *) malloc(SFic[LMS] * sizeof(Int32)) ;

   if (!*Best) 
   {
      fprintf(stderr,"Probleme de memoire impossible allouer Best\n") ;
      return 0 ;
   }
			/* *Best contient les indices des structures	*/
			/* apres Best contiendra les indices classes	*/
			/* selon l'ordre decroissant des taux associes	*/

   for (i = 0 ; i < SFic[LMS] ; i++)
	(*Best)[i] = i ;


      		/* chargement des taux de repetitions des structures	*/
		/* issues du modele aleatoire				*/

 
    fichier = fopen(NfichierAc,"r") ;
    if (!fichier)
    {
	fprintf(stderr, "Impossible d'ouvrir le fichier %s\n", NfichierAc) ;
	return 0 ;
    }

    fprintf(stdout, "Fichier modele aleatoire : %s\n", NfichierAc) ;

    for (i = 0 ; i < SFic[LMS] ; i++) 
    {
	fscanf(fichier,"%d %d", &x, &((*Model)[i])) ;
	fgets(Structure , ML, fichier) ;

					/* enleve blancs a droite	*/
        u = 0 ;
	long1 = strlen(Structure) ;
	while (u < long1 && Structure[u] == ' ')
               u++ ;
	
	(*SModel)[i][0] = '\0' ;

	if (u < long1)                   
            strcat((*SModel)[i], Structure + u) ;        
        
	   

	x = strlen((*SModel)[i]) ;
	if ((*SModel)[i][x-1] == '\n')
	    (*SModel)[i][x-1] = '\0' ;
   }

   return 1 ;
}



/* --------------------------------------------------------------------	*/
/* Sauvegarde des taux de chaque structure trouvee			*/
/* --------------------------------------------------------------------	*/

static void 
SauvegardeTaux(
        	char  *Nfichierst     ,
	        Int32 Fic[]           ,
		Int32 SFic[]  	      ,
		Int32 Tstruct[]       
	      )
{
   FILE *Fichier   ;
   Int32 Longueur  ;
   Int32 j         ;
   Int32 k         ;
   Int32 s         ; 
   Int32 m         ;

   fprintf(stderr, "Sauvegarde des taux de repetition des structures...\n") ;

   Longueur = 1 ;

   Fichier = fopen(Nfichierst,"w") ;

   for (j = 0 ; j < SFic[LMS] ; j++) 
   {

        if (j >= SFic[Longueur])
            Longueur ++ ;

        fprintf(Fichier,"%5d  %5d  ",j, Tstruct[j]) ;

        m = j - SFic[Longueur-1] ;

        for (k = Longueur-1 ; k >= 1 ; k--) 
	{
            s = m / Fic[k] ;
            m = m % Fic[k] ;

            fprintf(Fichier," %d",s) ;
        }

        fprintf(Fichier," 0\n") ;
   }

   fclose(Fichier) ;
}



/* --------------------------------------------------------------------	*/
/* calcul des NbStructures best accroissements selon modele aleatoire	*/
/* --------------------------------------------------------------------	*/

static void 
Accroissements(
	       Int32         SFic[]	  , 
	       Int32         Tstruct[]    , 
	       StructModel*  SModel	  , 
	       Int32*        Best 	  ,
	       Int32*	     Model        ,
	       int           NbStructures   /* # structures candidates	*/
					    /* a considerer		*/	
	      )
{

 Int32 	i		;
  Int32 	j		;
  Int32 	k		;
  float 	*Ac		;
  StructModel 	Structure 	;
  
   Ac = (float *) malloc(SFic[LMS] * sizeof(float)) ;

   if (!Ac) 
   {
      fprintf(stderr,"Probleme de memoire impossible allouer Ac\n") ;
      return ;
   }

  
		/* calcul des accroissements relatifs des taux 		*/
		/* entre model et donnees courantes			*/

   for (i = 0 ; i < SFic[LMS] ; i++) 
   {
        if (Tstruct[i] >= COUVERTURE_MIN && Model[i] >= COUVERTURE_MIN)
	   Ac[i] = (1.0 * Tstruct[i] - 1.0 * Model[i]) /
 		   (1.0 * Tstruct[i] + 1.0 * Model[i]) ; 
        else
           Ac[i] = MIN_PLAUSIBILITE ;
   }

		/* recherche des P meilleurs taux dans Tstruct		*/

   for (j = 0 ; j < NbStructures ; j++)
	for (i = SFic[LMS] - 1 ; i > j ; i--)

	     if (Ac[Best[i]] > Ac[Best[i-1]]) 
             	 SwapInt  (&(Best[i-1]), &(Best[i]))   ;
		
	     	/* les A premiers elements de Best designent		*/
		/* les A meilleurs accroissements dans Tstruct		*/


                /* affichage des meilleurs accroissements et 		*/
		/* structures corrrespondantes				*/

   fprintf(stdout,"\n\n%d Meilleurs accroissements relatifs\n",NbStructures) ;
   fprintf(stdout,"%%mod    %%data       Accroi    Structure\n\n") ;

   for (j = 0 ; j < NbStructures ; j++) 
   {
       k = Best [j] ;
    
       fprintf(stdout,"%5d   %5d    %10.5f     %s\n",
		      Model[k], Tstruct[k], Ac[k], SModel [k]) ;   
   }

   free(Ac)	  ;
}




/* --------------------------------------------------------------------	*/
/* calcul des NbStructures meilleures couvertures			*/
/* --------------------------------------------------------------------	*/

static void 
Couvertures(
	       Int32         SFic[]	  , 
	       Int32         Tstruct[]    , 
	       StructModel*  SModel	  , 
	       Int32*        Best 	  ,
	       Int32*	     Model        ,
               int           NbStructures   /* # structures candidates	*/
					    /* a considerer		*/	                             
	   )
{

  Int32 	i		;
  Int32 	j		;
  Int32 	k		;
  Int32*	Ac		;
  StructModel 	Structure 	;
 
  

   Ac = (Int32*) malloc(SFic[LMS] * sizeof(Int32)) ;

   if (!Ac) 
   {
      fprintf(stderr,"Probleme de memoire impossible allouer Ac\n") ;
      return ;
   }

  
		/* calcul des accroissements relatifs des taux 		*/
		/* entre model et donnees courantes			*/

   for (i = 0 ; i < SFic[LMS] ; i++)  
        Ac[i] =  Tstruct[i] ;       
 

		/* recherche des Nb_structures meill taux dans Tstruct  */

   for (j = 0 ; j < NbStructures ; j++)
	for (i = SFic[LMS] - 1 ; i > j ; i--)

	     if (Ac[Best[i]] > Ac[Best[i-1]]) 
             	 SwapInt(&(Best[i-1]), &(Best[i]))   ;
		
	     
		/* les A premiers elements de Best designent		*/
		/* les A meilleurs accroissements dans Tstruct		*/


                /* affichage des meilleurs accroissements et 		*/
		/* structures corrrespondantes				*/

   fprintf(stdout,"\n\n%d Meilleures couvertures\n",NbStructures) ;
   fprintf(stdout,"Couverture     Structure\n\n") ;

   for (j = 0 ; j < NbStructures ; j++)     
   {
       k = Best[j] ; 
       fprintf(stdout,"%5d     %s\n", Ac[k], SModel [k]) ;   
   }

   free(Ac)	  ;
}




/* --------------------------------------------------------------------	*/
/* calcul des longueurs tetes et boucles des palindromes		*/
/* --------------------------------------------------------------------	*/
static void 
CalculLongueurs(
       int   		lb	, 	/* temoin calcul longueur tete		*/
       int   		lt	,	/* temoin calcul longueur boucle	*/
       Int32 		N	,	/* Nombre de sequences			*/
       Int32*		NbPal	,	/* nb pals de chaque sequence		*/
       Palindrome**	Pal     ,	/* list pals de chaque sequence		*/
       int**		VTete   ,	/* valeurs tete des pals des sequences  */
       int**		VBoucle         /* valeurs boucles des pals des seq	*/
               )

{
  Int32 i 	   ;   
  Int32 m	   ;

   fprintf(stderr,"Calcule des longueurs...\n") ;

   for (i = 0 ; i < N ; i++)
				/* pour chaque sequence			*/

       for (m= 0 ; m < NbPal[i] ; m++) 
       {
				/* pour chaque pal de la sequence	*/
          if (lt) 
	  {

              if (Pal[i][m].FTete - Pal[i][m].DTete + 1 < LONGTETE)
                  VTete[i][m] = 0 ;
              else
		  VTete[i][m] = 1 ;
           }

	   if (lb)  
	   {

	      VBoucle[i][m] = (Pal[i][m].DQueue - Pal[i][m].FTete )
                              / INTERVBOUCLE ;

              if (VBoucle[i][m] >= NBVBOUCLE)
                  VBoucle[i][m] = NBVBOUCLE -1 ;               
	   }
       }
}


/* --------------------------------------------------------------------	*/
/* Calcule les distances entre 2 pals successifs d'une meme solution	*/
/* --------------------------------------------------------------------	*/
static int   
CalculDistances(
       int		td	, /* temoin de la distance		*/
       int**** 		d	, /* pointeur sur tableau des distances	*/
       Int32       	N	, /* N sequences		        */
       Int32* 		t       , /* nb solutions par seq i		*/
       Int32** 		p	, /* nb pal par solution j		*/
       Palindrome**	Pal	, /* palindromes des sequences		*/
       Int32*** 	Sol       /* tableau des solutions		*/
               )       
{

    Int32  i 	     ;
    Int32  j         ;
    Int32  r         ;
    Palindrome  p1         ;
    Palindrome  p2         ;


                                /* initialisation tableau distances	*/

   *d = (int ***) malloc(N * sizeof(int **)) ;
   CheckNullPtr("Pas assez de memoire pour allouer d\n",*d)


    for (i = 0 ; i < N ; i++) 
    {
         (*d)[i] = (int **) malloc(t[i] * sizeof(int *)) ;
  	  CheckNullPtr("Pas assez de memoire pour allouer d[i]\n",(*d)[i])

         for (j = 0 ; j < t[i] ; j++) 
	{

              (*d)[i][j] = (int *) malloc(p[i][j] * sizeof(int)) ;
              CheckNullPtr("Pas assez de memoire pour allouer d[i][j]\n",
                            (*d)[i][j])

         }
    }
  				/* calcule des distances		*/

   for (i = 0 ; i < N ; i++) 
				/* pour chaque sequence			*/

        for (j = 0 ; j < t[i] ; j ++) 
        {
                              /* pour chaque solution			*/

             (*d)[i][j][p[i][j]-1] = 0 ;

                                /* distance du pal droite a lui meme =0 */

              for (r = 0 ; r < p[i][j] -1 ; r ++) 
              {

                                /* pour chaque pal			*/

                  p1 = Pal[i][Sol[i][j][r]]   ; 
                  p2 = Pal[i][Sol[i][j][r+1]] ;
                  
                  if (Juxtapose(0, p1, p2)) 
                      (*d)[i][j][r] = p2.DTete - p1.FQueue - 1 ;

                  else if (Imbrique(0, p1 , p2)) 
                  {
                         if (td == 1 || td == 3) 
                             (*d)[i][j][r] = p2.DTete - p1.FTete - 1 ;
                         else
                             (*d) [i][j][r] = p1.DQueue - p2.FQueue - 1 ;
                  }
                         
                  else 
	          {     /* chevauche				*/
                            if (td == 1)
                                (*d) [i][j][r] = p2.DTete  - p1.FTete  - 1 ;
                            else if (td == 2) 
                                (*d) [i][j][r] = p1.DQueue - p2.FTete  - 1 ;
                            else 
                                (*d) [i][j][r] = p2.DQueue - p1.FQueue - 1 ;                                               
                  }

                  if (td == 1) 
		  {
                     (*d) [i][j][r] /= INTERVD1 ;

                     if ((*d) [i][j][r] >= NBVD1)
                        (*d) [i][j][r] = NBVD1 - 1 ;   
                  }

                  else if (td == 2) 
		  {
                     (*d) [i][j][r] /= INTERVD2 ;

                     if ((*d) [i][j][r] >= NBVD2)
                        (*d) [i][j][r] = NBVD2 - 1 ;   
                  }

                  else 
		  {
                     (*d) [i][j][r] /= INTERVD3 ;

                     if ((*d) [i][j][r] >= NBVD3)
                        (*d) [i][j][r] = NBVD3 - 1 ;   
                  }              
              }
        }
   return 1 ;
}



/* --------------------------------------------------------------------	*/
/* Libere espace memoire des tableaux de distance 			*/
/* --------------------------------------------------------------------	*/
static 
void FreeDistances(

       int*** 	d       ,	/* pointeur sur tableau des distances	*/
       Int32    N	, 	/* N sequences			        */
       Int32* 	t        	/* nb solutions par seq i		*/
                  )

{

   Int32 i ;
   Int32 j ;

   for (i = 0 ; i < N ; i++)

        if (d[i]) 
        {
            for (j = 0 ; j < t[i] ; j++)

                 if (d[i][j])
                     free(d[i][j]) ;
           free(d[i]) ;
        }

   free(d) ;
}



/* --------------------------------------------------------------------	*/
/* met a jour les appariements	correspondat aux structures candidates  */
/* --------------------------------------------------------------------	*/
static void 
MiseAjourAppariements
	    (
	     Int32 	 k		, 
	     Int32 	 NbInstances	, 
	     TInstance 	 Instances[]	, 
	     Int32** 	 Appariements	, 
	     Palindrome* Pals		,  
	     Int32** 	 Sols		
	    ) 
{
    Int32 a        = -1 ;
    Int32 b        = -1 ;
    Int32 u             ;
    Int32 v	        ;
    Int32 w	        ;
    Int32 Np       = -1 ;
    Int32 Longueur = -1 ;   
       
    for (u = 0 ; u < NbInstances ; u++)
	for (v = 0 ; v < k ; v++)
	{
	     Np       = Sols[Instances[u].Sol][Instances[u].Pal-v] ;
	     Longueur = Pals[Np].FTete - Pals[Np].DTete + 1 ;
		
	     for (w = 0 ; w <  Longueur ; w++)
	     {
	         a = Pals[Np].DTete  + w     ;
	         b = Pals[Np].FQueue - w     ;                 
	         (Appariements[a-1][b-1])++  ;                                   
             }
	}	 			       
}



/* --------------------------------------------------------------------	*/
/* Calcule les moments d'ordre 1 et 2 de la matrice d'appariements      */
/* --------------------------------------------------------------------	*/
static void 
Moments(
	int 	 Longueur	 ,
	Int32**  Appariements	 , 
	float*	 EX		 ,
	float*	 EcartT	         	 
       )
{

	int a				       ;
	int b				       ;
	float SX         =  0.0		       ;
	float SX2        =  0.0		       ;
	float EX2        = -1.0		       ;
	int   NbElements =    0		       ;
	
	for (a = 0 ; a < Longueur ; a++)
	     for (b = 0 ; b < Longueur ; b++)	     
	          if (Appariements[a][b] > 0)
		  {
		       SX  +=       Appariements[a][b]		            ;
		       SX2 += 1.0 * Appariements[a][b] * Appariements[a][b] ;
		       NbElements++					    ;
		  }	     

        if (NbElements > 0)
	{
		*EX     = SX  / NbElements		;
	 	 EX2    = SX2 / NbElements	        ;
		*EcartT = sqrt(EX2 - (*EX) * (*EX))	;
        }
	else
        {
		fprintf(stderr,"Elements matrice tous nuls\n") ;
		fprintf(stdout,"Elements matrice tous nuls\n") ;
                EX     = 0 ;
		EX2    = 0 ;
		EcartT = 0 ; 
	}
}



/* --------------------------------------------------------------------	*/
/* formatte une sortie 						        */
/* --------------------------------------------------------------------	*/
static void 
Formatte (
		FILE* 	fichier		, 
		int 	MaxChiffres	, 
		int 	Valeur
	 )
{
  
  int NbChiffres = 1  ;
  int i               ;

  while (Valeur >= pow(10, NbChiffres))
	 NbChiffres ++ ;       

  for (i = 1 ; i <= (MaxChiffres - NbChiffres) ; i ++)
      fprintf(fichier," ") ; 
       
}



/* --------------------------------------------------------------------	*/
/* Affiche la matrice des poids des appariements			*/
/* M[j][k] = p => l'appariement (j,k) apparait dans p structures	*/
/* --------------------------------------------------------------------	*/
static void 
AfficheAppariements(
   	            char* Seq 			, 
		    Int32 Longueur      	,
	            Int32 **Appariements	, 
		    int   Distingue		,
		    int   DebPart		,
		    int   FinPart		,
		    int   SeuilAppariement	
		   ) 
{
   Int32 a ;
   Int32 b ;
   
   fprintf(stdout, "Matrice des poids des appariements : \n\n") ;
   
				/* affichage en_tete colonne		*/

   if (Distingue)
   {   
        fprintf(stdout, "      ") ;
	
        for (b = 0 ; b < Longueur ; b++)
        {
	    if ((b+1) >= DebPart && (b+1) <= FinPart)
	       fprintf(stdout, "xxx") ;
	    else
	       fprintf(stdout, "   ") ; 
        }
        
	fprintf(stdout, "\n") ;	
   }
    
   fprintf(stdout, "     ") ;
   if (Distingue)
       fprintf(stdout, " ") ;
 
   for (b = 0 ; b < Longueur ; b++)
   {
	fprintf(stdout,"%d", b+1) ;	   
        Formatte(stdout,3,   b+1)   ;
   }

   fprintf(stdout,"\n     ") ;
   if (Distingue)
       fprintf(stdout, " ") ; 

   for (b = 0 ; b < Longueur ; b++)
	fprintf(stdout,"%c  ", Seq[b]) ;	
   
				/* affichage en_tete + ligne		*/
		    	
   for (a = 0 ; a < Longueur ; a++)
   {
       fprintf(stdout,"\n") ;       
       if (Distingue)
       {
	   if (((a+1) >= DebPart) && ((a+1) <= FinPart))
	       fprintf(stdout, "x") ;
	   else
	       fprintf(stdout, " ") ; 	   
       }
       fprintf(stdout,"%3d%c ", a+1,Seq[a]) ;
       
       for (b = 0 ; b < Longueur ; b++)
            if (Appariements[a][b] >= SeuilAppariement)
            {
 	        fprintf(stdout,"%d!", Appariements[a][b]) ;
                Formatte(stdout,2,Appariements[a][b])     ;         
            } 
            else
		fprintf(stdout, "   ") ;		       	    
   }

   fprintf(stdout,"\n") ;
}



/* --------------------------------------------------------------------	*/
/* Sauvegarde la matrice des poids des appariements			*/
/* M[j][k] = p => l'appariement (j,k) apparait dans p structures	*/
/* --------------------------------------------------------------------	*/
static int 
SauvegardeAppariements(
		        int     MAX_V		,
			int 	Alignees	, 
			Int32 	k		, 
			Int32 	NSeq		, 
			char* 	NFAppariement	, 
			Int32 	Longueur	,
	                Int32 **Appariements    
		      ) 
{

				/* NSeq = num Seq si non alignees, NbSeq sinon */
   Int32 a            ;
   Int32 b            ;
   FILE* fichier      ;
   char  numero[5]    ; 
   char  taille[5]    ; 
   char Nfichier[100] ;   
   int Pixel = 0      ;        /* valeur en Pixel correspondant au poids d'un */
                               /* appariement				      */				 
  
   fprintf(stderr,"nomfichier = %s\n",NFAppariement) ;
   Nfichier[0] = 0                 ;
   strcat(Nfichier, NFAppariement) ; 
   if (!Alignees)
   {   
       sprintf(numero,"%d",NSeq)       ;
       strcat(Nfichier,"_")            ;
   }
   sprintf(taille,"%d", k)         ;    
   strcat(Nfichier,taille)         ;
   strcat(Nfichier,"_")            ;
   strcat(Nfichier,numero)         ;
   strcat(Nfichier,".app")         ;

   fprintf(stderr,"%s %d\n",Nfichier, Longueur) ;

   fichier  = fopen(Nfichier,"w") ;
   
   if (!fichier)
   {
	fprintf(stderr,"Impossible de creer le fichier %s\n",Nfichier) ;
        return 0 ;
   }   
   
						/* sauvegarde				*/
   for (a = 0 ; a < Longueur ; a++) 
   {             
       for (b = 0 ; b < Longueur ; b++)   
       {           
	  if (Alignees)
             Pixel = (int) floor(0.5 + 1.0 * Appariements[a][b] / NSeq) ;
             
          else
	     Pixel =   (int) (MIN_PIXEL + (Appariements[a][b] - MIN_V) * 
                              1.0 * (MAX_PIXEL - MIN_PIXEL) / (MAX_V - MIN_V)) ;           

          if (Pixel > MAX_PIXEL) 
              Pixel = MAX_PIXEL ;
          fprintf(fichier, "%d\t", Pixel) ;    
       }

       fprintf(fichier,"\n") ;
   }

   fprintf(fichier,"\n") ;

   close(fichier) ;
   return 1 ;  
}


/* --------------------------------------------------------------------	*/
/* Prepare les donnees pour la representation DOT_PLOT			*/
/* --------------------------------------------------------------------	*/
static int 
Dot_plot	     (
			int     MAX_V		, /* poids absolu app.	*/
			int     dot_seq		, /* 1 plot / sequence	*/
                        char*   sequence        , /* Sequence		*/			
			StructModel Structure   , /* code structure	*/
			int 	Alignees	, /* flag aligne	*/
			Int32 	k		, /* longueur structure	*/
			Int32 	NSeq		, /* Numero Sequence	*/
						  /* ou nombre sequence */
			char* 	NFAppariement	, /* nom fichier	*/
			Int32 	Longueur	, /* Longueur sequence	*/
	                Int32 **Appariements    , /* matrice appariem.	*/
			int   Distingue		, /* ARN connu dans seq */
			int   DebPart		, /* debut ARN		*/
			int   FinPart		  /* fin ARN		*/
		      ) 
{

				/* NSeq = num Seq si non alignees, NbSeq sinon  */
   unsigned int a             ; /* indice ligne matrice				*/
   unsigned int b             ;	/* indice colonne matrice			*/
   FILE*   fichier       ;	/* nom interne fichier resultat			*/				
   char    numero[5]     ; 	
   char    taille[5]     ; 
   char    Nfichier[100] ;  
   char    Structure1[100] ; 
   char    Titre[200]	 ; 	/* Titre association a la representation dot 	*/
   float   Pixel = 0.0   ;      /* valeur en Pixel correspondant au poids d'un 	*/
                                /* appariement				 	*/
   float** pap = NULL    ;      /* poids des appariements		        */
   int     j		 ; 	/* indice nucleotide dans palindrome		*/
   int NbAppariements = 0 ; 	/* nombre d'appariements			*/
   int FinArnt	         ;      /* place derniere base ARNt dans la sequence	*/
   float MaxPoidsApp = 0.0 ;    /* appariement de plus fort poids		*/    


   				/* minuscules pour nucleotides hors ARN		*/
   if (Distingue)
   {
      for (a = 0 ; a < Longueur ; a++)
      {
	  if ((a+1 < DebPart || a+1 > FinPart) && sequence[a] < 95)                      
              sequence[a] = sequence[a] + 32 ;                       
      }
	
   }
    
	 pap = (float**) malloc(Longueur * sizeof(float*))                   ;
         CheckNullPtr("Pas assez de memoire pour allouer pap\n", pap);	
	  
	 for (a = 0 ; a < Longueur ; a++)
	 {
	     pap[a] = (float*) malloc(Longueur * sizeof(float))                  ;
             CheckNullPtr("Pas assez de memoire pour allouer pap[a]\n",pap[a]) ;	 
	 }

         				/* initialisation du tableau pap	*/	
					
          for (a = 0 ; a < Longueur ; a++) 
	      for (b = 0 ; b < Longueur ; b++)
	           pap[a][b] = 0.0 ;				 
  

					/* Recherche de la plus grande Valeur dans Appariements */
 
      for (a = 0 ; a < Longueur ; a++) 
	      for (b = 0 ; b < Longueur ; b++)
                  if ((1.0 * Appariements[a][b]) > MaxPoidsApp)
	             MaxPoidsApp =  1.0 * Appariements[a][b] ;	


					/* constitution du nom du fichier	*/

   EnleveBlancs(Structure) ;
   fprintf(stderr,"nomfichier = %s\n",NFAppariement) ;

   Nfichier[0] = 0                 ;   
   strcat(Nfichier, NFAppariement) ; 
   if (!Alignees)
   {   
       sprintf(numero,"%d",NSeq)       ;
       strcat(Nfichier,"_")            ;
       strcat(Nfichier,numero)         ;
   }

   if (!dot_seq)
   { 
   	strcat(Nfichier,"_") ;
   	strcat(Nfichier, Structure)     ;
   }

   strcat(Nfichier,".ps")          ;
   fprintf(stderr,"%s \n", Nfichier) ;

					/* constitution du titre		*/
   Titre[0] = 0 		   ;
   strcat(Titre, NFAppariement)    ;
   strcat(Titre, "   : Sequence  ") 	   ;
   strcat(Titre, numero)  	   ;    

   if (!dot_seq)
   {
   	strcat(Titre, "  Structure  ")  ;
   	strcat(Titre, Structure)        ;
   }

   fprintf(stderr,"%s %d\n",Nfichier, Longueur) ;

     
						/* constitution de pap		*/
   for (a = 0 ; a < Longueur ; a++) 
   {             
       for (b = 0 ; b < Longueur ; b++)   
       {  
	 if (MAX_V) 
	 {         
	  if (Alignees)
             pap[a][b] =  floor(0.5 + 1.0 * Appariements[a][b] / NSeq) ;
             
          else
	     pap[a][b] =  (MIN_PIXEL + (Appariements[a][b] - MIN_V) * 
                           1.0 * (MAX_PIXEL - MIN_PIXEL) / (MAX_V - MIN_V)) ;                        
         }
	 else		   /* poids relatif des appariements			*/
	     pap[a][b] =  (MIN_PIXEL + (Appariements[a][b] - MIN_V) * 
                          1.0 * (MAX_PIXEL - MIN_PIXEL) / (MaxPoidsApp - MIN_V)) ;    
       }
   }

 

   if (1)
   {     
	 FinArnt = Longueur-4 ;
						/* poids appariements tRNA	*/
  						/* sur diagonale inferieure	*/

						/* 1er palindrome		*/
	for (j = 0 ; j < 7 ; j++)	
 	    pap[DebPart+FinArnt-1-j-1][DebPart+1-1+j-1] = MAX_PIXEL ;

						/* 2eme palindrome		*/
	for (j = 0 ; j < 4 ; j++)	
 	    pap[DebPart+25-1-j-1][DebPart+10-1+j-1] = MAX_PIXEL ;	

						/* 3eme palindrome		*/
	for (j = 0 ; j < 5 ; j++)	
 	    pap[DebPart+43-1-j-1][DebPart+27-1+j-1] = MAX_PIXEL ;	

					        /* 4eme palindrome		*/
	for (j = 0 ; j < 5 ; j++)	
 	    pap[DebPart+FinArnt-7-1-j-1][DebPart+49-1+j-1] = MAX_PIXEL ;	

    }

   PS_dot_plot(Titre, sequence, pap, Nfichier) ;

   return 1 ;  
}




/* --------------------------------------------------------------------	*/
/* etablit les appariements impliques dans les structures candidates	*/
/* --------------------------------------------------------------------	*/
static int 
FAppariements(
       int MAX_V		, /* poids absolu des appariements	*/
       int dot_seq		, /* 1 "dot plot" /seq			*/
       int Alignees		, /* 1 si sequences alignees		*/
       int Distingue		, /* distingue partie d'un ARN		*/
       Int32* DebPart		, /* debut morceau ARN mis en evidence  */
       Int32* FinPart		, /* fin morceau ARN mis en evidence    */
       char* NFAppariement      , /* nom fichier des appariements	*/
       Int32  Lmax		, /* longueur max d'une sequence	*/
       Int32  N			, /* nombre de sequences		*/
       char   *s		, /* tableau des sequences		*/
       Int32  NbPal[]		, /* nombre pal de chaque sequence	*/
       Palindrome   *Pals[]     , /* pal de chaque sequence		*/
       WordBunchPtr AllStr	, /* positions des lintstructures	*/
       TPosition    Pos[]	, /* vecteur des positions des struc	*/
				  /* dans les sequences et solutions	*/
       Int32 ***Sols            ,
       Int32 ***M[]		, 
       StructModel* SModel	, /* classses des structures		*/
       Int32*       Best	, /* indices classes meilleures classes	*/
       int          NbStructures  /* # structures candidates considerees*/
              )
       

{       
  Int32  i		= -1	   ; /* incide des sequence		*/
  Int32  k		= -1	   ; /* taille courante structure	*/
  Int32  a		= -1	   ; /* ligne matrice appariements	*/
  Int32  b	 	= -1	   ; /* colonne matrice appariements	*/
  Int32  ea		= -1	   ; /* nb classes structures taille k  */
  Int32  c		= -1	   ; 
  StackiHdle Structures	= NULL	   ;
  Int32  val		= -1	   ;
  Int32 Longueur	= -1	   ;
  Int32 Seq		= -1	   ; /* numero sequence traitee		*/
  Int32 Sol		= -1	   ; /* numero solution traitee		*/
  Int32 Pal		= -1       ; /* numero palindrome traitee	*/
  int   premier		= -1	   ; /* temoin 1er elet d'une classe	*/
  Int32 u		= -1	   ; /*					*/
  TInstance* Instances  = NULL     ; /* Occurrences des structures	*/
  Int32 NbInstances     = 0	   ; /* Nb instances de structures	*/
  Int32** Appariements  = NULL     ; /* matrice des appariements	*/
  Int32 ClasseInstance  = 0        ; 
  Int32 MaxInstances    = 0        ;
  float EX	        = -1.0     ; /*  moyenne #appariements		*/
  float EcartT		= -1.0     ; /* Ecart type # appariements	*/
  int   SeuilAppariement= 0        ;
  StructModel Structure 	   ;
  int   ok_classe 	= 0	   ;
  StructModel Tampon		   ;	
  int   lu = 0			   ; /* element d'un pile est lu	*/
  int   iterer    = 1		   ; /* continue lire elts classe	*/
  int   instance   = 0		   ; /* pas d'instance de la structure	*/
				     /* courante			*/
  
 
 
  fprintf(stderr,"calcul des matrices d'appariements en cours...\n") ;
  
 
  fprintf(stdout,"Nombre de sequences : %d\n\n",N)  ;

 
			/* reinitialise pointeurs de piles en haut	*/
			

  CursiToTop(AllStr->nbwords) ;
  CurspToTop(AllStr->words)   ;

			/* pour chaque longueur de mots k		*/
    
  for (k = AllStr->maxlen ; k >= AllStr->minlen ; k--) 
  {
			
      fprintf(stdout,"\n\ntaille             : %d \n",k) ;

			/* lit le nombre de classes de taille k 	*/
     
      if (!ReadiDown(AllStr->nbwords,&ea)) 
      {
 	  fprintf(stderr,"Impossible de lire nombre ea (%d)\n", ea) ;
 	  return 0 ;	 
      }

      fprintf(stdout,"nombre de classes  : %d\n\n",ea)     ;

			/* ecrit les positions debut structure		*/
			/* (pal + a droite) de chaque classe		*/

      ReadpDown(AllStr->words,(Ptr *) &Structures) ;

		/* curseur en haut de chaque pile 			*/

     for (c = 0 ; c < ea ; c++)
          CursiToTop(Structures[c]) ;

     
     if (Alignees)
     {						/* longueur seq 0	*/
         Longueur = strlen(s) ;     
	 Appariements = NULL ;
	 Appariements = (Int32**) malloc(Longueur * sizeof(Int32*))                   ;
         CheckNullPtr("Pas assez de memoire pour allouer Appariements\n",Appariements);	
	  
	 for (a = 0 ; a < Longueur ; a++)
	 {
	     Appariements[a] = (Int32*) malloc(Longueur * sizeof(Int32))                  ;
             CheckNullPtr("Pas assez de memoire pour allouer Appariements\n",Appariements[a]) ;	 
	 }

         				/* initialisation du tableau des appariements	*/	
					
          for (a = 0 ; a < Longueur ; a++) 
	      for (b = 0 ; b < Longueur ; b++)
	           Appariements[a][b] = 0 ;
     }
   
							    /* pour chaque sequence	*/
     for (i = N-1 ; i >= 0 ; i--)
     {	 
	 fprintf(stderr,"sequence %d\n",i) ;
	 Longueur = strlen(s+i*Lmax)       ;


	 if (dot_seq && !Alignees)
			      {
				 NbInstances    = 0	   ; 
                                 ClasseInstance = 0        ; 
                                 MaxInstances   = 0        ;	                                    
                                 instance       = 0        ;	
				   
							    /* allocation matrice apparie.*/
					
	                         Appariements = (Int32**) malloc(Longueur * sizeof(Int32*))                   ;
                                 CheckNullPtr("Pas assez de memoire pour allouer Appariements\n",Appariements);	
	  
	                         for (a = 0 ; a < Longueur ; a++)
	                         {
	                          Appariements[a] = (Int32*) malloc(Longueur * sizeof(Int32))               ;
                                  CheckNullPtr("Pas assez de memoire pour allouer Appariements\n",Appariements[a]) ;	 
	                         }	     
					/* initialisation du tableau des appariements	*/	
					
                                 for (a = 0 ; a < Longueur ; a++) 
	                              for (b = 0 ; b < Longueur ; b++)
	                                  Appariements[a][b] = 0 ;	

				 MaxInstances = 0 ;

				for (c = 0 ; c < ea ; c++)
	    			    MaxInstances += Structures[c]->top ; 
						     	  			
				Instances =  malloc(MaxInstances * sizeof(TInstance))                   ;
        			CheckNullPtr("Pas assez de memoire pour allouer Instances\n",Instances) ;
			
                             }

        			 					
					/* pour chaque classe de longueur k		*/

         for (c = 0 ; c < ea ; c++)  
         {	
              if (!dot_seq)
              {	        				  	          
                 NbInstances    = 0	   ; 
                 ClasseInstance = 0        ; 
                 MaxInstances   = 0        ;
	                  
                 instance       = 0        ;	      
	      }	

	      iterer  = 1   ;   
	      premier = 1   ;
	         				           	     
	      
	      while (iterer && ReadiDown(Structures[c],&val))	      
	      {
		Seq  = Pos[val].Seq      ; 
		Sol  = Pos[val].Sol      ;
		Pal  = Pos[val].Pal      ;		  
	   	
	        if (premier) 
		{  
 		       premier = 0         ;			 			  
		       Structure[0] = '\0' ;	
		  	  				/* recupere classe structure	*/
		       for (u = k - 1 ; u >= 0 ; u--)	
 		       {		 
                            Tampon[0] = '\0' ;

                            if (u == 0)
				sprintf(Tampon, "%d", M[Seq][Sol][Pal][Pal-u])  ;
 			    else
				sprintf(Tampon, "%d ",M[Seq][Sol][Pal][Pal-u])  ;	

			    strcat(Structure,Tampon) ;
			}

			ok_classe = DansSelection(Structure, SModel, Best, NbStructures) ;
			  
			if (ok_classe && Seq == i)                     
 			{
			   fprintf(stderr,"%s\n", Structure) ;
      

			   if (!Alignees)
                           {
			      if (!dot_seq)
			      {
                                 Appariements = NULL       ;          
							    /* allocation matrice apparie.*/
					
	                         Appariements = (Int32**) malloc(Longueur * sizeof(Int32*))                   ;
                                 CheckNullPtr("Pas assez de memoire pour allouer Appariements\n",Appariements);	
	  
	                         for (a = 0 ; a < Longueur ; a++)
	                         {
	                          Appariements[a] = (Int32*) malloc(Longueur * sizeof(Int32))               ;
                                  CheckNullPtr("Pas assez de memoire pour allouer Appariements\n",Appariements[a]) ;	 
	                         }	     
					/* initialisation du tableau des appariements	*/	
					
                                 for (a = 0 ; a < Longueur ; a++) 
	                              for (b = 0 ; b < Longueur ; b++)
	                                  Appariements[a][b] = 0 ;	

			          MaxInstances = Structures[c]->top ;
                               					  			
	                          Instances =  malloc(MaxInstances * sizeof(TInstance))                   ;
                                  CheckNullPtr("Pas assez de memoire pour allouer Instances\n",Instances) ;

                             } /* fin de !dot_seq */

 	                   } /* fin de !Alignees */

                        } /* fin de ok_classe && Seq == i*/	                          			  
		        		        		             			      	      		                      
		}					/* fin if premier	*/
	
		if (ok_classe && Seq == i)  
                {    
		   		/* au moins une instance de la structure 	*/
		   instance = 1 ;                     		   					  
                   InsereInstance(ClasseInstance, k, Instances, 
				  Sol, Pal,  &NbInstances, Sols[i]) ;                  
                   		
		}
		else
                {                   
		   iterer = 0 ;	
		   	 		/* element lu appartient a sequence i-1			*/
					/* ou pas bonne classe					*/  
		     			/* on se repositionne sur element precedent 		*/
		   (Structures[c]->cursor)++ ;		
		}	
						
            } /* fin de While iterer Read ... */	  			   
             
	    if (!dot_seq && instance) 
            {		     		     
		    
		       ClasseInstance = NbInstances ;
		    
		          				    	/* instances identifiees		*/      
		       MiseAjourAppariements(k, NbInstances, Instances, Appariements, 
			                     Pals[i], Sols[i])               ;
		       Moments(Longueur , Appariements, &EX, &EcartT)	                            ;
		       SeuilAppariement = floor(EX)   ;	 
			    
		       fprintf(stdout, "sequence %d \n"   , i)		     ;
		       fprintf(stdout, "longueur %d \n"   , Longueur)	     ;			    
		       fprintf(stdout, "%d palindromes \n", NbPal[i])	     ;
		       fprintf(stdout, "%d instances \n"  , NbInstances)     ;	
		       fprintf(stdout, "taille %d \n"     , k)		     ;	
		       fprintf(stdout, "Moyenne = %.2f \n",EX)		     ;
		       fprintf(stdout, "Ecart   = %.2f \n",EcartT)	     ;    
		       fprintf(stdout, "Seuil   = %d \n"  ,SeuilAppariement) ;
      
                       if (!Alignees && NbInstances > 0)
                       {                        
                          Dot_plot ( 
				      MAX_V,
				      dot_seq,
				      s+i*Lmax, 
				      Structure,
				      Alignees, 
				      k, 
				      i, 
				      NFAppariement, 
				      Longueur,
	                              Appariements,
				      Distingue,
				      DebPart[i],
				      FinPart[i]
                                     ) ;
                        }    
             
	     				/* libere le tableau des appariements		*/
                       for (a = 0 ; a < Longueur ; a++)
	               free(Appariements[a])       ;
	               free(Appariements) ;     

		       free(Instances) ;
      
            } /* ! dot_seq && instance */                  		       
		     					                         			      
      } /* fin for c	*/ 
 
      if (dot_seq && instance)
      {
	               ClasseInstance = NbInstances ;
		    
		          				    	/* instances identifiees		*/      
		       MiseAjourAppariements(k, NbInstances, Instances, Appariements, 
			                     Pals[i], Sols[i])               ;
		       Moments(Longueur , Appariements, &EX, &EcartT)	                            ;
		       SeuilAppariement = floor(EX)   ;	 
			    
		       fprintf(stdout, "sequence %d \n"   , i)		     ;
		       fprintf(stdout, "longueur %d \n"   , Longueur)	     ;			    
		       fprintf(stdout, "%d palindromes \n", NbPal[i])	     ;
		       fprintf(stdout, "%d instances \n"  , NbInstances)     ;	
		       fprintf(stdout, "taille %d \n"     , k)		     ;	
		       fprintf(stdout, "Moyenne = %.2f \n",EX)		     ;
		       fprintf(stdout, "Ecart   = %.2f \n",EcartT)	     ;    
		       fprintf(stdout, "Seuil   = %d \n"  ,SeuilAppariement) ;
      
                       if (!Alignees && NbInstances > 0)
                       {                        
                          Dot_plot ( 
				      MAX_V,
				      dot_seq,
				      s+i*Lmax, 
				      Structure,
				      Alignees, 
				      k, 
				      i, 
				      NFAppariement, 
				      Longueur,
	                              Appariements,
				      Distingue,
				      DebPart[i],
				      FinPart[i]
                                     ) ;
                        }    
             
	     				/* libere le tableau des appariements		*/
                       for (a = 0 ; a < Longueur ; a++)
	               free(Appariements[a])       ;
	               free(Appariements) ;     
		       free(Instances) ;
      }
 
    }  /* fin for i	*/
       
    if (Alignees && NbInstances > 0)
    {            
       Dot_plot             (      MAX_V         ,       
				   dot_seq	 ,
				   NULL          ,
				   Structure     ,
				   Alignees      ,
			 	   k             , 
				   N             , 
				   NFAppariement , 
				   Longueur      , 
	    			   Appariements  ,
				   Distingue	 ,
				   DebPart[i]    ,
				   FinPart[i]
                            )    ;     
      
       for (a = 0 ; a < Longueur ; a++)
	   free(Appariements[a])       ;
	          
       free(Appariements) ;    
    }
  }								/* fin for k	*/
 
  fprintf(stderr,"Fin sauvegarde des resultats.\n") ;
  return 1	           ;
}								/* fin fonction	*/



/* --------------------------------------------------------------------	*/
/* Main									*/
/* --------------------------------------------------------------------	*/
int  main(
    		Int32     argc    ,
    		char**	  argv    
	 )
{
  /* variables								*/


  Int32    Lmax = 250   ;       /*  longueur max d'une sequence		*/
  char     Symb         ;       /* symbole d'un argument		*/
  char     Argument[50] ;       /* argument de main			*/
  Int32    N	 =    0	;	/* nombre de sequences			*/
  Int32   ns  	 =    0	;       /* nombre total de solutions		*/
  Int32    n	 =    0	;	/* nombre total de palindromes		*/
  Int32 *    t	 = NULL	;	/* nombre solutions seq i		*/
  Int32 **   p	 = NULL	;  	/* p[i][j] nombre pal seq i solu j  	*/
  char   *s      = NULL ;       /* tableau des sequences		*/
  Int32 *NbPal   = NULL ;       /* nombre pals sequence i		*/
  Palindrome **Pal     = NULL ;	/* Pal[i][m] = pal m seq i 		*/
  Int32 ***Sol   = NULL ;      	/* Sol[i][j][m] indice du pal m         */
				/* sol j seq i dans Pal[i]		*/
  Int32 **** M   = NULL	;       /* matrice de description des 		*/
				/* structures de chaque solution	*/
  TPosition *Pos = NULL ;	/* vecteur des positions des elements	*/
  Int32 Quorum	 =   80 ;	/* taux minimal de repetition		*/
  Int32 *Va      = NULL ;	/* index des classes 			*/ 
  Int32   Rmin   =    0 ;       /* nombre minimal de repetition		*/
  float Rminimal =    0 ;
  Int32      e   =    1 ;       /* nb classe des structures de taille k */
  Int32  minlen  =    1 ;       /* taille minimale des structures       */
  Int32  maxlen  =   25 ;       /* taille max des structures 		*/
  int    Verbose =    2 ;       /* taux de blabla			*/
  int    f       =    0 ;       /* temoin sauvegarde des structures	*/
				/* dans un fichier 			*/
  char   Nfichier[50]   ;       /* nom du fichier de sauvegarde		*/
  WordBunchPtr AllStr=NULL   ;  /* contient toutes les positions des 	*/
				/* structures repetees			*/
  FILE  *entree         ;       /* fichier d'entree (debugger CNRS)     */

  int st = 0            ;       /* temoin sauvegarde taux structures    */
  char Nfichierst[50]   ;       /* nom du fichier de taux               */

  Int32   Fic[7] = {0, 1, 3, 15, 105, 945, 10395} ;
				/* nb structures de longueur 0..6	*/

  Int32  SFic[7] = {0, 1, 4, 19, 124, 1069, 11464} ;
				/* nb structures de longuer <= 0..<=6   */ 
 
  Int32  *Tstruct= NULL ;       /* tableau des taux de chaque structure */
  Int32 i               ;

  int a= 0		;       /* temoin de calcul des accroissements	*/

  char NfichierAc[50]   ;       /* nom fichier modele des taux		*/

  int ast=0 		;       /*  temoin ancienne valeur de st	*/

  int lb       = 0	;	/* temoin tenir compte longueur boucle  */
  int lt       = 0      ;       /* temoin tenir compte longueur tete    */
  int purete   = 0      ;       /* temoin tenir compte purete		*/

  int **VTete    = NULL ;       /* tab valeurs associees aux tetes pals */
  int **VBoucle  = NULL ;       /* tab valeurs associees aux boucles    */

  int Moyenne  = 0 	;	/* Temoin affichage nombre moyen de 	*/
				/* solution par sequence contenant 	*/
				/* une structure			*/

  int Ecart = 0		;       /* ecart type du nombre moyen de sol	*/
				/* par seq contenant une structure	*/

  int td1 = 0           ; 	/* temoin distance 1			*/
  int td2 = 0		;       /* temoin distance 2			*/
  int td3 = 0		;       /* temoin distance 3			*/

  int ***d1 = NULL      ;       /* tableau des distances 1		*/
  int ***d2 = NULL      ;       /* tableau des distances 2		*/
  int ***d3 = NULL      ;       /* tableau des distances 3		*/

  int check = 0         ;       /* temoin verification distribution     */
                                /* structure   6000 TRNA                */

  int   Dtrna     = 0        ;  /* nb bases avant et apres TRNA         */
  Int32* DebPart  = NULL     ;  /* debut du morceau ARN mis en evidence	*/
  Int32* FinPart  = NULL     ;  /* fin du morceau ARN mis en evidence   */  
  int Appariement = 0        ;
  char NFAppariement[100]    ;    
                                /* nom fichier sauvegarde matrice app.	*/
  int  Distingue  =  0  ;       /* = 1 <=> distingue morceaux sequen.   */
  int  Alignees	  =  0  ;       /* sequences non alignees		*/
  StructModel* SModel	= NULL	     ; 
  Int32*       Best     = NULL	     ;
  Int32*       Model    = NULL	     ;
  int dot_seq           = 0          ; /* 1 dot plot / seq		*/
  int MAX_V		= 0  	     ; /* poids relatif (0)des appari.  */      
  int ToutesCandidates	= 0          ; /* considere seulement	 	*/
				       /* NB_STRUCTURES candidates	*/
  int NbStructures = 0               ; /* # NB_STRUCTURES candidates	*/
				       /* que l'on considere. 0=toutes	*/				       
  
 
				/* initialisation			*/



  fprintf(stderr," kmrstr8 (juin 98) \n") ;

  while (SelectArg("-q -f -v -l -m -n -a -p -b -t -o -e -h -i -j -c -d -k -z -w -x -y -u",
           &Symb,Argument,argc,argv)) 
  {

	  switch (Symb) 
          {
					/* taux minimum de repetition		*/

		 case 'q' : ARGUMENT(&Quorum,Symb,Argument)      ;
			    break	                        ;

					/* taux d'affichage			*/

		 case 'v' : ARGUMENT(&Verbose,Symb,Argument)     ;
			    break	                        ;

					/* sauvegarde position des structures	*/

		 case 'f' : f = 1 				;
			    if (strlen(Argument) == 0) 
                            {
			       fprintf(stderr,"Argument -f non valable%s\n",
					       Argument) 	;
			       return 0				;
			    }
			    strcpy(NfichierAc,Argument)		;
			    break				;

		  case 'a' : a = 1 				;
			     if (strlen(Argument) == 0) 
                             {
			       fprintf(stderr,"Argument -a non valable%s\n",
					       Argument) 	;
			       return 0				;
			     }
			     strcpy(NfichierAc,Argument)	;
			     fprintf(stdout,"modele aleatoire	: %s\n",NfichierAc) ;
			     break				;

		  case 'w' : 				;
			     if (strlen(Argument) == 0) 
                             {
			       fprintf(stderr,"Argument -w non valable%s\n",
					       Argument) 	;
			       return 0				;
			     }
			     strcpy(NFAppariement,Argument)	;
			     fprintf(stdout,"Nom fichier appariement	: %s\n",NFAppariement) ;
			     break				;

					/* sauvegarde des taux des structures  */

                 case 'e' : st = 1 				;
			    if (strlen(Argument) == 0) 
			    {
			       fprintf(stderr,"Argument -e non valable%s\n",
					       Argument) 	;
			       return 0				;
			    }
			    strcpy(Nfichierst,Argument)		;
			    break				;

 		case 'd' :  Appariement = 1   			;
                            if (strlen(Argument) == 0) 
			    {
			       fprintf(stderr,"Argument -d non valable%s\n",
					       Argument) 	;
			       return 0				;
			    }
			    strcpy(NfichierAc,Argument)         ;
 			    fprintf(stderr,"modele aleatoire%s\n",NfichierAc)  ;
			    break				;

		case 'x' :  dot_seq = 1   			;                            
			    break				;
		
		case 'y' :  ARGUMENT(&MAX_V,Symb,Argument)      ;                         
			    break				;

			/* max taille struct 	*/
		 case 'm' : ARGUMENT(&maxlen,Symb,Argument)     ;
			    break	                        ;

						/* min taille struct	*/

		 case 'n' : ARGUMENT(&minlen,Symb,Argument)      ;
			    break	                        ;

						/* longueur max seq	*/

                 case 'l' : ARGUMENT(&Lmax,Symb,Argument)        ;
			    break	                        ;

						/* tenir compte longueur tete	*/

                 case 't' : lt = 1 				;
			    break				;


						/* tenir compte longueur boucle	*/
                 case 'b' : lb = 1 				;
			    break				;

						/* tenir compte purete		*/
                 case 'p' : purete = 1 				;
			    break				;

						/* afficher nombre moyen de solution */
						/* par seq contenant une structure   */
						/* donne                             */
		 case 'o' : Moyenne = 1 			;
			    break				;
			
                  				/* calcul distance 1	*/
                 case 'h' : td1 = 1 				;
			    break				;

    						/* calcul distance 2	*/
                 case 'i' : td2 = 2 				;
			    break				;

    						/* calcul distance 3	*/
                 case 'j' : td3 = 3 				;
			    break				;

					        /* verifie distribution structure 6000	*/
                 case 'c' : check = 1 		                ;
                            Ecart = 1                           ;
                            if (Argument[0] != '0') 
                                Dtrna = 25                      ;  
                             fprintf(stderr,"%d\n",Dtrna)       ;
			    break				;

 		case 'k' :  Distingue = 1 		        ;
			    break				;

		case 'z' :  Alignees = 1 		        ;
			    break				;

						/* # structures candidates a envisager	*/

                 case 'u' : ARGUMENT(&NbStructures,Symb,Argument) ;
			    break	                          ;

	  }
    }

  entree = stdin ;
 
  if (a || f || Appariement) 
      {
     	ast = st ;
        st  = 1  ;
      }

  if (LectureDonnees(Distingue,&DebPart,&FinPart,entree,&N,Lmax,
                     &s,&ns,&n,&t,&p,&NbPal,&Pal,&Sol) &&
      ConstituerMatriceInitiale(&M,N,t,p) &&
      ConstituerPosition(&Pos,n,N,t,p) &&
      ConstituerVaInitial(n,&Va)) 
  {    
      if (lt) 
      {
         VTete = (int **) malloc(N * sizeof(int *)) ;
         CheckNullPtr("Pas assez de memoire pour allouer VTete\n",VTete)

         for (i = 0 ; i < N ; i++) 
         {
              VTete[i] = (int *) malloc(NbPal[i] * sizeof(int)) ;
	      CheckNullPtr("Pas assez de memoire pour allouer VTete[i]\n",VTete[i])
         }
      }

      if (lb) 
      {
           VBoucle = (int **) malloc(N * sizeof(int *)) ;
           CheckNullPtr("Pas assez de memoire pour allouer VBoucle\n",VBoucle)

           for (i = 0 ; i < N ; i++) 
	   {
               VBoucle[i] = (int *) malloc(NbPal[i] * sizeof(int)) ;
	       CheckNullPtr("Pas assez de memoire pour allouer Vboucle[i]\n",
                             VBoucle[i])
           }
      }

      if (lb || lt)
	  CalculLongueurs(lb,lt,N,NbPal,Pal,VTete,VBoucle) ;
       
      if (td1 && !CalculDistances(td1, &d1,N,t,p,Pal,Sol))
          return 0 ;

      if (td2 && !CalculDistances(td2, &d2,N,t,p,Pal,Sol))
          return 0 ;

      if (td3 && !CalculDistances(td3, &d3,N,t,p,Pal,Sol))
          return 0 ;
     
      Rminimal = N * Quorum / 100.0 ;
      Rmin = (Rminimal + 0.5) / 1 ;
      if (!Rmin)   
	 Rmin++ ;

      if (Verbose >= 1)
	 SetUpKmrNotify(Vrai) ;
      fprintf(stderr,"\n\n")  ;
      fprintf(stdout,"Kmrs version 8.0\n") ;
      fprintf(stdout,
             "%d sequence(s), au minimum %d repetition(s) (environ %d pour cent)             \n\n", N,Rmin,Quorum) ;

      
      (void) StrCpuTime(TIME_RESET) ;
      
      if (st || Appariement) 
      {
	    Tstruct =  (Int32 *) malloc(SFic[LMS] * sizeof(Int32)) ;

	    for (i = 0 ; i < SFic[LMS] ; i ++)
                 Tstruct[i] = 0 ;
      }

      AllStr = DoKmrS(Dtrna,check,td1, d1, td2, d2, td3, d3,
                      Ecart,Moyenne,lt,VTete,lb,VBoucle,purete,
                      st,Fic,SFic,Tstruct, (f || Appariement),
                      Verbose,Va, minlen, maxlen, 0, &e, n, N ,
		      Rmin,t,p,Pos,Pal,Sol,M) ;

      fprintf(stdout,StrCpuTime(TIME_RESET)) ;
      
       
      if (a || f || Appariement)
      {
         CreationModele(NfichierAc, &SModel, &Best, SFic, &Model) ;

         if (NbStructures == 0 )
            NbStructures = SFic[LMS] ;
         
        if (a) 
      		/* calcul et mise en evidences de P meilleurs 		*/
		/* accroissements relatifs				*/
	        Accroissements(SFic,Tstruct, SModel, Best, Model, NbStructures)  ;                  
        else
               /* on ordonne selon la couverture			*/
 	       Couvertures(SFic, Tstruct, SModel, Best, Model, NbStructures)      ;

       if (a || f || Appariement)
          st = ast ;
     
       if (f) 
       {

	 if (!AllStr) 
         {
	    fprintf(stderr,"Probleme de memoire dans Kmrs!\n") ;
	    return 0 ;
	 }
		/* ecriture resultats dans fichier de sauvegarde  	 */

	 if (
	     !SauvegardeStructures(
				   Distingue, DebPart, FinPart,
                                   Lmax, N, s, NbPal,
				   Pal, AllStr, Pos, Sol, M, 
				   SModel, Best, NbStructures
				  )
	    )
	     return 0 ;

	 if (AllStr)
	    FreeWordBunch(AllStr) ;
       }
      }

      if (st) 
      {
		 /* sauvegarde des taux de repetitions des structures    */
          SauvegardeTaux(Nfichierst, Fic, SFic, Tstruct) ;
      }
     
      if (Appariement)
	    if (!FAppariements(
	                       MAX_V, dot_seq, Alignees, Distingue, DebPart, FinPart,
			       NFAppariement, Lmax, N, s, NbPal, Pal,
			       AllStr, Pos, Sol, M,  SModel, Best, NbStructures
			      )
	       )
	       return 0 ;
    }
     else
	fprintf(stderr,"Probleme de memoire avant Kmrs!\n") ;
	
    if (a || f || Appariement)
    {
	free(Best)	;
	free(SModel)	;
        free(Model)	;
    }	
    
    if (lt && VTete) 
    {
	for (i = 0 ; i < N ; i++)
	    if (VTete[i])
		free(VTete[i]) ;

	free(VTete) ;
    }
  
    if (lb && VBoucle) 
    {
	for (i = 0 ; i < N ; i++)
	     if (VBoucle[i])
		free(VBoucle[i]) ;

	free(VBoucle) ;
    }

    if (td1 && d1)
	FreeDistances(d1,N,t) ;

    if (td2 && d2)
	FreeDistances(d2,N,t) ;

    if (td3 && d3)
	FreeDistances(d3,N,t) ;

    fprintf(stderr,"\nFin du programme.\n\n") ;

    return 1 ;
}  
