/* ==================================================== */
/*      Copyright (c) Atelier de BioInformatique        */
/*      Mar. 92                                         */
/*      File: libkmrs8.c                                */ 
/*      Purpose: KMR pour structures repetees           */
/*      History  :                                      */
/*      00/03/92 : <Gloup> first draft                  */
/*      15/08/93 : <Gloup> revised version              */
/*      12/02/95 : adapation aux structures             */
/*                 (modifications importantes)          */
/*      23/03/95 : version qui evite tests piles vides  */
/*      Juillet 95 :                                    */
/*                 Apprentissage des attributs          */
/*                 version avec sauvegarde des taux     */
/*                 de repetition des structures         */
/*                 et descriptions purete,longueurs	*/
/*	           moyenne et ecart type		*/
/*                 notion de distance entre pals	*/
/*		   on ne conserve qu'une classe         */
/*		   apres calcul des valeurs d'un goupe  */
/*		   d'attribut				*/
/* ---------------------------------------------------- */
/* les notations utilisees pour les variables sont      */
/* celles de l'article de Karp-Miller-Rosenberg         */
/* ==================================================== */

#ifndef THINK_C
#include <sys/types.h>
#endif

#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <math.h>


#include "../../include/kmrstr8.h"
#include "../../include/gtypes.h"
#include "../../include/pals6.h"
#include "../../include/libsysks.h"
#include "../../include/libstki.h"
#include "../../include/libstkp.h"
#include "../../include/libwbs.h"
#include "../../include/libkmrs8.h"



#define POP             PopiOut
#define PUSH            PushiIn
#define SWAPCURS        CursiSwap
#define TOPCURS         CursiToTop
#define DOWNREAD        ReadiDown

#define PUSHP           PushpIn


#define HUGE_POWER      0x10000000      /* a huge power of 2    */


/* ==================================== */
/*  VARIABLES GLOBALES AU MODULE        */
/* ==================================== */

int   Attributs    = 0    ;     /* attributs utilises ou pas      */
int   NbAttributs  = 0    ;     /* nombre d'attributs au pas k+1  */
int   Attributk1   = 0    ;     /* attribut ou pas au pas k = 1   */
char *Attributs1   = NULL ;     /* tab codes des attributs (k)    */
char *Attributs2   = NULL ;     /* tab codes des attriburts (k+1) */
int  Indent        = 0    ;     /* indentation des messages       */
int  Verb          = 0    ;     /* densite des messages           */


/* ==================================== */
/*        TYPES                         */
/* ==================================== */

typedef struct {
		Int32 Nr  ;            /* nombre de repetitions d'une classe  */
		Int32 Pos ;            /* position d'un element temoin        */
		Long Ns   ;	       /* nombre de solution contenant	      */
                                       /* contenant la classe                 */


                float  m1 ;            /* nb moyen de sol par seq representee */
                float  m2 ;            /* nb moyen de sol par sequence	      */
                float  e1 ;            /* ecart type nb moyen de sol par      */
				       /* par sequence representee	      */
                float  e2 ;            /* nb moyen de sol par sequence	      */  	 
} TStats ;




/* ==================================== */
/*        MACROS                        */
/* ==================================== */





/* ==================================== */
/* Prototypes des fonctions locales     */
/* ==================================== */



static StackiHdle DualVa           P(( Int32 va[], Int32 Na, Int32 ea ));

static StackiHdle DualVaCheck      P(( TPosition * Pos, Int32 k, Int32 va[],
                                       Int32 Na, Int32 ea ));


static Bool       CheckConstraint  P(( int Dtrna,Palindrome** Pal, Int32*** Sol,
	                     	       Int32**** M,
                                       int check, Int32 k,
                                       int Ecart, Int32 N,Int32 Classe, int Verbose,
                                       TStats *Stats,
				       TPosition *Pos,Int32 Rmin, StackiPtr stk ));


void MajAttributs P((
     int k             ,
     int  CodeAttribut ,
     Long ClasseQ      ,
     Long ClasseP      )) ;

static Bool       LoadInVab        P(( int Dtrna,Palindrome** Pal, Int32*** Sol,
                                       Int32**** M, 
                                       int check, Int32 k,
                                       int Ecart, Int32 N, int Verbose,
                                       TStats *Stats,
				       TPosition *Pos,Int32 Rmin,
				       Int32 vab[], StackiHdle ra,
                                       Int32 class ));


static Int32      BuiltVab         P(( int Dtrna,Palindrome** Pal, Int32*** Sol,
                                       Int32**** M,
                                       int check, Int32 k,
                                       int Ecart,Int32 N,int Verbose,
                                       TStats *Stats,
				       TPosition *Pos,Int32 Rmin,StackiHdle q,
				       Int32 vab[],
				       StackiPtr PilesQ ));



/* -------------------------------------------- */
/* Empile les elts dans pile "q" et les numeros */
/* de piles dans "PilesQ"			*/
/* -------------------------------------------- */

int             Empile     P((
    Int32       ClasseQ ,
    int         TQProv[],
    StackiPtr   PilesQ  ,
    StackiHdle  q 	,
    Int32       classe	,
    Long        code    ,
    Int32       r
			   )) ;


static Int32 Couverture P((

       Int32      classe ,
       StackiHdle qProv  ,
       TPosition  Pos[]  ,
       Int32      Rmin   )) ;



static StackiHdle BuiltQ P((
	Int32  Rmin		,
        int    td1              ,    /* temoin calcul distance 1        */
	int    *** d1           ,    /* tableau distances 1             */
        int    td2              ,    /* temoin calcul distance 2        */
	int    *** d2           ,    /* tableau distances 2             */
        int    td3              ,    /* temoin calcul distance 3        */
	int    *** d3           ,    /* tableau distances 3             */
       	Int32   ***Sol			,
	int     purete			,
	Palindrome    **Pal		,
	int     lt			,
	Int32   ** VTete		,
	int     lb			,
	Int32   ** VBoucle		,
	Int32   **** M                  ,
	TPosition   *Pos                ,
	Int32       k                   ,
	StackiHdle  p                   ,
	Int32       ep                  ,
	Long       Nas			,
	Long       eq                  	,
	StackiPtr  *PilesQ


			  )) ;



static void ResetVa P((Int32 Va[], Int32 n)) ;

static Int32 ApplyLemma1 P((
        int    Dtrna            ,
        int    check            ,    /* temoin distribution trna        */
        int    td1              ,    /* temoin calcul distance 1        */
	int    *** d1           ,    /* tableau distances 1             */
        int    td2              ,    /* temoin calcul distance 2        */
	int    *** d2           ,    /* tableau distances 2             */
        int    td3              ,    /* temoin calcul distance 3        */
	int    *** d3           ,    /* tableau distances 3             */
	int      Ecart          ,    /* temoin calcul ecart type	*/
	int      lt		,    /* temoin tenir compte des tetes   */
	int    **VTete		,    /* tab des tetes des pals		*/
	int      lb		,    /* temoin tenir compte boucles	*/
	int    **VBoucle	,    /* tab des boucls des pals		*/
	int	purete		,    /* temoin tenir compte purete pal	*/
	int     Verbose         ,
	TStats  **Stats         ,
	Int32      Rmin         ,
	Int32   va[]            ,
	Int32          k        ,
	Int32         ea        ,
	Int32          n        ,    /* nb total pals           */
	Int32          N        ,    /* nb sequences            */
	Int32        * t        ,    /* nb solutions            */
	Int32       ** p        ,    /* nb Pals                 */
	TPosition    * Pos      ,    /* indices Seq,Sol,Pal     */
	Palindrome  ** Pal      ,    /* Palindromes             */
	Int32    *** Sol        ,    /* indices Pal dans Sol    */
	Int32     **** M        ,    /* Matrices                */
	Long      * Nex              /* nb classes theoriques   */
				     /* de taille k+1           */
	)) ;

static void       SwapVectors      P(( StackiHdle *v1, StackiHdle *v2 ));
static StackiHdle AbortKmrSimple   P(( Int32 va[], char *msg, Int32 wlen,
					Int32 n));
static WordBunchPtr AbortKmrAll    P(( Int32 va[], WordBunchPtr w,
				       char *msg, Int32 wlen, Int32 n ));
static void      AfficherStats     P((

        int    td1              ,    /* temoin calcul distance 1        */
	int    *** d1           ,    /* tableau distances 1             */
        int    td2              ,    /* temoin calcul distance 2        */
	int    *** d2           ,    /* tableau distances 2             */
        int    td3              ,    /* temoin calcul distance 3        */
	int    *** d3           ,    /* tableau distances 3             */

                        int Ecart	     ,
			int Moyenne	     ,
                        Int32   ***Sol       ,
			int	purete       ,
       			Palindrome     **Pal ,
      			int	lt	     ,
      			int      **VTete     ,
      			int      lb	     ,
      			int      **VBoucle   ,
			int   st             ,
			Int32 Fic[]          ,
			Int32 SFic[]         ,
			Int32 Tstruct[]      ,
			FILE  *stream        ,
			Int32     eab        ,
			TStats    *Stats     ,
			Int32     n          ,
			Int32     k          ,
			TPosition * Pos      ,
			Int32     **** M
				    )) ;

static int       Sup P(( Int32 a , Int32 b)) ;


/* ==================================================== */
/* Affiche contenu d'un message                         */
/* ==================================================== */
void Affiche P((

     char *Message  ,
     int  Increment )) ;

/* ==================================================== */
/* statiques locales                                    */
/* ==================================================== */

static Int32    sRepet = 2;            /* # d'occurence */





/* ==================================================== */
/* Affiche contenu d'un message                         */
/* ==================================================== */
void Affiche(Message,Increment)

     char *Message  ;
     int  Increment ;
{
     int i ;

     if (Verb > 2) {

	Indent += Increment ;

	if (Message) {

	   for (i = 0 ; i < Indent ; i++)
		fprintf(stderr," ") ;

	   fprintf(stderr,"%s\n",Message) ;
	}

     }
}




/* ==================================================== */
/* Etapes KMR a -> a+b quelconques                      */
/* ==================================================== */

/* -------------------------------------------- */
/* -- KMR etape 1 ---                           */
/* -------------------------------------------- */

/* -------------------------------------------- */
/* calcul du dual de V                          */
/* version Check => on n'empile une position    */
/* que si elle represente une structure         */
/* extensible d'un palindrome a gauche          */
/* donc commencant au moins a  k + 1            */
/* positions du debut d'une solutions           */
/*                                              */
/* @static: DualVa                              */
/* -------------------------------------------- */




static StackiHdle DualVaCheck(
	TPosition *Pos  ,
        Int32 k         ,
	Int32  va[]     ,
	Int32  Na       ,
	Int32  ea            )
{
	Int32           i   ;
	StackiHdle      stk ;
	Int32           c   ;

	Affiche("DualVaCheck",3) ;

	if (! (stk = NewStackiVector(ea, kMinStackiSize)))
	    return NULL;


	for (i = 0 ; i < Na ; i++) 
        {
	    if (va[i] >= 0 && Pos[i].Pal >= k) 
            {
					/* structure extensible		*/
		     if (!PUSH(stk + va[i], i))
		         return NULL ;
            }
	}

        Affiche("Fin dual va check",0) ;
	Affiche("",-3) ;
	return stk;

} /* end of DualVaCheck */




static StackiHdle DualVa(va, Na, ea)
	Int32 va[]    ;
	Int32 Na, ea  ;

{
	Int32      i   ;
	StackiHdle stk ;
	Int32      c   ;

	if (! (stk = NewStackiVector(ea, kMinStackiSize)))
	    return NULL;


	for (i = 0 ; i < Na ; i++) 
        {

					/* structure extensible		*/
		 if (va[i] >= 0)
		     if (!PUSH(stk + va[i], i))
		         return NULL ;
      	}

	return stk;





} /* end of DualVa */





/* -------------------------------------------- */
/* Empile les elts dans pile "q" et les numeros */
/* de piles dans "PilesQ"			*/
/* -------------------------------------------- */

int Empile(ClasseQ,TQProv,PilesQProv, qProv, classe,code,r)
    Int32       ClasseQ  ;
    int         TQProv[] ;
    StackiPtr   PilesQProv   ;
    StackiHdle  qProv 	 ;
    Int32       classe	 ;
    Long        code     ;
    Int32       r        ;
{

    if (!TQProv[ClasseQ]) {

	TQProv[ClasseQ] = 1 ;
	if (!PUSH(&PilesQProv,ClasseQ))
	    return NULL	     ;
    }

    return(PUSH(qProv+classe+code,r)) ;

}



/* -------------------------------------------- */
/* renvoie couverture d'une classe              */
/* -------------------------------------------- */

static Int32 Couverture(classe, qProv, Pos, Rmin)

       Int32      classe ;
       StackiHdle qProv  ;
       TPosition  Pos[]  ;
       Int32      Rmin   ;


{
       Int32  PrevSeq = -1  ;
       Int32  Couvert =  0  ;
       Int32  val	    ;
       Int32  i		    ;

       CursiToTop(qProv[classe]) ;

       while (ReadiDown(qProv[classe], &val)) 
       {
	     i = Pos[val].Seq ;

	     if (i != PrevSeq) 
	     {
		Couvert++   ;
		PrevSeq = i ;
	     }
       }

       if (Couvert >= Rmin)
	   return Couvert ;
       else
	   return 0 ;
}



/* -------------------------------------------- */
/* -- KMR etape 2 ---                           */
/* -------------------------------------------- */

/* -------------------------------------------- */
/* creation de la pile q et ventilation de la   */
/* pile p vers la pile q suivant les classes    */
/* indiques dans v                              */
/* -------------------------------------------- */
/* Note : p est depile                          */
/* -------------------------------------------- */
/* Note : pour le Lemme1 :                      */
/*      v = va, ep = eq = ea , shift = b        */
/*        pour le Lemme2 :                      */
/*      v = vb, ep = ea, eq = eb , shift = a    */
/*                                              */
/* @static: BuiltQ                              */
/* -------------------------------------------- */

static StackiHdle BuiltQ(Rmin,td1, d1, td2, d2, td3, d3,
                         Sol,purete,Pal,lt,VTete,lb,VBoucle,
			 M,Pos,k,p,ep, Nas,eq, PilesQ)

	Int32  Rmin		;    /* taux minimal de couverture	*/
        int    td1              ;    /* temoin calcul distance 1        */
	int    *** d1           ;    /* tableau distances 1             */
        int    td2              ;    /* temoin calcul distance 2        */
	int    *** d2           ;    /* tableau distances 2             */
        int    td3              ;    /* temoin calcul distance 3        */
	int    *** d3           ;    /* tableau distances 3             */
	Int32   ***Sol			;
	int     purete			;
	Palindrome    **Pal	        ;
	int     lt			;
	Int32   ** VTete		;
	int     lb			;
	Int32   ** VBoucle		;
	Int32   **** M                  ;
	TPosition   *Pos                ;
	Int32       k                   ;
	StackiHdle  p                   ;
	Int32       ep                  ;
	Long       Nas			;
	Long       eq                  	;
	StackiPtr  *PilesQ              ;

{
	Int32    ClasseQ          ;
	Long     QBest    = -1    ;
	Int32    Taux     = 0	  ;
	Int32    BestTaux = 0     ;
        Int32     r		  ;
	Long      c		  ;
	Int32     classe          ;
	Int32      i,j,m          ;
	StackiHdle q              ;
	StackiHdle qProv	  ;
        Int32     s               ;
	Long f[NBATTRIBUTS-1]     ;
        Long shift = 0            ;
	Int32 pq                  ;  /* numero d'une pile Q */
        Long attribut             ;
	int  CodeAttribut         ;  /*  numero de code associe a l'attribut */
	int  a                    ;
	StackiPtr PilesQProv      ;
	Int32 code		  ;
	int   *TQprov             ; /*   tab temoin des classes Q non vides  */


        NbAttributs = 0 ;

	Affiche("BuiltQ",3) ;

	if (Attributs) {


	   if (purete) {
	       shift += NBVPURETE ;
	       NbAttributs++ ;
	   }

	   f[0] = shift ;

	   if (lt) {
	       shift += NBVTETE ;
	       NbAttributs++ ;
	   }

	   f[1] = shift ;


	   if (lb) {
		shift +=NBVBOUCLE ;
		NbAttributs++ ;
	   }

	   f[2] = shift ;

	   if (k > 0) 
           {
	      if (td1) 
              {
		 shift += NBVD1 ;
		 NbAttributs++  ;
	      }

	      f[3] = shift ;


	      if (td2) 
              {
		 shift += NBVD2 ;
		 NbAttributs++  ;
	      }

	      f[4] = shift ;

	      if (td3) 
              {
		  shift += NBVD3 ;
		  NbAttributs++ ;
	      }
	   }
	   else {/* k == 0 */
	      f[3] = f[2] ;
	      f[4] = f[3] ;
	   }

           if (k == 0) 
           {
               if (NbAttributs == 0)
                  Attributk1 = 0 ;
               else
                  Attributk1 = 1 ;
	   }

	}

	if (NbAttributs) 
        {

	   qProv = NewStackiVector(Nas,kMinStackiSize) ;
	   if (!qProv)   
	      return NULL ;

	   PilesQProv = NewStacki(kMinStackiSize) ;
	   if (!PilesQProv)
	      return NULL ;


	   TQprov = (int *) malloc( (2*k + 1) * sizeof(int)) ;

	}


	q = NULL ;

	*PilesQ = NewStacki(kMinStackiSize) ;
	if (! *PilesQ) return NULL ;

	q = NewStackiVector(eq, kMinStackiSize);
	if (! q) return NULL;


	for (c = 0 ; c < ep ; c++) 
       {
            if (NbAttributs) 
            {
         
             /* intialisation du tableau  TQProv                    */

                for (i = 0 ; i <= 2*k ; i++) 
                    TQprov[i] = 0 ;

            }

	    while (POP(p + c, &r)) {    /* poping out p         */

		  i = Pos[r].Seq ;
		  j = Pos[r].Sol ;
		  m = Pos[r].Pal ;

                  s = Sol[i][j][m-k] ;    /* indice pal		*/

		  classe = M[i][j][m][m-k] ;

					  /* precise groupe attributs	*/
					  /* selon code structure	*/

		     if (NbAttributs) 
                     {
			ClasseQ = classe ;
			classe *= shift  ;

			if (purete)
			    if (!Empile(ClasseQ,TQprov,PilesQProv,qProv,
					classe,Pal[i][s].Code,r))
				 return NULL ;


			if (lt)
			    if (!Empile(ClasseQ,TQprov,PilesQProv,qProv,
					classe,f[0] + VTete[i][s],r))
				 return NULL ;


			if (lb)
			    if (!Empile(ClasseQ,TQprov,PilesQProv,qProv,
					classe,f[1] + VBoucle[i][s],r))
				 return NULL ;

			if ( k > 0 ) {

			    if (td1)
			       if (!Empile(ClasseQ,TQprov,PilesQProv,qProv,
					   classe, f[2] + d1[i][j][m-k],r))
				   return NULL ;

			    if (td2)
			       if (!Empile(ClasseQ,TQprov,PilesQProv,qProv,
					   classe, f[3] + d2[i][j][m-k],r))
				   return NULL ;

			    if (td3)
			       if (!Empile(ClasseQ,TQprov,PilesQProv,qProv,
					   classe, f[4] + d3[i][j][m-k],r))
				return NULL ;
			}

		     }
		     else {	/* pas d'attributs */
			 if (!(q[classe]->top))
			      if (!PUSH(PilesQ,classe))
				 return NULL ;
			 if (!PUSH(q+classe,r))
				 return NULL ;
		     }

	    }      			/* fin while		*/
					/* met separateur -1 dans les piles Q     */
					/* possedant elts pile P courante         */

	    /* selection de la classe ayant la plus forte couverture	*/

	    if (NbAttributs) {

	       QBest    = - 1 ;
	       BestTaux =   0 ;

	       CursiToTop(PilesQProv) ;

	       while (ReadiDown(PilesQProv,&ClasseQ)) 
	       {
		      for (code = 0 ; code < shift ; code++) 
		      {

			  Taux = Couverture(ClasseQ*shift + code,qProv,Pos,Rmin) ;

			  if (Taux > BestTaux) 
			  {

			     QBest    = code ;
			     BestTaux = Taux ;
			  }

		      } /* fin de for code */


		      if (QBest != -1) 
		      {   /* il y a une meilleur classe */

			 /* memorise pile Q + classe non vide		 */

			 if (!(q[ClasseQ]->top))
			      if(!PUSH(PilesQ,ClasseQ))
				 return NULL ;

			  while (POP(qProv+ClasseQ*shift+QBest,&r)) {
				   /* transfert des elts vers pile Q	*/

				if (!PUSH(q+ClasseQ,r))
				   return NULL ;

			  } /* fin de while POP */

			  /* memorise provenance P	*/
			  if (!PUSH(q+ClasseQ,c))
			      return NULL ;


			  /* recherche code attribut	*/

			  CodeAttribut = 0 ;

			  for (a = 0 ; a < NBATTRIBUTS-1 ; a++)
			       if(QBest >= f[a])
				  CodeAttribut = a + 1 ;

			  if (!PUSH(q+ClasseQ,CodeAttribut))
			      return NULL ;

			  if (!PUSH(q+ClasseQ,-1))
			      return NULL ;

		      }  /* fin de QBest != -1 */

	       } /* fin de while ReadiDown */

				/* remise a zero des piles QProv	*/
		while (POP(&PilesQProv,&ClasseQ))

		      for (code = 0 ; code < shift ; code ++)
			   qProv[ClasseQ*shift + code]->top = 0 ;

	    }  /* fin de if NbAttributs	*/

	    else { /* pas d'attributs */

		 CursiToTop(*PilesQ) ;

		 while(ReadiDown(*PilesQ,&classe))
		       if (!PUSH(q+classe,-1))
			  return NULL ;

	    }  /* fin else */

	} 				/* fin for */


	if (NbAttributs) 
        {
	    FreeStackiVector(qProv,Nas) ;
	    FreeStacki(PilesQProv) ;
	    free(TQprov) ;
	}

	Affiche("Fin BuiltQ",0) ;
	Affiche("",-3)	        ;
	return q ;
}

/* -------------------------------------------- */
/* -- KMR etape 3 ---                           */
/* -------------------------------------------- */

/* -------------------------------------------- */
/* teste si une pile d'index verifie une        */
/* certaine contrainte. Ici la contrainte est   */
/* juste qu'elle contienne au moins Rmin        */
/* elements repartis dans au moins Rmin         */
/* sequences                                    */
/*                                              */
/* @static: CheckConstraint                     */
/* -------------------------------------------- */

static Bool CheckConstraint(Dtrna,Pal,Sol,M,check,k,Ecart,N,Classe,Verbose,Stats,Pos,Rmin,stk)
        int Dtrna           ;
        Palindrome** Pal    ;
        Int32***     Sol    ;
        Int32****    M      ;
        int         check   ;
        Int32      k        ;
        int    Ecart    ;
	Int32  N	;
	Int32  Classe   ;
	int  Verbose    ;
	TStats   *Stats ;
	TPosition *Pos  ;
	Int32    Rmin   ;
	StackiPtr stk   ;

{
         Int32  i             ;
	 Int32  s = 0         ;
         Int32  sbis = 0      ;
	 Int32  val           ;
	 Int32  strnabis = 0  ;
	 Int32  PrevSeq     = -1  ;
	 Int32  PrevSol     = -1  ;
	 Int32  PrevSeqTrna = -1  ;
	 Int32  PrevSolTrna = -1  ;
	 Long   Ns = 0        ;
         Long   se1,se2       ;
         Int32* NsSeq= NULL   ;
         Int32* NsSeqTrna = NULL ;
         int    trna          ;
         Int32  i1,j1,p1      ;
         int    premier = 1   ;
         Palindrome pal1      ;
	 Int32 strna  = 0     ;
	 Int32 NsTrna = 0     ;
         int   HorsTrna       ;
         int   u              ;


	    Affiche("Check constraint",3) ;
             
	    CursiToTop(stk) ;

            if (k != 3)
                trna = 0   ;
            else               /* construit longueur 4 */
                trna = 1   ;

            if (Ecart) {

                NsSeq = (Int32 *) malloc(N * sizeof(Int32)) ;

                if (!NsSeq) {
                    fprintf(stderr,"Erreur allocation NsSeq !!!\n") ;
                    return Faux ;
                }

                if (trna) 
		{

                   NsSeqTrna = (Int32 *) malloc(N * sizeof(Int32)) ;

                   if (!NsSeqTrna) 
		   {
                       fprintf(stderr,"Erreur allocation NsSeqTrna !!!\n") ;
                       return Faux ;
                   }


                }

                for (i = 0 ; i < N ; i++)
                     NsSeq[i] = 0 ;

                if (trna) {

                     for (i = 0 ; i < N ; i++)
                          NsSeqTrna[i] = 0 ;

                }


						   /* initialisation tab ecart	   */
            }

	    while (ReadiDown(stk,&val)) 
	    {          /* tq il existe elt dans pile   */

                  if (check && trna) {             /* verif distribution trna      */

                     i1 = Pos[val].Seq ;
                     j1 = Pos[val].Sol ;
                     p1 = Pos[val].Pal ;

                     if (premier) {

                         trna = (M[i1][j1][p1][p1-3] == 6)  *
                                (M[i1][j1][p1][p1-2] == 0)  *
                                (M[i1][j1][p1][p1-1] == 0)  *
                                (M[i1][j1][p1][p1]   == 0)    ;

                         premier = 0 ;
                     }

		     pal1 = Pal[i1][Sol[i1][j1][p1-3]] ;

		     HorsTrna = (pal1.DTete < Dtrna + 1 ) || (pal1.FQueue >= Dtrna + 80) ;

                  }

		  if (Pos[val].Seq != PrevSeq) {   /* si appartient a autre seq    */
		      s++ ;                        /* incrementer s                */

		      PrevSeq = Pos[val].Seq ;
		      PrevSol = - 1          ;
		  }


		  if (check && trna && !HorsTrna &&
		      Pos[val].Seq != PrevSeqTrna) {   /* si appartient a autre seq    */
		      strna++ ;                        /* incrementer s                */

		      PrevSeqTrna = Pos[val].Seq ;
		      PrevSolTrna = -1           ;
		  }



                  if (Pos[val].Sol != PrevSol) { /* si appartient a autre sol    */
		        Ns++ ;                   /* incrementer Ns               */

                        if (Ecart)
                           NsSeq[Pos[val].Seq]++ ;

		        PrevSol = Pos[val].Sol ;
		  }

		  if ( check && trna && !HorsTrna &&
		       Pos[val].Sol != PrevSolTrna) { /* si appartient a autre sol    */
			NsTrna++ ;                /* incrementer Ns               */

			NsSeqTrna[Pos[val].Seq]++ ;

			PrevSolTrna = Pos[val].Sol ;
		  }


            }


	    if (s >= Rmin) {

               if (check && trna) {


                                              /* affichage distribution 6000 */
                       fprintf(stdout,"\n\n") ;
                       fprintf(stdout,"Longueur                           : %4d\n",k+1) ;
                       fprintf(stdout,"Nb bases avant et apres trna       : %4d\n", Dtrna) ;
                       fprintf(stdout,"Structure                          :  ") ;
                       for (u = p1 - k  ; u <= p1 ; u++) {
	                   fprintf(stdout," %2d", M[i1][j1][p1][u]) ;

                       }

                       fprintf(stdout,"\n") ;
		       fprintf(stdout,"Numero de classe                   : %4d \n",Classe) ;
		       fprintf(stdout,"sequences representees             : %4d \n",s)      ;
		       fprintf(stdout,"sequences trna vraies representees : %4d \n",strna)  ;
		       fprintf(stdout,"nb solutions                       : %4d \n",Ns)  ;
		       fprintf(stdout,"nb solutions trna vraies           : %4d \n",NsTrna)  ;


		       for (i = 0 ; i < N ; i++) {

                            fprintf(stdout,"%4d : %4d, %4d  ",
                                    i,NsSeq[i], NsSeqTrna[i] ) ;

                            if (NsSeq[i] > 0) {
                                fprintf(stdout,"*") ;


                            if (NsSeq[i] > 0)
				sbis++ ;


                            if (NsSeqTrna[i] > 0)
				strnabis++ ;

                                if (NsSeq[i] != NsSeqTrna[i])
                                     fprintf(stdout," !") ;



                            }

                            fprintf(stdout,"\n") ;
                        }


                        fprintf(stdout,"sequences representees             : %4d\n\n",sbis) ;
			fprintf(stdout,"sequences reellement representees  : %4d\n\n",strnabis) ;


                        fprintf(stdout,"\n") ;



             }


	       if (Verbose >= 2) {
		  Stats[Classe].Nr   = s    ; /* sauve indice de classe      */
		  Stats[Classe]. Pos = val  ; /* sauve position temoin d'une */
					     /* structure                    */
                  Stats[Classe].Ns = Ns    ;
                  Stats[Classe].m1 = 1.0 * Ns / s ;
                  Stats[Classe].m2 = 1.0 * Ns / N ;


                  if (Ecart) {

                     			    /* calcul de l'ecart type	   */

                      se1 = 0 ;
                      se2 = 0 ;

              

                      for (i = 0 ; i < N ; i++) {

                           

                           se2 +=  (NsSeq[i] - Stats[Classe].m2) *
                                   (NsSeq[i] - Stats[Classe].m2) ;

                           if (NsSeq[i])
                               se1 +=  (NsSeq[i] - Stats[Classe].m1) * 
                                       (NsSeq[i] - Stats[Classe].m1) ;
                      }

                       Stats[Classe].e1 = sqrt(1.0 * se1 / s) ;
                       Stats[Classe].e2 = sqrt(1.0 * se2 / N) ;



                      free(NsSeq) ; 

                      if (k == 3 && NsSeqTrna)
                          free(NsSeqTrna) ;

                  }
               }
               
	       Affiche("Fin check constrainst",0) ;
	       Affiche("",-3) ;
	       return Vrai ;
	    }
	    else {

                if (Ecart && NsSeq)
                    free(NsSeq) ;

                if (k == 3 && NsSeqTrna)
                    free(NsSeqTrna) ;

	       Affiche("Fin check constraint",0) ;
	       Affiche("",-3) ;
	       return Faux ;
            }

} /* end of CheckConstraint */



/* -------------------------------------------- */
/* stockage conditionnel d'une classe dans Vab  */
/* a partir d'une pile r                        */
/* On stocke la nouvelle classe si              */
/* Ra verifie la contrainte 'CheckConstraint'   */
/* et on renvoie un booleen indiquant si le     */
/* stockage a ete effectue                      */
/*                                              */
/* @static: LoadInVab                           */
/* -------------------------------------------- */

static Bool LoadInVab(Dtrna,Pal,Sol,M,check,k, Ecart,N,Verbose,Stats,Pos,Rmin,vab, r, class)
        int    Dtrna        ;
        Palindrome** Pal    ;
        Int32***     Sol    ;
        Int32****    M      ;
        int         check   ;
        Int32      k        ;
        int    Ecart	    ;
	Int32  N	    ;
	int    Verbose      ;
	TStats *Stats       ;
	TPosition *Pos      ;
	Int32     Rmin      ;
	Int32  vab[]        ;
	StackiHdle  r       ;
	Int32       class   ;
        

{
	Int32   pos;
	Bool     okConstraint ;
       

	Affiche("Load in Vab",3) ;

	okConstraint = CheckConstraint(Dtrna,Pal,Sol, M, check, k,
                                       Ecart,N,class,Verbose,Stats,Pos,Rmin,*r);

	if (okConstraint) 
        {
	    while(POP(r, &pos))
            {
		  vab[pos] = class     ;                   
            }
	}

	else
	   (*r)->top = 0;

	Affiche("Fin Load In Vab",0) ;
	Affiche("",-3) ;

	return okConstraint;

} /* end of LoadInVab */




/* met a jour le ligne du tableau des attributs correspondant a ClasseQ */
void MajAttributs(k,CodeAttribut,ClasseQ,ClasseP)
     int k             ;
     int  CodeAttribut ;
     Long ClasseQ      ;
     Long ClasseP      ;

{
       Long ind        ;
       Long Depart1    ;
       Long Depart2    ;

       Affiche("Maj Attributs",3) ;

       Depart1 = ClasseP * k ;
       Depart2 = ClasseQ * (k+1) ;

       Attributs2[Depart2] = CodeAttribut ;

       for (ind = 1 ; ind <= k ; ind++)
           if (ind < k || Attributk1 > 0)
	      Attributs2[Depart2 + ind] = Attributs1[Depart1 + ind - 1] ;

       Affiche("Fin Maj Attributs",0) ;
       Affiche("",-3) ;
}




/* -------------------------------------------- */
/* depilage de q et stockage dans vab           */
/* -------------------------------------------- */
/* Note : q est depile                          */
/* -------------------------------------------- */
/* Note : pour le Lemme1 : v = va; eq = ea      */
/*        pour le Lemme2 : v = vb; eq = eb      */
/*                                              */
/* @static: BuiltVab                            */
/* -------------------------------------------- */

static Int32 BuiltVab(Dtrna,Pal,Sol, M, check, k,
		      Ecart,N,Verbose,Stats,Pos,Rmin,q, vab,PilesQ)
        int Dtrna           ;
        Palindrome** Pal    ;
        Int32***     Sol    ;
        Int32****    M      ;
        int         check   ;
        Int32      k        ;
         int    Ecart			;
	Int32  N			;
	int    Verbose                  ;
	TStats  *Stats                  ;
	TPosition *Pos                  ;
	Int32    Rmin                   ;
	StackiHdle  q                   ;
	Int32       vab[]      		;
	StackiPtr  PilesQ               ;




{
	Int32      i, pos, eab, class, prev;
	StackiPtr  r 			;
	Int32 ClasseP1 = -1  		;
	Int32 ClasseP2 = -1		;
	int  CodeAttribut               ;

	Affiche("Built Vab",3) ;

	r = NewStacki(kMinStackiSize);          /* stack temporaire     */
	if (! r) return -1;


	eab = 0;                                /* compteur de classe   */

	 while (POP(&PilesQ, &i)){

	    while (POP(q + i, &pos))    {       /* poping out q         */



                if (pos == -1) 
		{
						 /* nouvelle classe      */
		   if (NbAttributs) 
		   {
		      POP(q+i,&CodeAttribut) ;   /* Attribut associe a la pile */

		      POP(q+i,&ClasseP2) ;       /* provenance P         */
		   }

		   if (r->top && LoadInVab(Dtrna,Pal,Sol,M,check,k,
				 Ecart,N,Verbose,Stats,Pos,Rmin,vab, &r, eab)) 
		   {
						/* charge ds vab, reset  */
		       if (NbAttributs)
			  MajAttributs(k,CodeAttribut,eab,ClasseP1) ;

		       eab++;                   /* r et incremente eab  */

		    }

		   ClasseP1 = ClasseP2 ;
		}

                else

		 if(!PUSH(&r, pos))
		     return NULL ;                        /* stocke pos dans r    */
	    }

				     /* dernier groupe ?     */
	       if (r->top && LoadInVab(Dtrna,Pal,Sol,M,check,k,
			     Ecart,N,Verbose,Stats,Pos,Rmin,vab, &r, eab)) 
	      {
		  /* charge ds vab, reset */

		  if (NbAttributs)
		     MajAttributs(k,CodeAttribut,eab,ClasseP1) ;

		     eab++;                        /* r et incremente eab  */                 
	       }

	}

FreeStacki(r);                          /* liberation memoire   */

	Affiche("Fin BuiltVab",0) ;
	Affiche("",-3) ;

	return eab;                             /* # groupes trouves    */

} /* end of BuiltVab */



/* ---------------------------------------------------- */
/* superiorite d'une valeur sur une autre     		*/
/* ---------------------------------------------------- */
static int Sup(a , b)
	    Int32 a     ;
	    Int32 b     ;
{
  return a > b ;
}




#define SUPT(a,b) Sup(Pal[i][Sol[i][j][a]].DQueue,Pal[i][Sol[i][j][b]].DTete)
#define SUPQ(a,b) Sup(Pal[i][Sol[i][j][a]].DQueue,Pal[i][Sol[i][j][b]].DQueue)



/* ---------------------------------------------------- */
/* extension a la taille k + 1 des matrices decrivant   */
/* les structures                                       */
/* ---------------------------------------------------- */
static void MAJMatrice(N,t,p,M,Pal,Sol,k)
       Int32      N      ;    /* nb sequences           */
       Int32    * t      ;    /* nb solutions           */
       Int32   ** p      ;    /* nb Pals                */
       Int32 **** M      ;    /* Matrices               */
       Palindrome ** Pal ;    /* Palindromes            */
       Int32  *** Sol    ;    /* indices Pal dans Sol   */
       Int32      k      ;    /* taille structures      */


{
 int i ;
 int j ;
 int m ;

 Affiche ("Mise a jour Matrice",3) ;

 for (i = 0 ; i < N ; i++)
     /* pour chaque sequence                    */

     for (j = 0 ; j < t[i] ; j++)
	 /* pour chaque solution                */

	 for (m = k ; m < p[i][j] ; m++)
	     /* pour chaque pal d'une solu      */

	     M[i][j][m][m-k] = M[i][j][m-1][m-k] +
			       SUPT(m-k,m) + SUPQ(m-k,m) ;

 Affiche("Fin Mise a jour Matrice",0) ;
 Affiche("",-3) ;
}




/* effectue Reset du tableau de Piles Va */

static void ResetVa(Va, n)
       Int32 Va[]    ;
       Int32      n  ;
{
       Int32 i ;

       Affiche("Reset Va",3) ;

       for (i = 0 ; i < n ; i++)
	    Va[i] = -1 ;

       Affiche("Fin Reset Va",0) ;
       Affiche("",-3) ;
}


/* ==================================================== */
/* LEMMES KMR                                           */
/* ==================================================== */

/* -------------------------------------------- */
/* LEMME 1                                      */
/* -------------------------------------------- */
/* note pour les structures b = 1               */
/* sauf initialement ou b = 0                   */
/*                                              */
/* construction de Va+b a partir de Va          */
/* x Eab y <=> x Ea y and x+b Ea y+b            */
/* a >= b >= 0                                  */
/* returns eab;  (-1 means memory failure)      */
/* -------------------------------------------- */
/* note :                                       */
/* le cas b = 0 sert a nettoyer V1 au boot.     */
/*                                              */
/* @static: ApplyLemma1                         */
/* -------------------------------------------- */

static Int32 ApplyLemma1(Dtrna,check,td1, d1, td2, d2, td3, d3,
                         Ecart,lt,VTete,lb,VBoucle,purete,
                         Verbose,Stats,Rmin,va,k,ea,n,N,t,p,Pos,Pal,
			 Sol,M,Nex)
        int    Dtrna            ;
        int    check            ;    /* distribution structure 6000     */
        int    td1              ;    /* temoin calcul distance 1        */
	int    *** d1           ;    /* tableau distances 1             */
        int    td2              ;    /* temoin calcul distance 2        */
	int    *** d2           ;    /* tableau distances 2             */
        int    td3              ;    /* temoin calcul distance 3        */
	int    *** d3           ;    /* tableau distances 3             */
        int      Ecart		;    /* temoin calcul ecart type	*/
	int      lt		;    /* temoin tenir compte des tetes   */
	int    **VTete		;    /* tab des tetes des pals		*/
	int      lb		;    /* temoin tenir compte boucles	*/
	int    **VBoucle	;    /* tab des boucls des pals		*/
	int	purete		;    /* temoin tenir compte purete pal	*/
	int     Verbose         ;
	TStats    ** Stats      ;
	Int32      Rmin         ;
	Int32   va[]            ;
	Int32          k        ;
	Int32         ea        ;
	Int32          n        ;    /* nb total pals           */
	Int32          N        ;    /* nb sequences            */
	Int32        * t        ;    /* nb solutions            */
	Int32       ** p        ;    /* nb Pals                 */
	TPosition    * Pos      ;    /* indices Seq,Sol,Pal     */
	Palindrome      ** Pal  ;    /* Palindromes             */
	Int32    *** Sol        ;    /* indices Pal dans Sol    */
	Int32     **** M        ;    /* Matrices                */
	Long      * Nex         ;    /* nombre de structures    */
				     /* possibles de taille k+1 */

{
	Int32           eab      ;
	StackiHdle      pa, qa   ;
	Long 		Nas,Nas1 ;
	StackiPtr       PilesQ   ;

					/* ---------------------------- */
					/*     E t a p e  1             */
					/* construction de pa           */
					/* Dual de va                   */
					/* ---------------------------- */
         
	Affiche("Apply Lemma 1",3) ;

	if (k > 0)
	   MAJMatrice(N,t,p,M,Pal,Sol,k) ;

	pa = DualVaCheck(Pos,k,va, n, ea);

	if (! pa)
	    return -1;

					/* ---------------------------- */
					/*     E t a p e  2             */
					/* ventilation de pa sur qa     */
					/* ---------------------------- */



	 Nas = 0 ;

			/* nombre d'extensions gauches possibles en     */
		        /* ajoutant un pal a une structure               */

         if (purete)
             Nas += NBVPURETE ;
			/* nombre d'extension gauche tenant aussi 	*/
			/* compte de la purete des pals			*/


	 if (lt)
	     Nas  += NBVTETE  ;
			/* nombre d'extension gauche tenant aussi 	*/
			/* compte longueur tete des pals		*/


	 if (lb)
             Nas  += NBVBOUCLE ;

			/* nombre d'extension gauche tenant aussi 	*/	
			/* compte longueur boucle des pals		*/

         if ( k > 0) 
	 {

              if (td1)
                  Nas += NBVD1 ;

			/* tient compte nb valeurs distance 1		*/

 	      if (td2)
                  Nas += NBVD2 ;

			/* tient compte nb valeurs distance 2		*/
   	      if (td3)
                  Nas += NBVD3 ;

			/* tient compte nb valeurs distance 3		*/
         }



			/* nb classes theoriques de longueur k+1	*/

	 Nas *= 2*k+1 ;

	 Nas1 = 2*k+1 ;


	*Nex *= Nas1 ;



	qa = BuiltQ(Rmin,td1, d1, td2, d2, td3, d3,
		    Sol,purete,Pal,lt,VTete,lb,VBoucle,
		    M,Pos,k,pa,ea,Nas,Nas1,&PilesQ) ;

		       /* il y a Nas * ea nouvelles classes possibles   */



	FreeStackiVector(pa, ea);

	if (NbAttributs)
	   Attributs2 = NEWN(char, Nas1 * ea * (k+1)) ;

	if (! qa) {
	    if (PilesQ)
		FreeStacki(PilesQ) ;
	    return -1;
	}

					/* ---------------------------- */
					/*     E t a p e  3             */
					/* depilage qa  et construction */
					/* de vab                       */
					/* ---------------------------- */


	if (Verbose >= 2) {
	   *Stats = NEWN(TStats,Nas1*ea) ;
	   if (!(*Stats))
	      return - 1 ;
	}
	else
	   *Stats = NULL ;


        ResetVa(va,n) ;

	eab = BuiltVab(Dtrna,Pal, Sol, M, check, k,
		       Ecart,N,Verbose,*Stats,Pos,Rmin,qa, va, PilesQ);

	FreeStackiVector(qa, Nas1);

	FreeStacki(PilesQ) ;

	Affiche("Fin Apply Lemma1",0) ;
	Affiche("",-3) ;

	return eab;

} /* end of ApplyLemma1 */




/* ==================================================== */
/* algorithme KMRS                                      */
/* ==================================================== */

/* -------------------------------------------- */
/* Kmr internal swap vector function            */
/*                                              */
/* @static: SwapVectors                         */
/* -------------------------------------------- */

static void SwapVectors(v1, v2)
	StackiHdle *v1 ;
	StackiHdle *v2 ;
{
	StackiHdle swp;

	swp = *v1;
	*v1 = *v2;
	*v2 = swp;

} /* end of SwapVectors */



/* -------------------------------------------- */
/* KmrS internal abort fonction                 */
/*                                              */
/* @static: AbortKmrSimple                      */
/* -------------------------------------------- */

static StackiHdle AbortKmrSimple(va, msg, wlen,n)
	Int32           va[]  ;
	char            *msg  ;
	Int32           wlen  ;
	Int32           n     ;
{
	if (va)  free(va);

	NotifyKmrError(msg, wlen);

	return NULL;

} /* end of AbortKmrSimple */




/* -------------------------------------------- */
/* Kmr internal abort fonction                  */
/*                                              */
/* @static: AbortKmrAll                         */
/* -------------------------------------------- */

static WordBunchPtr AbortKmrAll(va, w, msg, k,n)
	Int32           va[]    ;
	WordBunchPtr    w       ;
	char            *msg    ;
	Int32           k       ;
	Int32           n       ;
{
	if (va)  free(va);

	if (w)   (void) FreeWordBunch(w);

	NotifyKmrError(msg, k);

	return NULL;

} /* end of AbortKmrSimple */

 

/* -------------------------------------------- */
/* affiche les  statistiques                    */
/* concernant les structures repetees           */
/*                                              */
/* @function: AfficherStats                     */
/* -------------------------------------------- */


static void AfficherStats(td1, d1, td2, d2, td3, d3,
                          Ecart,Moyenne,Sol,purete,Pal,lt,VTete,lb,VBoucle,
                          st,Fic,SFic,Tstruct,stream,eab,Stats,N,k,Pos,M)

        int    td1              ;    /* temoin calcul distance 1        */
	int    *** d1           ;    /* tableau distances 1             */
        int    td2              ;    /* temoin calcul distance 2        */
	int    *** d2           ;    /* tableau distances 2             */
        int    td3              ;    /* temoin calcul distance 3        */
	int    *** d3           ;    /* tableau distances 3             */
       int      Ecart		;
       int      Moyenne		;
       Int32    ***Sol		;
       int	purete		;
       Palindrome     **Pal     ;
       int	lt		;
       int      **VTete		;
       int      lb		;
       int      **VBoucle	;
       int      st              ;
       Int32    Fic[]           ;
       Int32    SFic[]          ;
       Int32    Tstruct[]       ;
       FILE     *stream         ;
       Int32     eab            ;
       TStats    *Stats         ;
       Int32     N              ;
       Int32     k              ;
       TPosition    * Pos       ;    /* indices Seq,Sol,Pal             */
       Int32      **** M        ;    /* Matrices                        */

       

{
   Long  Depart      ;
   float Per         ;
   Int32   i         ;
   Int32   j         ;
   Int32   l         ;
   Int32   m         ;
   Int32   u         ;
   Int32   c         ;
   Int32   val       ;
   Int32   s         ;
   Int32   indice    ;
   Int32   t         ;
   Long    f1,f2,f3  ;
   Long    Ns        ;
   int     ind       ;
   int     g         ;

    Affiche("Affichage resultats",3) ;

   for (c = 0 ; c < eab ; c++) 	
   {
       /* pour chaque classe            */

       fprintf(stream,"\n") ;

       s   = Stats[c].Nr   ;
       Per = 100.0 * s / N ;

       val = Stats[c].Pos ;

       i   = Pos[val].Seq ;
       j   = Pos[val].Sol ;
       m   = Pos[val].Pal ;

       Ns  = Stats[c].Ns ;

         
       for (u = m - k + 1 ; u <= m ; u++) 
       {
	   fprintf(stream," %2d", M[i][j][m][u]) ;	 	 
       }

       fprintf(stream, " :  %4d/%4d ,  (%6.2f%%)\n",s,N,Per) ;

      
       /* Affichage des valeurs de l'attribut de chaque palindrome */

       if (NbAttributs) 
       {
	   fprintf(stream,
	   "                                          valeurs attributs : ") ;
	   Depart = c * k  ;

	   for (u = m-k+1, ind = 0 ; u <= m  ; u++, ind++) 
	   {

              if (ind < k-1 || Attributk1 > 0) {

		if (Attributs2[Depart + ind] == 0)
		    fprintf(stream," %2d", Pal[i][Sol[i][j][u]].Code) ;

		else if (Attributs2[Depart + ind] == 1)
		    fprintf(stream," %2d", VTete[i][Sol[i][j][u]]) ;

		else if (Attributs2[Depart + ind] == 2)
		   fprintf(stream," %2d", VBoucle[i][Sol[i][j][u]]) ;

		else if (Attributs2[Depart + ind] == 3)
		   fprintf(stream," %2d", d1[i][j][u]) ;

		else if (Attributs2[Depart + ind] == 4)
		   fprintf(stream," %2d", d2[i][j][u]) ;

		else if (Attributs2[Depart + ind] == 5)
		   fprintf(stream," %2d", d3[i][j][u]) ;
		else
		   fprintf(stream,"Erreur valeur attribut !!!\n") ;
              }
	     }
	    fprintf(stream,"\n") ;

	    /* affichage des codes des attributs			*/

            fprintf(stream,
	   "                                            codes attributs : ") ;

	    for (ind = 0 ; ind < k ; ind++)
                 if (ind < k-1 || Attributk1 > 0)
		    fprintf(stream,"  %d",Attributs2[Depart + ind]) ;

	    fprintf(stream,"\n") ;
	 }

       if (Moyenne)  
       {


           fprintf(stream,
         "      Nombre de solutions contenant la classe               : %12ld\n",Ns) ;


           fprintf(stream,
         "      Nombre moyen d'apparition par sequence representee    : %12.5f\n",
         Stats[c].m1) ;

       }


       if (Ecart)

          fprintf(stream,
         "                                               Ecart type   : %12.5f\n",Stats[c].e1) ;



         if (Moyenne) {

            fprintf(stream,
         "      Nombre moyen d'apparition par sequence                : %12.5f\n",
         Stats[c].m2) ;

          fprintf(stream,
	 "                                            classe numero   :  %d\n",
	       c) ;



         }


          if (Ecart)

          fprintf(stream,
	 "                                               Ecart type   : %12.5f\n",Stats[c].e2) ;


			/* sauvegarde taux de chaque structure      */

       if (st && k <= LMS) 
       {

	  indice = 0 ;

	  for (l = m-1 ; l >= m-k+1 ; l-- )
	      indice += M[i][j][m][l] * Fic[m-l] ;

	  indice += SFic[k-1] ;

	  Tstruct[indice] =  Per * 100 + 0.5 ;
       }

   }
   fprintf(stream,"\n\n") ;

   Affiche("Fin resultats",0) ;
   Affiche("",-3) ;

}



/* ---------------------------------------------------- */
/* algorithme KMRS                                      */
/* ---------------------------------------------------- */
/* arguments :                                          */
/* va    vecteur Int32 : data[i] = index dans un        */
/*          alphabet de l'element i de la sequence      */
/*          (l'indexation doit etre realisee par le     */
/*           programme appelant). Pour interdire des    */
/*          positions : mettre l'index -1.              */
/* Rmin   : # de fois ou un mot doit etre repete        */
/* maxlen : <- taille des mots a trouver :              */
/*          maxlen >  0 : mots de taille maxlen         */
/*          maxlen <= 0 : mots de plus grande taille    */
/*        : -> taille des mots trouves                  */
/* e      : <- taille de l'alphabet                     */
/*        : -> recoit le nombre de classes trouvees     */
/*             ou -1 si erreur memoire                  */
/*                                                      */
/*       les positions des mots                         */
/*    ou NULL si erreur ou rien trouve                  */
/* ---------------------------------------------------- */
/* notations de l'article Karp et al.                   */
/* vk : vecteur des pos des mots de long. k             */
/* ek : nbre de classes de mots de long. k              */
/*                                                      */
/* renvoi eventuellement (selon temoin f) les           */
/* positions de chaque structure repetees               */
/* @function: DoKmrS                                    */
/* ---------------------------------------------------- */

#define ShootInFlam(msg, step,n)    return AbortKmrAll(va, AllStr, msg, step,n)
#define CheckNullPtr(ptr,n)         if (! (ptr)) ShootInFlam("Not enough Memory", 0,n)

WordBunchPtr DoKmrS(Dtrna,check,td1, d1, td2 ,d2, td3, d3,
                    Ecart,Moyenne,lt,VTete,lb,VBoucle,purete,
                    st,Fic,SFic,Tstruct,f,Verbose,va,minlen,
		    maxlen,k,e,n,N,Rmin,t,p,Pos,Pal,Sol,M)

        int    Dtrna            ;    /* nb bases avant et apres TRNA    */
        int    check            ;    /* distribution structure 6000     */
        int    td1              ;    /* temoin calcul distance 1        */
	int    *** d1           ;    /* tableau distances 1             */
        int    td2              ;    /* temoin calcul distance 2        */
	int    *** d2           ;    /* tableau distances 2             */
        int    td3              ;    /* temoin calcul distance 3        */
	int    *** d3           ;    /* tableau distances 3             */
	int      Ecart		;    /* temoin calcul ecart type	*/
        int      Moyenne        ;    /* temoin calcul moyenne		*/
	int      lt		;    /* temoin tenir compte des tetes   */
	int    **VTete		;    /* tab des tetes des pals		*/
	int      lb		;    /* temoin tenir compte boucles	*/
	int    **VBoucle	;    /* tab des boucls des pals		*/
	int	purete		;    /* temoin tenir compte purete pal	*/
	int      st             ;    /*  temoin sauvegarde taux         */
	Int32    Fic[]          ;    /*  nb structures long 0..5        */
	Int32    SFic[]         ;    /*  nb structures long <=0 ..<= 5  */
	Int32    Tstruct[]      ;    /* tableau taux de chaque structur */
	int      f              ;    /* temoin de sauvegarde            */
	int      Verbose        ;    /* taux de blabla...               */
	Int32    va[]           ;    /* index des classes de taille 1   */
	Int32      minlen       ;    /* taille minimale des structures  */
	Int32      maxlen       ;    /* taille maximale structures      */
	Int32           k       ;    /* taille courante structures      */
	Int32         *e        ;    /* Nb de classe de taille 1        */
	Int32          n        ;    /* nb total pals                   */
	Int32          N        ;    /* nb sequences                    */
	Int32      Rmin         ;    /* nb minimal de repetitions       */
	Int32        * t        ;    /* nb solutions                    */
	Int32       ** p        ;    /* nb Pals                         */
	TPosition    * Pos      ;    /* indices Seq,Sol,Pal             */
	Palindrome  ** Pal      ;    /* Palindromes                     */
	Int32    *** Sol        ;    /* indices Pal dans Sol            */
	Int32     **** M        ;    /* Matrices                        */

{
	Int32      ea, eab        ;
	TStats     *Stats  = NULL ;
	WordBunchPtr AllStr       ;
	Long       Nex = 1        ;

				/* ---------------------------- */
				/* initialisation timing        */
				/* ---------------------------- */
	#ifndef CNRS
	(void) UserCpuTime(TIME_RESET);
	(void) SysCpuTime(TIME_RESET);
	#endif

				/* ---------------------------- */
				/* initialisation memoire       */
				/* ---------------------------- */

       Verb = Verbose ;
       Indent = 0 ;

       Affiche("Initialisation vab",3) ;

       if (purete || lb || lt || td1 || td2 || td3)
	  Attributs = 1 ;

	AllStr = NULL    ;

	fprintf(stderr,"\n\nRecherche des structures repetees (version 8.0) : \n\n") ;

	fprintf(stdout,"nombre d'entrees du dictionnaire %d\n",n) ;
	
	if (f) 
        {
	   AllStr = NewWordBunch(kMinStackiSize,kMinStackpSize) ;
	   CheckNullPtr(AllStr,n) ;
	}


	ea = *e ;


	*e = -1;                        /* assume error         */

				/* ---------------------------- */
				/* les structures de taille 1   */
				/* sont deja presentes          */
				/* ---------------------------- */



	do 
	{

	      /* recherche des structures de taille k + 1       */

	    eab = ApplyLemma1(Dtrna,check,td1, d1, td2, d2, td3, d3,
                               Ecart,lt,VTete,lb,VBoucle,purete,
                               Verbose,&Stats,Rmin,va,
			      k,ea,n,N,t,p,Pos,Pal,Sol,M,&Nex) ;

	    if (eab < 0)
		ShootInFlam("Not enough Memory", k+1,n);

	    if (eab > 0) 
	    {
                NotifyKmrStep(k+1, eab,Nex);

		if (Verbose >= 2 && k+1 >= minlen) 
                {
                    AfficherStats(td1, d1, td2, d2, td3, d3,
                                  Ecart,Moyenne,Sol,purete,Pal,
                                  lt,VTete,lb,VBoucle,
                                  st,Fic,SFic,Tstruct,stdout,eab,
                                  Stats,N,k+1,Pos,M) ;
		   free(Stats) ;
		}

		ea = eab ;

		if (NbAttributs) 
                {
		   if (k > 0)
		      free(Attributs1) ;
		   Attributs1 = Attributs2 ;
		}

		k++      ;

		if (f && (k >= minlen) && (k <= maxlen)) 
                {
		   if (!PUSH(&(AllStr->nbwords),ea))
		       return NULL ;
		   if (!PUSHP(&(AllStr->words), (Ptr) DualVa(va,n,ea)))
		      return NULL  ;
		}
	    }
	}
     while (eab > 0 && k < maxlen) ;

 
				/* ---------------------------- */
				/* fin traitement des longueurs */
				/* intermediaires               */
				/* le dernier resultat est dans */
				/* va et le # de classe ds ea   */
				/* et sa longueur est k         */
				/* ---------------------------- */

	*e     = ea ;
	maxlen = k  ;

	if (f) 
	{
	   AllStr->minlen = minlen ;
	   AllStr->maxlen = maxlen ;
	}

	NotifyKmrEnd(k, k,UserCpuTime(TIME_NO_RESET) + SysCpuTime(TIME_NO_RESET));	

	if (va)
	   free(va)  ;

	return AllStr ;

} /* end of DoKmrSimple */

#undef ShootInFlam
#undef CheckNullPtr  
