/* ==================================================== */
/*	Copyright (c) Atelier de BioInformatique	*/
/*	Mar. 92 					*/
/* 	File: libsysks.c				*/
/*	Purpose: library to deal with sys calls		*/
/*		 and notifications			*/
/*	History:					*/
/*	00/03/92 : <Gloup> first draft			*/
/*	15/08/93 : <Gloup> revised version		*/
/*      20/02/95 : adaptations mineurs aux structures	*/
/* ==================================================== */


#ifndef CNRS 

#ifndef THINK_C
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif

#endif




#include <stdio.h>
#include <stdlib.h>




#include "../../include/gtypes.h"
#include "../../include/libsysks.h"



#define ERROUT	stderr		/* error output stream	*/


/* ====================================================	*/
/* statiques locales					*/
/* ====================================================	*/

static Bool 	sNotifState = Faux;	/* notification state	*/

/* ====================================================	*/
/* Kmr notification					*/
/* ====================================================	*/

/* -------------------------------------------- */
/* setup etat de notification			*/
/*						*/
/* @function: SetUpKmrcNotify			*/
/* -------------------------------------------- */

void SetUpKmrNotify(notif)
	Bool	notif;
{
	sNotifState = notif;

} /* end of SetUpKmrcNotify */

/* -------------------------------------------- */
/* get etat de notification			*/
/*						*/
/* @function: GetKmrNotify			*/
/* -------------------------------------------- */

Bool GetKmrNotify()
{

	return sNotifState;

} /* end of GetKmrNotify */
   
/* -------------------------------------------- */
/* notification de l'etat d'avancement de kmr	*/
/*						*/
/* @function: NotifyKmrStep			*/
/* -------------------------------------------- */

void NotifyKmrStep(wlen, wnb, Nex)
	Int32 wlen, wnb ;
	Long Nex ;
{
        float Per ;

        Per = 100.0 * wnb / Nex ;

	if (! sNotifState)
	    return;

	fprintf(ERROUT, "Kmr: %4d class(es) / %7ld (%6.2f %%) of length %d\n",
                wnb, Nex, Per, wlen) ;
	fprintf(stdout, "Kmr: %4d class(es) / %7ld (%6.2f %%) of length %d\n",
                wnb, Nex, Per, wlen) ;

            
        
	
} /* end of NotifyKmrStep */

/* -------------------------------------------- */
/* notification de fin de kmr			*/
/*						*/
/* @function: NotifyKmrEnd			*/
/* -------------------------------------------- */

void NotifyKmrEnd(Int32 wlen, Int32 maxlen, float tcpu)
	{
	if (! sNotifState)
	    return;

	if (wlen == maxlen)
	    fprintf(ERROUT, "Kmr: reached maximum length %d\n\n", maxlen);
	else
	    fprintf(ERROUT, "Kmr: reached end at length %d\n\n", wlen);

        #ifndef CNRS

            if (tcpu >= 0.)	
	        fprintf(ERROUT, "total duration = %.2f sec.\n", tcpu);

        #endif

        if (wlen == maxlen)
	    fprintf(stdout, "Kmr: reached maximum length %d\n\n", maxlen);
	else
	    fprintf(stdout, "Kmr: reached end at length %d\n\n", wlen);

           #ifndef CNRS

            if (tcpu >= 0.)	
	        fprintf(stdout, "total duration = %.2f sec.\n", tcpu);

        #endif



} /* end of NotifyKmrEnd */

/* -------------------------------------------- */
/* notification d'une erreur			*/
/*						*/
/* @function: NotifyKmrError			*/
/* -------------------------------------------- */

void NotifyKmrError(msg, wlen)
	char 	*msg;
	Int32	wlen;
{
	fprintf(ERROUT, "Kmr: Error at step %d\n", wlen);
	fprintf(ERROUT, "Kmr: %s\n", msg);

} /* end of NotifyKmrError */




#ifndef CNRS


/* ====================================================	*/
/* Cpu time sys calls					*/
/* ====================================================	*/

/* -------------------------------------------- */
/* Get(/Reset) User Cpu time			*/
/*						*/
/* @function: UserCpuTime			*/
/* -------------------------------------------- */

float UserCpuTime(reset)
	Bool reset;
{
	static  float sLast = 0;

	float  now, ust;

#ifdef THINK_C
	now = (float) TickCount();
	ust = (now - sLast) / (float) TICKS_PER_SEC;
#else
     	struct rusage rusage;

	(void) getrusage(RUSAGE_SELF, &rusage);

     	now =   (float) rusage.ru_utime.tv_sec
	      + ((float) rusage.ru_utime.tv_usec / 1000000.);
	ust = now - sLast;
#endif

	if (reset)
	   sLast = now;

	return ust;
}

/* -------------------------------------------- */
/* Get(/Reset) Sys Cpu time			*/
/*						*/
/* @function: SysCpuTime			*/
/* -------------------------------------------- */

float SysCpuTime(reset)
	Bool reset;
{
	static  float sLast = 0;
	
	float  now, ust;

#ifdef THINK_C
	ust = 0.;
#else
     	struct rusage rusage;
     	
	(void) getrusage(RUSAGE_SELF, &rusage);

     	now =   (float) rusage.ru_stime.tv_sec
	      + ((float) rusage.ru_stime.tv_usec / 1000000.);
	ust = now - sLast;   
#endif

	if (reset) 
	   sLast = now;

	return ust;
}

/* -------------------------------------------- */
/* Get a Cpu Time string			*/
/*						*/
/* @function: StrCpuTime			*/
/* -------------------------------------------- */

char *StrCpuTime(reset)
 	Bool	reset;
{
	static char buffer[256];

	float ust, syt, tot;

	ust = UserCpuTime(reset);
	syt = SysCpuTime(reset);
	tot = ust + syt;

	(void) sprintf(buffer, "@ Cpu time user: %f sys: %f tot: %f\n",
				ust, syt, tot);

	return buffer;
}

#endif   

