/* ==================================================== */
/*	Copyright (c) Atelier de BioInformatique	*/
/*	Mar. 94 					*/
/* 	File: libwbs.c					*/
/*	Purpose: Word bunch handling library		*/
/*	History:					*/
/*	15/08/93 : <Gloup> first draft from libkmr	*/
/*        /02/95 : adaptation aux structures 		*/
/* ---------------------------------------------------- */
/* un "WordBunch" est la representation d'un ensemble	*/
/* de "groupes" de mots, chaque groupe contenant 	*/
/* les index des occurences des instances de ce		*/
/* groupe sur la sequence. un groupe peut etre une	*/
/* classe (kmr), une clique (kmrc) etc..		*/
/* la structure de donnees employee, reutilise les	*/
/* piles ("stacks") de libstki et libstkp.		*/
/*							*/
/* typedef struct {					*/
/*   Int32	minlen, maxlen; longueurs min et max	*/
/*				des mots		*/
/*   StackiPtr  nbwords;  stack du # de mots trouves	*/
/*			  pour chaque longueur		*/
/*   StackpPtr  words;	stack de pointeurs :		*/
/*			chaque elt du stack (un 	*/
/*			groupe de mot) pointe sur	*/
/*			un stack d'entiers donnant les	*/
/*			positions (index) des mots	*/
/* } WordBunch;						*/
/* ====================================================	*/

#ifndef THINK_C
#include <sys/types.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#

#include "../../include/gtypes.h"
#include "../../include/libsysks.h"
#include "../../include/libstki.h"
#include "../../include/libstkp.h"
#include "../../include/libwbs.h"





/* -------------------------------------------- */
/* liberation d'un WordBunch			*/
/*						*/
/* always returns NULL 				*/
/*						*/
/* @function: FreeWordBunch 			*/
/* -------------------------------------------- */

WordBunchPtr FreeWordBunch(wb)
	WordBunchPtr wb;
{
	Int32	   n;
	StackiHdle w;
	
	if (! wb)
	    return NULL;

	while (    PopiOut(&(wb->nbwords), &n)
	    	&& PoppOut(&(wb->words), (Ptr *) &w))
		FreeStackiVector(w, n);

	FreeStacki(wb->nbwords);
	FreeStackp(wb->words);
	    	
	FREE(wb);
	
	return NULL;

} /* end of FreeWordBunch */

/* -------------------------------------------- */
/* creation d'un WordBunch			*/
/* stackxSize : taille <initiale> des stack	*/
/*						*/
/* @function: NewWordBunch			*/
/* -------------------------------------------- */

WordBunchPtr NewWordBunch(stackiSize, stackpSize)
	Int32	stackiSize, stackpSize;
{
	WordBunchPtr	wb;
	
	if (! (wb = NEW(WordBunch)))
	    return NULL;
	    
	wb->nbwords = NULL;
	wb->words   = NULL;
	
	if (! (wb->nbwords = NewStacki(stackiSize)))
		return FreeWordBunch(wb);
		
	if (! (wb->words = NewStackp(stackpSize)))
		return FreeWordBunch(wb);

	return wb;
	
} /* end of NewWordBunch */   
