/* PALS6.C								*/
/* fichier contenant les entetes et les corps de fonctions concernant	*/
/* la gestion des palindromes.						*/
/* 03/09/94								*/
/* 21/02/95								*/
/* nouvelle version adaptee aux Int32 et changement DTete DTete		*/
/* pour etre compatible avec kmrstr5 et libkmr5				*/


#include <stdio.h>


#include "../../include/gtypes.h"
#include "../../include/pals6.h"




/* definition des fonctions */


/* verifie que 2 palindromes P1 et P2 sont juxtaposes et qu'il   */
/* n'existe aucun palindrome juxtapose entre			  */
int  Juxta0(int *Mjuxta[],int i, int j)

     {
        int t ;

	if (Mjuxta[i][j] == 1) {
	    t = i+1 ;
            while (t < j && !(Mjuxta[i][t] && Mjuxta[t][j]))
                   t++ ;
            return (t==j) ;

        }
        else
            return 0 ;
     }


  /* verifie la juxtaposition de 2 palindromes */
int Juxtapose(int Tolerance,Palindrome P1,Palindrome P2)
    {
      return (P1.FQueue - Tolerance < P2.DTete) ;
    }


 /* verifie l'imbrication d'un palindrome dans un autre */
int Imbrique(int Tolerance,Palindrome P1, Palindrome P2)
    {
       return (P1.FTete - Tolerance < P2.DTete &&
	       P2.FQueue - Tolerance < P1.DQueue) ;
    }


 /* verifie si la tete de P2 est incluse dans la boucle de P1 */
int Quinconce(int Tolerance,Palindrome P1, Palindrome P2)
    {
       return (P1.FTete - Tolerance < P2.DTete &&
	       P2.FTete - Tolerance < P1.DQueue &&
	       P1.FQueue - Tolerance < P2.DQueue) ;
    }

 /* verifie si un palindrome contient au moins une liaison impure */
int Impure(Palindrome P1)
    {
      return (P1.Code == 0) ;
    }

void PrintPal(FILE * sortie,Palindrome P,chaine suffixe)
{

   fprintf(sortie,"[%d-%d-%d-%d-%d]",
	   P.DTete,P.FTete,
	   P.DQueue,P.FQueue,
	   P.Code) ;
   fprintf(sortie,suffixe) ;
}


/* verifie l'egalite de 2 palindromes	*/
int EgalPal(Palindrome P1, Palindrome P2)
{
   return (P1.DTete == P2.DTete &&
	   P2.FTete == P2.FTete &&
	   P1.DQueue == P2.DQueue &&
	   P1.FQueue == P2.FQueue &&
	   P1.Code == P2.Code
	  ) ;
}




/* verifie que 2 palindromes sont compatibles (strictement disjoints) */
int Compatible(int Tolerance, Palindrome P1, Palindrome P2)

{
    return(
	    Place(Tolerance, P2.DTete,P2.FTete,P1) &&
	    Place(Tolerance, P2.DQueue,P2.FQueue,P1)
	  ) ;
}


/* verifie qu'un ensemble de palindromes est compatibles 2 a 2 		*/
int Compatibilite(int Tolerance,Palindrome P[], int n)
{
   int i ;
   int k ;
   int ok ;

   ok = 1 ;
   i = 0 ;
   while (ok && i < n-1) {
	 k = i + 1 ;
	 while (ok && k < n)
	       if (!Compatible(Tolerance, P[i],P[k]))
		  ok = 0 ;
	       else
		  k++ ;
	 i++ ;
   }
   return ok ;
}




/* lit une sequence  AVEC SCORE 					*/
int  LireSeqScore(FILE *entree,Palindrome Pals[],char *Seq,int *NbPals)

{
  char sep[5]  ;
  int j        ;
  int k        ;

   (*NbPals) = 0 ;
   fscanf(entree,"%s",Seq) ;
   if (!feof(entree)) {

      while (fscanf(entree,"%d %d %d %d %d %d",
		    &(Pals[*NbPals].DTete), &(Pals[*NbPals].FTete),
		    &(Pals[*NbPals].DQueue),&(Pals[*NbPals].FQueue),
		    &(Pals[*NbPals].Code),&(Pals[*NbPals].Score))
	    )
	     (*NbPals)++ ;
      fscanf(entree,"%s",sep) ;
      return 1 ;
   }
   else
      return 0 ;

}





int Place(int Tolerance, int Debut, int Fin, Palindrome P)
{
   return(
	 P.FQueue < Debut + Tolerance ||
	 (P.FTete < Debut + Tolerance  && Fin < P.DQueue + Tolerance) ||
	 Fin < P.DTete + Tolerance
	) ;
}   
