/* ---------------------------------------------------- */
/* @File : dbrand.c 					*/
/* ---------------------------------------------------- */

/* -------------------------------------------- */
/* randomize une Swiss prot DB / CD ROM / DD	*/
/* usage dbrand database > database_randomisee	*/
/* -------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>

#include "libdb.h"
#include "swiss.h"

/* prototype	*/
static void shuffle_seq(char *s, int times);
static void print_seq(char *seq, int lenline);
static void clean_seq(char *seq);

/* -------------------------------------------- */
/* main						*/
/* -------------------------------------------- */
main(argn, argv)
 	int  argn;
 	char **argv;
{

	int 	status, irec, len;
		
	char	*record, *data, seq[BUFSIZ];

					/* ---------------------------- */
					/* check args			*/
					/* ---------------------------- */
	
	if (argn <= 1) {
	   fprintf(stderr, "usage: dbtest database_name [> database_rand]\n");
	   exit(1);
	}

	if (access(argv[1], R_OK)) {
	   fprintf(stderr, "dbtest: cannot access %s\n",  argv[1]);
	   exit(2);
	}
	

					/* ---------------------------- */
					/* open data base		*/
					/* ---------------------------- */

	status = db_open(argv[1], 0);		
	if (status != eDB_NoErr) {
	    fprintf(stderr, "dbtest: cannot open  %s ; errno %d \n",  
			     argv[1], status);
	   exit(2);
	}

					/* ---------------------------- */
					/* get entry			*/
					/* ---------------------------- */

	irec = 0;

	while (record = db_next_record(swiss_move_mark)) {
	
		irec ++;

		data = swiss_seqname(record);
		fprintf(stderr, "Record: %s\n", data);
		swiss_unlock_record(record);
		 
		data = swiss_get_keyword(record, "ID");
		printf("%s\n", data);
		swiss_unlock_record(record);


		data = swiss_get_keyword(record, "DE");
		printf("%s\n", data);
		swiss_unlock_record(record);

		data = swiss_get_keyword(record, "SQ");
		printf("%s\n", data);
		swiss_unlock_record(record);

		data = swiss_seqdata(record);
		strcpy(seq, data);
		swiss_unlock_record(record);

				/* shuffle_seq sequence	*/
		shuffle_seq(seq, 10);

				/* clean sequence	*/
		clean_seq(seq);

				/* print sequence	*/
		print_seq(seq, 60);


				/* print final		*/
		printf("//\n");

	}

	fprintf(stderr, "\n%d Records\n", irec);
	fprintf(stderr, "*End*\n");
		
			

					/* ---------------------------- */
					/* close DB file		*/
					/* ---------------------------- */
		
	db_close();


	exit(0);
}



static void shuffle_seq(char *s, int times)
{
   char		temp;
   double	big;
   int		i, len, sup, randoffset1, randoffset2;

   len = (int)strlen(s);
   sup = times*len;		/* on tourne times fois	longueur seq */
   big = (double)pow(2,31);

   for (i = 0; i < sup; i++) {
      randoffset1 = ((double)random()/big)*len;
      randoffset2 = ((double)random()/big)*len;
      temp = s[randoffset1];
      s[randoffset1] = s[randoffset2];
      s[randoffset2] = temp;
   }
}

static void print_seq(char *seq, int lenline)
{
   int len, i, lines;
   char	*s, *ss, temp;


   len = (int)strlen(seq);

   lines = len/lenline;

   for (i = 0, s = seq, ss = s + lenline; i < lines;
			i++, s += lenline, ss += lenline) {
	temp = *ss;
	*ss = '\000';
	printf("     %s\n", s);
	*ss = temp;
   }

   printf("     %s\n", s);

   return;
}

static void clean_seq(char *seq)
{
   int len = (int)strlen(seq);
   char *s = seq;


   seq--;
   while (*++seq) {
      if (isalpha(*seq))
	*s++ = *seq; 
   }
   *s = '\000';
}
