/* ============================================================ */
/*	   							*/
/* fichier distribPE.c						*/
/* calcul la distribution des longueurs de pal et des energies  */
/* palindromes situes dans un(e zone determine d'un) ARN	*/
/*								*/
/* entree	: stdin (fichier issu de Palamou)		*/
/* sortie	: stdout (fichier contenant distributions	*/
/*                moyenne sur l'ensemble des sequences)		*/
/* date 	: 15/10/97					*/
/* ============================================================ */


		

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

		/* fichiers inclus 				*/

#include "../Param/param.h"
#include "../../../Include/gtypes.h"
#include "../pals5.h"


#define LONGSEQ 500
#define MAXNOM   30
#define NB_MAX   25



/* ==================================================================== */
/* 			  MACROS					*/
/* ==================================================================== */

#define PointeurNul(a,b) {if(!(a)) {fprintf(stderr,"Probleme allocation sur %s\n",(b)) ; return (0) ; }}


/* ==================================================================== */
/* 			  TYPES					        */
/* ==================================================================== */
typedef int TDistribution[NB_MAX] ;



/* ==================================================================== */
/* 	               FONCTIONS LOCALES				*/
/* ==================================================================== */




/* lit une sequence  AVEC SCORE 					*/
static int  LireSequence(int Comment, int Inv, int Distingue, char* Commentaire,
                         Palindrome Pals[],char *Seq,int *NbPals,
                         int* DebPart, int* FinPart)

{
  char sep[5]  ;

   (*NbPals) = 0 ;

    if (Comment)
        fscanf(stdin, "%s", Commentaire) ;  

    if (Inv)
        fscanf(stdin,"%d %d", DebPart, FinPart) ;

    fscanf(stdin,"%s",Seq) ;

    if (!Inv && Distingue)
        fscanf(stdin,"%d %d", DebPart, FinPart) ;


   if (!feof(stdin)) 
   {    
      while (fscanf(stdin,"%d %d %d %d %d %d",
		    &(Pals[*NbPals].Dtete) , &(Pals[*NbPals].Ftete) ,
		    &(Pals[*NbPals].Dqueue), &(Pals[*NbPals].Fqueue),
		    &(Pals[*NbPals].Code)  , &(Pals[*NbPals].Score))
	    )
	     (*NbPals)++ ;

      fscanf(stdin,"%s",sep) ;
      return 1 ;
   }
   else
      return 0 ;

}


static int DansIntervalle(Int32 n, Int32 Debut, Int32 Fin)
{
  return (n >= Debut && n <= Fin) ; 
 }


static int DansZone(Palindrome Pal,Int32 LongPal, Int32 DebPart, Int32 FinPart)
{
   Int32 k ;

   k = 0 ;

   while ( k < LongPal && 
           !DansIntervalle(Pal.Dtete  + k, DebPart,FinPart) &&
           !DansIntervalle(Pal.Dqueue + k, DebPart,FinPart)
         )
            k++ ;

   return (k < LongPal) ;
}


static void CalculeDistribution(Int32 NbPals , Palindrome* Pals, 
                                Int32 DebPart, Int32 FinPart, int* DistributionPals,
                                int* DistributionEnergies) 
{

    int   j	    ;  /* indice de palindromes				*/
    int   k         ;  /* indice des longueures de palindromes		*/
    int LongPal     ;  /* longueur du palindrome courant		*/
    int Energie     ;  /* energie du palindrome	en kcal/mol		*/
    

  
				/* initialisation des distributions	*/

    for (k = 0 ; k < NB_MAX ; k++)
    {
         DistributionPals[k]     = 0 ;
	 DistributionEnergies[k] = 0 ;
    }

    for (j = 0 ; j < NbPals ; j++)
    {
         LongPal = Pals[j].Ftete - Pals[j].Dtete + 1 ;
	
         if (DansZone(Pals[j], LongPal, DebPart, FinPart))
	 {
	     Energie = (int)(0.5 + Pals[j].Score / 100)   ;
             DistributionEnergies[Energie]++              ; 
             DistributionPals[LongPal]++                  ;            
	 }
    }
}



/* ==================================================================== */
/*  				main					*/
/* ==================================================================== */

int main (argc,argv)
          int argc     ;
          char *argv[] ;
{   
   Palindrome *Pals             ;  /* Palindromes de chaque sequence    */
   char Seq[LONGSEQ]       	;  /* sequence				*/
   int  MaxPal = 500		;  /* nombre max pal dans sequence	*/
   int  NbPals  	   	;  /* nb palindromes dans sequence	*/   			   
   Int32      i                 ;  /* indice des sequences		*/
   Int32      NbSeq 		;  /* nombre de sequence a traiter	*/
   char       Symb              ;  /* symbole argument de main		*/
   char      Argument[50]	;  /* argument de main			*/
   Int32     DebPart		;
   Int32     FinPart		; 
   TDistribution* DistributionPals     = NULL   ; /* distribution des longueurs	*/
   TDistribution* DistributionEnergies = NULL   ; /* distribution des energies	*/
   float SDistributionPals[NB_MAX]     ;
   float SDistributionEnergies[NB_MAX] ;		
   Int32 k	                ;    
   Int32* NbPalDansZone         ;      
   Int32 SXPals		        ;		
   Int32 SX2Pals	        ;
   float* MuPals     = NULL	; 
   float* VarPals    = NULL	;
   float* Pourcent   = NULL     ;   
   float SMuPals		;
   float SVarPals		;
   Int32 SXEnergies	        ;		
   Int32 SX2Energies	        ;
   float* MuEnergies  = NULL	; 
   float* VarEnergies = NULL	; 
   float SMuEnergies		;
   float SVarEnergies		;
   float SNbPalDansZone		;  
   float SPourcent		;
   int   Distingue = 0		; /* 1 si zone particuliere RNA 	*/
   int   Verbose   = 0          ;
   int   PasZone   = 0          ; /* 1 si zone qu' on ne considere pas  */
   int   Inv	   = 0          ; /* inversion de lecture de la zone	*/
   int   Comment   = 0          ; /* commentaire en debut sequence	*/
   char  Commentaire[300]	;
  

   /* debut du programme principal */

   fprintf(stderr,"Debut du programme distribPE (version Octobre 97)\n") ;

   Commentaire[0] = 0 ;

   while (SelectArg("-k -v -p -i", &Symb, Argument, argc, argv)) 
   {
	  switch (Symb) 
          {
 		case 'k' :  Distingue = 1 ;
			    break	  ; 

		case 'v' :  Verbose   = 1 ;
			    break	  ; 

		case 'p' :  PasZone   = 1 ;
                            Distingue = 1 ;
			    break	  ; 

                case 'i' :  Inv       = 1 ;
                            Distingue = 1 ;
			    break	  ; 
	  }
    }

   fscanf(stdin,"%d",&NbSeq) ;

   Pals = (Palindrome*) malloc (MaxPal * sizeof(Palindrome)) ;
   PointeurNul(Pals,"Pals")    
						/* allocation					*/
   MuPals = (float*) malloc (NbSeq * sizeof(float)) ;
   PointeurNul(MuPals,"MuPals")
 
   MuEnergies = (float*) malloc (NbSeq * sizeof(float)) ;
   PointeurNul(MuEnergies,"MuEnergies")

   VarPals = (float*) malloc (NbSeq * sizeof(float)) ;
   PointeurNul(VarPals,"VarPals")
 
   VarEnergies = (float*) malloc (NbSeq * sizeof(float)) ;
   PointeurNul(VarEnergies,"VarEnergies") 

   Pourcent = (float*) malloc (NbSeq * sizeof(float)) ;
   PointeurNul(Pourcent,"Pourcent")   
  
   DistributionPals = (TDistribution*) malloc (NbSeq * sizeof(TDistribution)) ;
   PointeurNul(DistributionPals,"DistributionPals")  

   DistributionEnergies = (TDistribution*) malloc (NbSeq * sizeof(TDistribution)) ;
   PointeurNul(DistributionEnergies,"DistributionEnergies")  

   NbPalDansZone = (int*) malloc(NbSeq * sizeof(int)) ;
   PointeurNul(NbPalDansZone, "NbPalDansZone")
 		
   for (i = 0 ; i < NbSeq ; i++)    	      
   {	
      
      LireSequence(Comment, Inv, Distingue, Commentaire, Pals, Seq,
                   &NbPals, &DebPart, &FinPart) ;

      fprintf(stderr,"sequence : %d\n",i+1) ;
	  
      if (!Distingue || PasZone)
      {
       DebPart = 1 ;
       FinPart = strlen(Seq) ;
      }
		              	    
      CalculeDistribution(NbPals, Pals,DebPart, FinPart, 
                             DistributionPals[i], DistributionEnergies[i]) ;

				/* calcul des moments	et affichage			*/

      SXPals           = 0 ;
      SX2Pals          = 0 ; 	     
      NbPalDansZone[i] = 0 ;
      SXEnergies       = 0 ;
      SX2Energies      = 0 ;

      for (k = 0 ; k < NB_MAX ; k++)
      {                 		
	   NbPalDansZone[i] += DistributionPals[i][k]               ;
	   SXPals           += (DistributionPals[i][k] * k)         ;
	   SX2Pals          += (DistributionPals[i][k] * k * k)     ;

	   SXEnergies       += (DistributionEnergies[i][k] * k)     ;
	   SX2Energies      += (DistributionEnergies[i][k] * k * k) ;
       }
			     
       MuPals[i]      = 1.0 * SXPals  / NbPalDansZone[i]            ; 
       VarPals[i]     = 1.0 * SX2Pals / NbPalDansZone[i]  
                              - MuPals[i] * MuPals[i]               ;	

       MuEnergies[i]  = 1.0 * SXEnergies  / NbPalDansZone[i]        ; 
       VarEnergies[i] = 1.0 * SX2Energies / NbPalDansZone[i]  
                               - MuEnergies[i] * MuEnergies[i] 	    ;	
	      	
       Pourcent[i] = 1.0 * NbPalDansZone[i] / NbPals ;

       if (Verbose)
       {
					/* affichage des resultats pour chaque seq   	       */
        fprintf(stdout,"Sequence %d\n", i)                                                                ;
        fprintf(stdout,"NbPals (dans zone) = %3d (%2.f%%)\n"       ,NbPalDansZone[i],100 * Pourcent[i]) ;
        fprintf(stdout,"Palindromes        : Mu = %6.2f, Var = %6.2f\n"  ,MuPals[i], VarPals[i])          ; 
        fprintf(stdout,"Energies           : Mu = %6.2f, Var = %6.2f\n\n",MuEnergies[i], VarEnergies[i])  ;

        fprintf(stdout,"   k  Pals Energies\n\n") ;

        for (k = 0 ; k < NB_MAX ; k++)
              fprintf(stdout,"%5d %3d %3d\n", k, 
                     DistributionPals[i][k], DistributionEnergies[i][k]) ;

        fprintf(stdout,"\n") ;
       }                   	   					     	      	                   
    }						 		/* fin i	*/
				
				/* initialisation				*/

    SMuPals        = 0.0      ;
    SVarPals       = 0.0      ;
    SMuEnergies    = 0.0      ;
    SVarEnergies   = 0.0      ;
    SPourcent      = 0.0      ;
    SNbPalDansZone = 0.0      ;

    for (k = 0 ; k < NB_MAX ; k++)	     
    {
	 SDistributionPals[k]     = 0.0 ;
	 SDistributionEnergies[k] = 0.0 ; 
    }
				/* moyennes sur les sequences			*/

    for (i = 0 ; i < NbSeq ; i++)
    {
	for (k = 0 ; k < NB_MAX ; k++)
	{	     
	     SDistributionPals[k]     += DistributionPals[i][k] ;
	     SDistributionEnergies[k] += DistributionEnergies[i][k] ;
	}
	
	SMuPals        += MuPals[i]        ;
	SVarPals       += VarPals[i]       ; 
        SMuEnergies    += MuEnergies[i]    ;
	SVarEnergies   += VarEnergies[i]   ; 
	SPourcent      += Pourcent[i]      ;
        SNbPalDansZone += NbPalDansZone[i] ;	    
    }

    SMuPals        /= (1.0 * NbSeq) ;
    SVarPals       /= (1.0 * NbSeq) ;
    SMuEnergies    /= (1.0 * NbSeq) ;
    SVarEnergies   /= (1.0 * NbSeq) ;
    SPourcent      /= (1.0 * NbSeq) ;
    SNbPalDansZone /= (1.0 * NbSeq) ;
    
    

    for (k = 0 ; k < NB_MAX ; k++)	     
    {
	 SDistributionPals[k]     /= (1.0 * NbSeq) ;
	 SDistributionEnergies[k] /= (1.0 * NbSeq) ; 
    }
	
					/* affichage des resultats			*/
     fprintf(stdout,"Moyenne sur les sequences :\n") ;
     
     fprintf(stdout,"NbPals (dans zone) = %6.2f (%2.f%%)\n",SNbPalDansZone,100 * SPourcent )  ;
     fprintf(stdout,"Palindromes        : Mu = %6.2f, Var = %6.2f\n",SMuPals, SVarPals)           ; 
     fprintf(stdout,"Energies           : Mu = %6.2f, Var = %6.2f\n\n",SMuEnergies, SVarEnergies) ;
      
     fprintf(stdout,"   k  Pals Energies\n\n") ;
     for (k = 0 ; k < NB_MAX ; k++)
         fprintf(stdout,"%5d %6.2f % 6.2f\n", k, 
                 SDistributionPals[k], SDistributionEnergies[k]) ;

              
					/* desalocation					*/          					

    free(Pals)                  ;
    free(MuPals)                ;
    free(VarPals)               ;
    free(MuEnergies)            ;
    free(VarEnergies)           ;
    free(Pourcent)              ;
    free(NbPalDansZone)         ;   
    free(DistributionPals)      ;   
    free(DistributionEnergies)  ;
            
    fprintf(stderr,"\nfin du programme distribPE\n") ;

    return 1 ;
 
}      
