/* ------------------------------------- */
/* Shuffle2                              */
/* Shuffles Fasta sequences, keeping the */
/* dinucleotide frequencies constant.	 */
/* ------------------------------------- */

#ifndef THINK_C
#include <sys/types.h>
#include <sys/param.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#include "Gtypes.h"
#include "libsys.h"
#include "libfasta.h"

#ifdef THINK_C
extern int __ccommand(char ***argv);
#define LINE_FEED '\r'
#else
#define LINE_FEED '\n'
#endif

                        /* ------------------------------------ */
                        /* messages printout                    */
                        /* ------------------------------------ */
#define VERBOSE 1
#define VERBOSMAX 0
#if VERBOSE
#define NOTIFY(msg)             BANNER(msg)
#define FMTNOTIFY(fmt, msg)     FMTBANNER(fmt, msg)
#else
#define NOTIFY(msg)
#define FMTNOTIFY(fmt, msg)
#endif
#if VERBOSMAX
#define NOTIFMAX(msg)             BANNER(msg)
#define FMTNOTIFMAX(fmt, msg)     FMTBANNER(fmt, msg)
#else
#define NOTIFMAX(msg)
#define FMTNOTIFMAX(fmt, msg)
#endif

#define  SEQ_ASKED_OUT		/* pour sortir la sequence	*/

#undef DEBUG


/* -------- */
/* defines  */
/* -------- */

#define ALPHABET	"GATC"
#define T_ALPH 4

#define OUTPUT_REQUIRED 1

#define CHAR_PER_LIGNE 50

#define MAX_SEQ 256000

/*----------------------- */
/* definitions des types  */
/* ---------------------- */

/* ----------------------------------- */
/* definitions des variables globales  */
/* ----------------------------------- */

static char *alphabet = ALPHABET;

/* ----------------- */
/* local prototypes  */
/* ----------------- */

static void		Erreur P((char *msg, int stat));

static int		rang P((char *c));

static void		Shuffle2 P((char *s, int times));

static void		SortieSeq P((FastaSequence *seqFasta, FILE *streamout));


/* ------- */
/* erreur  */
/* ------- */

static void Erreur(msg, stat)
	char *msg;
	int stat;
{
   fprintf(stderr, "Error: %s\n", msg);
   if (stat) exit(stat);
}



/* ---------------------------------------------- */
/* calcule le rang d'un caractere dans l'alphabet */
/* ---------------------------------------------- */

static int rang (c)
	int	c;
{
	int	rng;
	char	*ptr;

	if (ptr=strchr(alphabet,c))
		rng=(int)(ptr-alphabet);
	else
		rng = -1;

	return rng;

}

/* ------------------------------------------------------ */
/* shuffle en preservant les frequences des dinucleotides */
/* ------------------------------------------------------ */

static void shuffle2(fsq, times)
	FastaSequence	*fsq;
	int		times;
{
   char		*s,temp;
   int		i, j, len, sup, randpos, avt, apr, stock;
   int		pos[T_ALPH][T_ALPH];

   s = fsq->seq;

   for (i=0 ; i<T_ALPH ; i++)
   for (j=0 ; j<T_ALPH ; j++)
	pos[i][j]=0;

   len = (int) strlen(s);
   sup = times*len;		/* on tourne ( times x len ) fois	*/

   for (i = 0; i < sup; i++) 
   {

	avt=apr=-1;

	while (avt==-1 || apr==-1) 
        {
		randpos = random() % (len-2) + 1;
		avt=rang(s[randpos-1]);
		apr=rang(s[randpos+1]);
	}

	if ((stock=pos[avt][apr])==0)
		pos[avt][apr]= randpos;
	else 
        {
		temp=s[randpos];
		s[randpos]=s[stock];
		s[stock]=temp;
		pos[avt][apr]=0;
	}

   }

}


/* ------------------------- */
/* Impression des resultats  */
/* format FASTA		     */
/* ------------------------- */

static void SortieSeq(seqFasta,streamout)
   FastaSequence	*seqFasta;
   FILE		*streamout;

{
   int		i,Compteur;
   Int32	longueur, nlines ;
   char		*buf, ligne[CHAR_PER_LIGNE+1];

   					/* ecriture nom & commentaires	 */

   fprintf(streamout, ">S2-%s ",seqFasta->name) ;
   fprintf(streamout, " / shuffled / %s%c",seqFasta->comment,LINE_FEED) ;

   					/* ecriture de la sequence */
    
   longueur = seqFasta->length ;
   nlines = (Int32) longueur/CHAR_PER_LIGNE; 
   buf= seqFasta->seq ;
   for (i = 0 ; i < nlines ; i++) { 
        strncpy(ligne, buf+(i*CHAR_PER_LIGNE), CHAR_PER_LIGNE) ;
	ligne[CHAR_PER_LIGNE] = '\000';
        fprintf(streamout,"%s%c",ligne, LINE_FEED) ;
   }
   
   if (longueur%CHAR_PER_LIGNE) {
        strcpy(ligne, buf + (i*CHAR_PER_LIGNE)) ;
        fprintf(streamout,"%s%c",ligne, LINE_FEED) ;
   }

   fprintf(streamout, "//%c", LINE_FEED);
   return;   
}


/* ----------------------------------------------- */
/* ----------------------------------------------- */
/* Programme principal				   */
/* ----------------------------------------------- */
/* ----------------------------------------------- */

main(argn, argv)
   int  argn;
   char *argv[];

{


   FastaSequence	*seqFasta ;
   FILE			*streamin,*streamout ;
   int			shtimes;
   char			arg,optarg;




#ifdef THINK_C
   argn = __ccommand(&argv);
#endif

			/* parametres par defaut	*/

	streamin  = stdin;
	streamout = stdout;
	shtimes = 10;

           NOTIFY("-------------------------------------------------\n");
           NOTIFY("   Atelier de BioInformatique\n\n");
           NOTIFY("   Shuffle2 : Shuffle une banque Fasta avec\n");
           NOTIFY("   conservation des frequences des dinucleotides\n");
        FMTNOTIFY("   last compiled: %s\n", DATE);
        FMTNOTIFY("              on: %s\n", HOST);
           NOTIFY("-------------------------------------------------\n");



	if (argn != 2)
		Erreur("usage : shuffle2 fasta_file [> output_file]", 1);

	if ( !(streamin = fopen (argv[1],"r") ) )
		Erreur("cannot open fasta source file", 14);

	srandom(2 * (long) time(NULL) + 1);

	seqFasta = NewFastaSequence();

	
	while (ReadFastaSequence(streamin, seqFasta)) {
		shuffle2(seqFasta,shtimes);
		SortieSeq(seqFasta,stdout);
	}

	FreeFastaSequence(seqFasta);

     BANNER("--------------------------------------------\n");
  FMTBANNER("%s", StrCpuTime(TIME_RESET));
     BANNER("--------------------------------------------\n");

  exit (0);

}
/* fin de la fonction  main */
