/* ------------------------------------- */
/* fichier shuffleC.c			 */
/* programme pour permutter des sequenc. */
/* en conservant enventuellement des	 */
/* morceaux en 5' et 3'			 */
/* -n<entier> :  nombre de sequences	 */
/*               aleatoires / seq bio	 */
/* -k	      :  permutation dans 	 */
/*               partie degeneree	 */
/* -c         :  commentaire debut seq   */
/* ------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "../Param/param.h"
#include "../../Include/gtypes.h"
#include "../Include/libsys.h"

#define MAX_RND 0x7fffffff      /* plus grand entier aleatoire		*/
#define NB_BASES 50	   	/* nb base sur une ligne		*/
#define LONG 500		/* longueur max d'une sequence		*/
#define LINE_FEED 10		/* retour chariot			*/
#define ALPHABET "ACGTN"	


static void ARGUMENT(int* a, char b, char c[])
{
     if (c[0] != '0') 
     {
          *a = atoi(c) ;
          if (!a) 
          {
		fprintf(stderr,"Mauvais argument -%c %s\n",b,c); 
                exit(0) ;
	  }
     }
}		 



static void Permutte(int part1, int part2, char Sequence1[], int LongSeq,
 	      int nA, int nT, int nG, int nC, int nN) 
{
   int i             ;   		/* position base dans sequence	*/
   int c             ; 	  		/* base				*/
   int lA            ;
   int lT            ;
   int lG 	     ;
   int lC 	     ;
   int lN            ;
   long   x          ; 
    
   Sequence1[LongSeq] = 0 ;

 

   for (i = part1 ; i <= part2 ; i++)
   {

	lA = nA       ;			/* limite des A dans espace de	*/
					/* proba [1,lN]			*/
   	lT = lA + nT  ;
   	lG = lT + nG  ;
   	lC = lG + nC  ;
   	lN = lC + nN  ;
					/* tirage au sort de la base  	*/	
	x = random() ;
	c = (int) (1 + 1.0 * (x - 0) * (lN-1) / (MAX_RND-0)) ;

        if 	(c <= lA)
	{
		Sequence1[i] = 'A' ;
		nA--  ;
	}	
 	else if (c <= lT)
	{
		Sequence1[i] = 'T' ;
		nT -- ;
	}	
	else if (c <= lG)
	{
		Sequence1[i] = 'G' ;
		nG--  ;
	}	
	else if (c <= lC)
	{
		Sequence1[i] = 'C' ;
		nC-- ;
	}	
	else if (c <= lN)
	{
		Sequence1[i] = 'N' ;
		nN-- ;
	}	
			
   }
   
    
}


/* bascule booleenne	*/
static void inverse(int* i) 
{
     if (*i) 
	*i = 0 ;
     else
	*i = 1 ;
    
}


main(int argc, char * argv[])  
{
 int 	LongSeq 	; /* longueur des sequences			*/
 char   Seq[LONG]	; /* sequence originale				*/
 char   Sequence1[LONG]	; /* sequence permuttee				*/
 int    j		; /* indice de base				*/
 char 	c		; /* base					*/	 
 char   Symb            ; /* symbole argument de main 			*/
 char   Argument[50]    ; /* argument de main	     			*/
 int    nA		; /* nb bases A dans une sequence		*/
 int    nT		; 
 int    nG		;
 int    nC		;
 int    nN		; 
 int    N = 0		; /* nb de sequences				*/
 int    Longueur        ;
 int    repetition = 1	; /* nb sequences shufflees par sequence bio	*/
 int    r		; /* indice de repetition d'une meme sequence	*/
 int    NSorties = 0    ; /* nb sequences en sortie			*/
 int    variables	; /* nb nucleotides variables			*/
 int    partie_variable	; /* 1 si dans partie degeneree de l'ARN	*/
 char*  base		; /* caractere lu est une base			*/
 int    part1  	  = -1	; 
 int 	part2 	  = -1 	;
 int    Comment   = 0   ;
 int    Distingue = 0   ;
 char   Commentaire[LONG] ; /* Commentaire du fichier			*/
 int    Debut = -1	;   /* debut partie variable			*/
 int    Fin   = -1      ;   /* fin partie variable			*/

  fprintf(stderr,"debut du programme shuffle \n")   ;
   
					/* Initialisation		*/
  srandom((int)time(NULL)) ;
  Commentaire[0] = 0 ;

    
  while (SelectArg("-n -k -c",&Symb,Argument,argc,argv)) 
  {
	  switch (Symb) 
          {
		 case 'n' : ARGUMENT(&repetition,Symb,Argument)      ;
			    break	                             ;	
		 case 'k' : Distingue = 1		             ;
			    break	                             ;	
	 	 case 'c'  : Comment = 1			     ;
			     break				     ;				  
	  }
  }

 fprintf(stderr,"nombre de repetitions %d\n",repetition) ; 

 if (Comment)
     fscanf(stdin,"%s", Commentaire) ;

 while (!feof(stdin))
  {
						/* lecture de la sequence courante	*/
	LongSeq   = 0 ;
	Longueur  = 0 ;
        variables = 0 ;

	fscanf(stdin,"%d",&LongSeq) ;
		
        nA 		= 0 ; 
	nT 		= 0 ; 
	nG 		= 0 ; 
	nC 		= 0 ;
	nN		= 0 ;
	partie_variable = 0 ;

	while (Longueur < LongSeq - 2)
	{
	 	fscanf(stdin,"%c",&c)  ;
	        
		if (c == '?')
		{
		   inverse(&partie_variable) ;
		   fscanf(stdin,"%c",&c)    ;

		   if (partie_variable)
		       Debut =  Longueur ;
		   else
		       Fin   =  Longueur-1 ;
		}

		if (partie_variable || !Distingue)
 		{
	 	   switch (c) 
                   {
		 		case 'A' :  nA++  ;
			    		    break ;
				case 'T' :  nT++  ;     	  	
					    break ;
		 		case 'G' :  nG++  ;	      	  	 
					    break ;
				case 'C' :  nC++  ;	      	
					    break ;
			        case 'N' : nN++   ;
	            }
		
			variables++ ;
		 }
		
		base = strchr(ALPHABET,c) ;

		if (base) 
		{
		  if (!partie_variable && Distingue)
			 Sequence1[Longueur] = c ; 
		    Longueur++ 			 ; 
		}     	 								
	}

	fscanf(stdin,"%s",Seq)	;		/* lecture du #				*/
		
	if (Longueur > 0)
	{
	    N++ ;
	    fprintf(stderr,"%d\n",N) ;
 
            if (!Distingue)
            {
		part1 = 0           ;
		part2 = Longueur -1 ;
            }
            else
            {
		part1 = Debut ;
                part2 = Fin   ;
 	    }

	    for (r = 1 ; r <= repetition ; r++)
	    {
	   	Permutte(part1, part2, Sequence1, Longueur, nA, nT, nG, nC, nN) ;

				/* ecrit la nouvelle sequence	*/

		fprintf(stdout,"%d\n",Longueur)                        ;   
  
                if (Distingue)
		{
  		    fprintf(stdout,"%d %d\n",Debut+1, Fin+1)           ;		
		
		}

		for (j = 1  ; j <= Longueur ; j++)
		{
			fprintf(stdout,"%c",Sequence1[j-1])            ;
			if ((j % NB_BASES) == 0) 
	           		fprintf(stdout,"\n") ;
		}
	
		if ((Longueur % NB_BASES) != 0) 
	        	fprintf(stdout,"\n") ;		               

		fprintf(stdout,"#\n") ;

	        NSorties++ ;
	    }
	}
 }

  fprintf(stderr,"nombre de sequences traitees  : %d\n",N) ;
  fprintf(stderr,"nombre de sequences en sortie : %d\n", NSorties) ;
  fprintf(stderr,"fin du programme\n") ;	

  return 1 ;
}

